/**
 * Interface des classes qui implémentent le code d'un agent ubiquitaire
 * 
 * Définition : 
 * 	-> un agent ubiquitaire est un agent dont :
 * 		-le support des interactions (capteurs/actuateurs) sont des dispositifs physiques diffusés dans l'envrionnement (c'est à dire des devices au sens de UDA).
 * 		-le support des interactions sociales (entre agent) sont des avatars (au sens de Ubiquity)
 * 
 * 	--> au niveau de l'agent chaque support à un interacteur physique ou social est appelé un slot : "slot physique" pour un interacteur physique (device) et
 * "slot social" pour un interacteur social (avatar).
 *  --> l'affectation d'un slot physique à un device est établie par des règles dites "Plug rules". Ces règles permettent à tout instant
 *  d'identifier le device le plus approprié qui doit être associé au slot, c'est pourquoi on appelle ces liaisons slot->device des "liaisions 
 *  opportunes".
 *  --> le code de l'agent est écrit pour fonctionner avec 0, 1 ou plusieurs avatars. Chacun de ces avatar est identifié par le nom du slot social auquel il correspond.
 * 
 * L'agent peut être dans 3 états :
 *  - OFF : l'agent n'est pas instancié : tous les slots sociaux de l'agent sont déconnectés, aucun pulsor n'existe.
 *  	Les slots physiques qui valide au moins une règle sont virtuellement pluggué dans le sens ou l'agent n'est pas
 *  	notifié des plug/unplug, ni des changements d'état et des changements de valeurs des sensors. 
 *  - RUN : l'agent est en cours d'exécution : tous ses pulses sont actifs, les slots sociaux sont tous pluggué (tous les avatars 
 *  		sont créé), et tous slots physiques validant au moins une règle opportune sont pluggué, l'agent reçoit les changement d'état;
 *  		les notifications plug/unplug et les changements de valeurs des sensors.
 *  		NB: si un local_id est spécifié pour un slot social, et qu'un participant de même id local existe au moment ou le plug social doit être
 *  		établie, alors le plug ne sera pas effectif et le deviendra automatiquement si le local_id se libère (c.a.d si le participant occupant ce local_id disparait).
 *  - PAUSE : l'agent est instancié, mais il n'est pas en cours d'exécution dans le sens ou tous ses pulsors sont mis en pause,
 *  	 les slots physiques validant au moins une règle opportune sont pluggué, l'agent recoit les changements d'état
 *  	 et les notifications plug/unplug de ces slots, mais pas les notifications des changements de valeurs des sensors.
 *  	 Tous les slots sociaux de l'agent sont déconnectés.
 * 
 * NB: les classes filles doivent indiquer la spécification des slots de l'agent en définisant :
 * 	- une propriété public final static de nom physical_specif et de type String
 *  - une propriété public final static de nom social_specif et de type String
 *  Le format de ces chaines sont les suivants :
 * 	
 * 			 public final static String physical_specif =
 * 						"name_slot1 : type_device1 ;" 				//pour un slot physique non optionnel
 * 					 +  "name_slot2 : type_device2 , optional ;" 	//pour un slot physique optionnel
 * 		  	 ;  
 * 
 * 			 public final static String social_specif =
 * 						"name_slot1 : space_name ;" 			//pour un slot social sans local_id spécifique
 * 					 +  "name_slot2 : space_name , local_id ;" 	//pour un slot avec local_id spécifique
 * 
 * 			;
 * 
 * 		NB: type_deviceX est :
 * 					- soit un entier, qui donne donc directement la valeur du code du type de device
 * 					- soit le nom d'une constante de type de device UDA (par exemple : LIGHT pour la constante UDA.LIGHT)
 * 
 * 		NB: il n'est pas possible d'utiliser un space privé (avec mot de passe) dans la spécification des slots sociaux de l'agent. Ces types de spaces
 * 		doivent être fixé dynamiquement en configurant l'agent.
 * 
 * En cas d'absence de ces spécifications, l'agent sera considéré comme n'ayant aucun slot (donc aucun device et aucun avatar).
 */
package skuad.agentu;

import skuad.uda.*;
import skuad.ubiquity.space.*;
import skuad.ubiquity.space.mailbox.*;
import skuad.util.time.Pulse;
import skuad.util.value.Value;


public interface AgentU {
		
	//Constante de réponse ( [<0 ; 10]	
	public static byte NOT_POSSIBLE = -2;	//action impossible
	public static byte ERROR = -1;			//erreur critique
	public static int OK = 1;	
	
	//Constantes d'état
	public static byte NOT_EXIST = 0;		//l'agent n'existe pas
	public static byte OFF = 1;				//arrété (non instancié) -> les slots ne sont pas associé aux devices
	public static byte PAUSE = 2;			//en pause : plus de pulse behavior, les avatars sont déconnecté et plus de notification sensorValue. -> les slots restent associés aux devices
	public static byte RUN = 3;				//en cours d'exécution (aprés un start ou un resume)
	
	//Constantes d'évènement
	public static byte PAUSED = 10;		//évènement pause
	public static byte RESUMED = 11;		//évènement reprise aprés pause
	public static byte ALL_PLUGGED = 12;	//évènement : tous les plugs non optionels sont établies
	public static byte MISSING_PLUG = 13;//évènement : rupture de ALL_PLUGGED
	
	//Constantes des modes d'exécution 
	public static byte MANUAL = 20;
	public static byte ALLWAYS_RUN = 21;
	public static byte BREAKPLUG_STOP = 22;
	public static byte BREAKPLUG_PAUSE = 23;
	
	//Constantes de contextualisation des plug/unplug
	public static byte OP_START = 30;
	public static byte OP_RESUME = 31;
	public static byte OP_PAUSE = 32;
	public static byte OP_STOP = 33;	
	public static byte OP_REMOVE = 34;
	public static byte OP_SPACE_CHANGE = 35;
	public static byte OP_LOCAL_ID_RELEASE = 36;
	//]
	
	//Constantes de contextualisation des changements d'évènement des agents (dernier argument de la méthode AgentUObserver.agentStateChange(...))
	public static byte RUN_MODE_UNPLUG = 40;
	public static byte RUN_MODE_PLUG = 41;
	public static byte DISTANT_ORDER = 42;
	public static byte LOCAL_ORDER = 43;
	public static byte INIT_SERVER = 44;
	public static byte SHUTDOWN = 45;
	public static byte PLUG_NOTIFICATION = 46;
	public static byte UNPLUG_NOTIFICATION = 47;
	//]
	
	//[Utilitaire
	public static String stateToString(byte vstate) {
		switch(vstate) {
		case AgentU.NOT_EXIST: return("NOT_EXIST");
		case AgentU.OFF: return("OFF");
		case AgentU.PAUSE: return("PAUSE");
		case AgentU.RUN: return("RUN");
		}
		return("- unknow -");
	}
	
	public static String ctxToString(byte vctx) {
		switch(vctx) {
		case AgentU.RUN_MODE_UNPLUG : return("RUN_MODE_UNPLUG");
		case AgentU.RUN_MODE_PLUG : return("RUN_MODE_PLUG");
		case AgentU.DISTANT_ORDER : return("DISTANT_ORDER");
		case AgentU.LOCAL_ORDER : return("LOCAL_ORDER");
		case AgentU.INIT_SERVER : return("INIT_SERVER");
		case AgentU.SHUTDOWN : return("SHUTDOWN");
		case AgentU.PLUG_NOTIFICATION : return("PLUG_NOTIFICATION");
		case AgentU.UNPLUG_NOTIFICATION : return("UNPLUG_NOTIFICATION");
		}
		return("- unknow -");
	}	
	//]
		
	//Cycle de vie de l'agent
	public void start(AgentUDescriptor aud, int ctx);	//invoqué au démarrage de l'agent
	public void stop(AgentUDescriptor aud, int ctx);	//invoqué lors de la mise à l'arret de l'agent
	
	public void event(AgentUDescriptor aud, int code, int ctx); //code = PAUSED / RESUMED / ALL_PLUGGED / MISSING_PLUG
	
	//Exécution du comportement de l'agent
	public void behavior(AgentUDescriptor aud, Pulse pulse); //exécute le comportement de l'agent suite à l'activation d'un pulse, pulse = le descripteur de la pulsation qui a initié cet appel	
			//NB: ces déclenchements ne peuvent avoir lieu que si l'agent est dans l'état RUN
	
	//Devices attachés à l'agent
		//évènement d'attachement de device
	public void notifyPlugDevice(AgentUDescriptor aud, String slot, Device device);
	public void notifyUnplugDevice(AgentUDescriptor aud, String slot);
	public void notifyChangeDevice(AgentUDescriptor aud, String slot, Device device);
	
	//Ecoute des devices
		//NB: stateDevice est déclenché que si l'état de l'agent est RUN ou PAUSE
	public void stateDevice(AgentUDescriptor aud, String slot, Device dev, boolean ready, boolean config_changed); //invoqué dès que l'état du device évolue
		//NB: les sensorChange ne sont déclenché que si l'état de l'agent est RUN
	public void sensorChange(AgentUDescriptor aud, String slot, int sensor_id, Value value);

	/**
	 * Value = HASH
	 * 	- label : HASH
	 * 		state : int			//SenSorTarget.OK // SenSorTarget.NO_PLUG
	 * 		slot : String		//nom du slot physique 
	 * 		sid : int			//sensor id
	 * 		value : int / float / byte[]	
	 */
	public void sensorReading(AgentUDescriptor aud, String query_label, Value value);

	
	//Avatars associés à l'agent
	public void notifyPlugAvatar(AgentUDescriptor aud, String name, Avatar avatar, int contexte);
	public void beforeUnplugAvatar(AgentUDescriptor aud, String name, Avatar avatar, int contexte); //change_space==true s'il s'agit d'un changement d'avatar
	public void notifyUnplugAvatar(AgentUDescriptor aud, String name, int contexte);
	public void receive(AgentUDescriptor aud, Avatar avatar, MailBox box, Message mess);
	
	//[A supprimer
	/**
	 * @deprecated remplacé par sensorChange(...)
	 */
	@Deprecated
	public void sensorValue(AgentUDescriptor aud, String slot, Device dev, int sensor_id, int value);
	/**
	 * @deprecated remplacé par sensorChange(...)
	 */
	@Deprecated
	public void sensorValue(AgentUDescriptor aud, String slot, Device dev, int sensor_id, float value);
	/**
	 * @deprecated remplacé par sensorChange(...)
	 */
	@Deprecated
	public void sensorValue(AgentUDescriptor aud, String slot, Device dev, int sensor_id, byte[] value);
	//]
}