package skuad.hubudp;

import skuad.util.databytes.*;


/**
* A <em>Hub</em> is the object which represent the <em>overlay network</em> that connects network nodes, called <em>spots</em>.
* This <em>Hub</em> can be in different states : <ul>
* 	<li>UNKNOW 			= 0</li>
*	<li>HUB_READY 		= 1</li>
*	<li>PORTAL_ENABLED 	= 2</li>
*	<li>PORTAL_DISABLED = 3</li>
*	<li>MANAGER_ADDED 	= 4</li>
*	<li>MANAGER_REMOVED = 5</li>
*	<li>LOG_WARNING 	= 6</li>
*	<li>LOG_ERROR 		= 8</li>
*	<li>LOG_INFO 		= 9</li>
*	<li>INITIALIZING	= 10</li>
*	<li>TRY_JOIN_PORTAL	= 11</li>
*	<li>HUB_BREAK 		= 12</li>
*	<li>HUB_ERROR 		= 13</li>
*	<li>BAD_KEY 		= 14</li>
*</ul>
* @see Spot
* @see LocalSpot
*/

public abstract class Hub {
	//[Constante d'option pour l'envoie des messages
	public final static byte NONE			= 0;
	public final static byte ALL			= (byte) 0xFF;
	public final static byte NEED_ACK		= 1;
	public final static byte AUTO_DISPOSE	= 1 << 1;
	public final static byte SERVE_LOCAL	= 1 << 2;
	public final static byte NO_SERVE_LOCAL	= NEED_ACK | AUTO_DISPOSE;
	//]
	
	//[Constante pour la valeur event des méthodes HubListener
	public final static byte UNKNOW 			= 0;	
	public final static byte HUB_READY 			= 1;	//-> args = hub name
	public final static byte PORTAL_ENABLED 	= 2;
	public final static byte PORTAL_DISABLED 	= 3;
	public final static byte MANAGER_ADDED 		= 4;	//-> args = manager name
	public final static byte MANAGER_REMOVED 	= 5;	//-> args = manager name
	public final static byte LOG_WARNING 		= 6;	//-> args = message log
	public final static byte LOG_ERROR 			= 8;	//-> args = message log
	public final static byte LOG_INFO 			= 9;	//-> args = message log
	public final static byte INITIALIZING		= 10;
	public final static byte TRY_JOIN_PORTAL	= 11;
	public final static byte HUB_BREAK 			= 12;
	public final static byte HUB_ERROR 			= 13;
	public final static byte BAD_KEY 			= 14;
	public final static String DEFAULT_HUB		= "default";	//-> args = hub name
	public static String eventName(byte code) {
		switch(code) {
		case UNKNOW :			return("UNKNOW");
		case HUB_READY :		return("HUB_READY");
		case PORTAL_ENABLED :	return("PORTAL_ENABLED");
		case PORTAL_DISABLED :	return("PORTAL_DISABLED");
		case MANAGER_ADDED :	return("MANAGER_ADDED");
		case MANAGER_REMOVED :	return("MANAGER_REMOVED");
		case LOG_WARNING :		return("LOG_WARNING");
		case LOG_ERROR :		return("LOG_ERROR");
		case LOG_INFO :			return("LOG_INFO");
		case INITIALIZING :		return("INITIALIZING");
		case TRY_JOIN_PORTAL :	return("TRY_JOIN_PORTAL");
		case HUB_BREAK :		return("HUB_BREAK");
		case HUB_ERROR :		return("HUB_ERROR");
		case BAD_KEY :			return("BAD_KEY");
		default : return("-undefined-");
		}
	}
	//]
	
	protected byte state;
	
	public Hub() {
		state = Hub.UNKNOW;
	}
	
	/**
	* To get the <em>state</em> of the <em>Hub</em>. 
	* Those States and byte codes are described in the <em>Hub</em> definition.
    * @return byte
    * 			 The code that corresponds to the current state of the Hub
    * @see Hub
    *
    */
	public byte getState() { return(state); }
	/**
	* To get the <em>name</em> of the <em>Hub</em>. 
    * @return String
    * 			 The name of the Hub
    * @see Hub
    *
    */
	public abstract String getHubName();
	public abstract boolean isPrivate();
	/**
	* Will be called the next time when the <em>Hub</em> is in the state READY 
	* Immediatly called if the <em>Hub</em> is already ready.
    * @param HubReadyListener
    * 		The Listener that handle the READY event.
    * @see Hub
    *
    */
	public abstract void whenReady(HubReadyListener hrl); //hrl.ready(...) sera déclenché la prochaine fois que le hub passera à l'état ready (immédiatement si le hub est déjà dans cet état).
	
		//envoies de message en mode broadcast
	/**
	* To send a message, with type <em>DataReader</em> to all <em>Spots</em> of the <em>Hub</em> (broadcast).
	* @param code_protocole
    *            The code of the used protocol to send the message
    * @param content
    *            The content of the Data Reader type message to send
    * @param option (OPTIONAL)
    *            It's possible to add options to that message
    * @return boolean
    * 			 TODO
    * @see Spot
    *
    */
	public boolean broadcast(int channel, DataReader content){ return(broadcast(channel, content, ALL)); }	//-> emission du message en mode broadcast sur l'ensemble du hub, avec toutes les options activées par défaut
	/**
	* To send a message, with type <em>DataReader</em> to all <em>Spots</em> of the <em>Hub</em> (broadcast).
	* @param code_protocole
    *            The code of the used protocol to send the message
    * @param content
    *            The content of the DataReader type message to send
    * @param option (OPTIONAL)
    *            It's possible to add options to that message
    * @return boolean
    * 			 TODO
    * @see Spot
    *
    */
	public abstract boolean broadcast(int channel, DataReader content, int option);	//-> emission du message en mode broadcast sur l'ensemble du hub, dispose() sera appliqué à l'instance content que si auto_dispose==true

	/**
	* To send a message, with type <em>Byte</em> to all <em>Spots</em> of the <em>Hub</em> (broadcast).
	* @param code_protocole
    *            The code of the used protocol to send the message
    * @param content
    *            The content of the byte type message to send
    * @param option
    *            It's possible to add options to that message
    * @return boolean
    * 			 TODO
    * @see Spot
    *
    */
	public abstract boolean broadcastByte(int channel, byte content, int option);	//-> emission du message ne contenant qu'un byte sur l'ensemble du hub
	
	/**
	* To send a message, with type <em>byte buffer</em> to all <em>Spots</em> of the <em>Hub</em> (broadcast).
	* @param code_protocole
    *            The code of the used protocol to send the message
    * @param buff
    *            The byte buffer that you want to send
    * @param option
    *            It's possible to add options to that message
    * @return boolean
    * 			 TODO
    * @see Spot
    *
    */
	public boolean broadcastBuff(int channel, byte[] buff){	//-> emission du buffer en mode broadcast sur l'ensemble du hub, avec toutes les options activées par défaut
		return(broadcastBuff(channel, buff, 0, buff.length, ALL));
	}
	/**
	* To send a message, with type <em>byte buffer</em> to all <em>Spots</em> of the <em>Hub</em> (broadcast).
	* @param code_protocole
    *            The code of the used protocol to send the message
    * @param buff
    *            The byte buffer that you want to send
    * @param option
    *            It's possible to add options to that message
    * @param offset
    *            The offset of the byte buffer that you want to send
    * @param size
    *            The size of the buffer that you want to send
    * @return boolean
    * 			 TODO
    * @see Spot
    *
    */
	public abstract boolean broadcastBuff(int channel, byte[] buff , int offset, int size, int option); //-> emission du buffer en mode broadcast sur l'ensemble du réseau

		//gestion des spots
	/**
	* To get the descriptor of the <em>Local Spot</em>.
    * @return LocalSpot
    * 			 The descriptor of the Local Spot.
    * @see Spot
    *
    */
	public abstract LocalSpot localSpot();		//-> retourne le descripteur du spot local
	/**
	* To assign a <em>SpotObserver</em> to a <em>Hub</em>, which will notify <em>SMD</em> changes and UP <em>events</em>.
	* @param observer
    *            The observer you want to assign to the Hub.
    * @see SpotObserver
    * @see Hub
    *
    */
	public abstract void addSpotObserver(SpotObserver observer); //dès son enregistrement l'observer recoit tous les évènements up pour chaque spot UP et les IdContext déjà produit sur ces spots, ainsi qu'un updateSMD(spot, true) pour les spots qui ont déjà inintialisés leur SMD
	/**
	* To unassign a <em>SpotObserver</em> to a <em>Hub</em>.
	* @param observer
    *            The observer you want to unassign.
    * @see SpotObserver
    */
	public abstract boolean removeSpotObserver(SpotObserver observer);
	public abstract void explore(SpotExplorer cb);
	
	/**
	* To get the <em>Spot</em> object by providing the Target ID
	* @param idTarget
    *            The Target ID of the spot you want to get
    * @return Spot
    * 			 The Spot you want to get
    * @see Spot
    *
    */
	public abstract Spot getSpotByIdTarget(String idTarget); 			//retourne le descripteur du spot correspondant à l'id idTarget indiqué, null si non trouvé dans ce hub
	/**
	* To get the <em>Spot</em> object by providing the Local ID
	* @param local_id
    *            The Local ID of the spot you want to get
    * @return Spot
    * 			 The Spot you want to get
    * @see Spot
    *
    */
	public abstract Spot getSpotByIdLocal(int local_id);				//retourne le descripteur du spot correspondant à l'id local indiqué, null si non trouvé dans ce hub
	/**
	* To get the <em>Spot</em> object by providing the Global ID
	* @param global_id
    *            The Global ID of the spot you want to get
    * @return Spot
    * 			 The Spot you want to get
    * @see Spot
    *
    */
	public abstract Spot getSpotByIdGlobal(String global_id);			//retourne le descripteur du spot correspondant à l'id global indiqué, null si non trouvé dans ce hub
	/**
	* To get the <em>Spot</em> object by providing the Context ID
	* @param id_context
    *            The Context ID of the spot you want to get
    * @return Spot
    * 			 The Spot you want to get
    * @see Spot
    *
    */
	public abstract Spot getSpotByIdContext(String chaine_contexte, String id_context);
	
		//administration du hub
	/**
	* To assign a <em>Message Receiver</em>, that will handle received messages in the given channel, to the <em>Hub</em>.
	* @param channel
    *            The message channel to handle 
    * @param name
    *            The name of the receiver you assign
    * @param manager
    *            The Message Receiver that you want to assign to the Hub
    * @see Hub
    * @see MessageReceiver
    *
    */
	public abstract boolean addReceiver(int channel, String name, MessageReceiver manager);		//ajoute un gestionnaire de réception pour le spot local à ce hub (-> manager.receive(sender, content_paquet, channel) sera invoqué sur chaque paquet recu,)
	/**
	* To unassign a <em>Message Receiver</em> of the <em>Hub</em>.
	* @param channel
    *            The message channel to handle 
    * @param name
    *            The name of the receiver you assign
    * @param manager
    *            The Message Receiver that you want to unassign
    * @see Hub
    * @see MessageReceiver
    *
    */
	public abstract boolean removeReceiver(int channel, String name, MessageReceiver manager);	//supprime le gestionnaire précédemment enregistré

	/**
	* To connect the <em>Hub</em> to a <em>Portal</em>
	* @param ip_portal
    *            The IP address of the portal to connect to the Hub 
    * @param port_portal
    *            The port of the portal to connect to the Hub
    * @param link_key
    *            The key to provide to validate the link
    * @return boolean
    * 			 TO-DO
    * @see Hub
    * @see MessageReceiver
    *
    */
	public abstract boolean linkPortal(String ip_portal, int port_portal , String link_key);//<- relie le hub au portail ip_portal:port_portal, la bonne valeur de link_key doit être fournit.
//	public abstract boolean localPortal(int port_portal);	//<- crée un portail relié au hub "hub", le spot local devient alors immédiatement le spot relais (cela revient à enchainer les appels HubUdp.newPortal(...) et hup.linkPortal(...) dans le même contexte d'exécution).
//	public abstract boolean fusion(Hub hub_b);
	//]
}