package skuad.hubudp;

import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import skuad.util.databytes.*;
import skuad.util.thread.ThreadStack;
/**
 * 	A <i>Spot</i> is the representation of an execution context in a <i>Hub</i>. It can be defined
 *  by 2 types of Metadatas : SMD (Small MetaDatas), which are used to store essentials and simplest IDs. TAG are more 
 *  complicated MetaDatas which can describe the <i>Spot</i> more precisely.<br/><br/>
 *  
 *	<ul>About SMD ranges :
 *		<li> <b>01 to 20</b>: Automatically Provided <i>(1=Global_ID; 2=OS; 3=Soft)</i>
 *		<li> <b>21 to 40</b>: General Informations <i>(21=HostName; 22=SpotName; 23=Hardware)</i>
 *		<li> <b>41 to 60</b>: Additional Protocols <i>(41=TAG; 42=Device; 43=AgentU Server; 44=Ubiquity)</i>
 *		<li> <b>61 to 80</b>: Services <i>(61=Web Server)</i>
 *	</ul>
 */
public abstract class Spot {

	//[Constante SMD
	//Renseigné automatiquement (1-20)
	public final static int GLOBAL_ID = 1;		//chaine aléatoire (8 lettres alphanumérique minuscule/majuscule, soit +218 000 milliards de combinaison)
	//NB: c'est un identifiant unique, fixé systématiquement dans le SMD d'un spot, il est garanti unique 
	//(éventuellement actualisable en cas de conflit) mais qui n'est pas persistant (change d'une exécution à l'autre)
	//--> pour un id persitant il faut utiliser les idContext
	public final static int OS 		=	2;		//famille de l'os du porteur , valeur = [code_OS]
	public final static int SOFT 	=	3;		//langage logiciel du porteur , valeur = [code_soft]

	//Informations générales	(21-40)
	public final static int HOST_NAME = 21;		//nom de la machine porteuse, valeur = {...name...}
	public final static int SPOT_NAME =	22;		//nom de du spot, valeur = {...name...}
	public final static int HARDWARE  = 23;		//information sur l'hardware, valeur = [code_HARDWARE]

	//Protocole complémentaire	(41-60)
	public final static int TAG 			= 41;	//support du protocole TAG ,   valeur = [1] ssi le protocole TAG est supporté
	public final static int DEVICE 			= 42;	//device UDA UDP exposé, valeur = [device_id]
	public final static int SERVER_AGENTU 	= 43;	//support SERVER_AGENTU disponible si valeur == 1
	public final static int UBIQUITY	 	= 44;	//support Ubiquity disponible si valeur présente, elle donne la valeur md5key de l'hote

	//Service					(61-80)
	public final static int SWEB 	=	61;		//webserver actif, valeur = [port_h][port_f]  (+d'info en TAG: name = "sweb" , value = { port : name : [url:, image:, label:]  } ) 	
	//<100
	//]

	//Valeur pour code_OS
	public final static int LINUX 	= 1;
	public final static int MAC 	= 2;
	public final static int WINDOWS = 3;
	public final static String decodeOs(int code, String unknow) {
		switch(code) {
		case LINUX : return("Linux");
		case MAC : return("Mac");
		case WINDOWS : return("Windows");
		default: return(unknow);
		}
	}
	
	//Valeur pour le code_soft
	public final static int JAVA 		= 1;
	public final static int JS	 		= 2;
	public final static int C_ARDUINO	= 3;	
	public final static String decodeSoft(int code, String unknow) {
		switch(code) {
		case JAVA : return("Java");
		case JS : return("JS");
		case C_ARDUINO : return("C_Arduino");
		default: return(unknow);
		}
	}

	
	//Valeur pour code_HARDWARE
	public final static int RPI 	= 1;		//c'est une machine raspberry pi
	public final static int ESP8266 = 2;		//c'est une plateforme ESP8266
	public final static String decodeHardware(int code, String unknow) {
		switch(code) {
		case RPI : return("RPi");
		case ESP8266 : return("ESP8266");
		default: return(unknow);
		}
	}
	//]

	protected Vector<SpotObserver> observers;		//les observers de spot
	private int ctarget;
	protected Map<String, String> idContexts;	//map des idContext du spot 
	protected String global_id;
	protected boolean is_up;
	private String idTarget;
	
	protected Spot(int vct) {
		ctarget = vct;
		observers = new Vector<SpotObserver>();
		idContexts = new HashMap<String, String>();
		global_id=null;
		is_up=false;
		idTarget = null;
	}
	protected int getCTarget() { return(ctarget); }
	
	protected void changeIdTarget(int ct){
	    this.ctarget = ct;
	    this.idTarget = null;
	}
	
	/**
	 * Get the <i>Hub</i> where the <i>Spot</i> is registered.
	 * @return <b>Hub</b> The <i>Hub</i> of the <i>Spot</i>.
	 */
	public abstract Hub getHub();
	/**
	 * To know if the spot is <em>local</em> or not. If it is, the <i>Spot</i> will return <b>true</b>.
	 * @return <b>boolean</b>
	 * 			Return <em>true</em> if the spot is local. Return <em>false</em> if not.
	 */
	public boolean isLocalSpot() { return(false); }	//true s'il s'agit de le spot local
	/**
	 * To know if the <i>host</i> of the <i>Spot</i> is <em>local</em> or not. In the case of a <i>
	 * Local Spot</i>, it will return <b>true</b>.
	 * @return <b>boolean</b>
	 * 			Return <em>true</em> if the host is local. Return <em>false</em> if not.
	 */
	public abstract boolean isLocalHost(); //true s'il s'agit de l'hote local
	/**
	 * To know if the <i>Spot</i> is ready to communicate.
	 * @return <b>boolean</b> Return <i>true</i> if the <i>Spot</i> is ready.
	 */
	public boolean isUp() { return(is_up); }
	/**
	 * To get the <i>host IP Address</i> of the <i>Spot</i>.
	 * @return <b>String</b> Return the String that contains the <i>IP address</i>.
	 */
	public abstract String getIp();
	/**
	 * To get the <i>Port</i> of the <i>Spot</i> associated to the execution context.
	 * @return <b>int</b> Return the <i>integer</i> variable that contains the <i>Port</i>.
	 */
	public abstract int getPort();		
	/**
	 * To get the <i>IP Address and Port</i> of the <i>Spot</i> associated to the execution context.
	 * @return <b>String</b> Return the <i>String</i> variable that contains the addition of the <i>IP Address and Port</i>.
	 */
	public abstract String getIpPort();		//ip:port du contexte portant ce spot
	/**
 	* To add a <i>Spot Observer</i> on a <i>Spot</i>, that will notify UP events and SMD updates of this <i>Spot</i>.
 	* @param obs The <i>Spot Observer</i> that you want to assign.
	* @see SpotObserver
 	*/
	public void addSpotObserver(SpotObserver obs) { synchronized(observers) { 
		if(!observers.contains(obs)) {
			observers.add(obs);
			//on fire le updateSMLD(...) et les idContext(...) déjà produit
			if(readySMD()) ThreadStack.perform(() -> obs.updateSMD(this, true));
			exploreAllIdContext(obs);
		}
	}}
	/**
	 * 
	 * @param obs
	 */
	public void exploreAllIdContext(IdContextExplorer obs) {
		if(!idContexts.isEmpty()) synchronized(idContexts) {
			for(String ch_ctx : idContexts.keySet()) {
				ThreadStack.perform(() -> obs.idContext(this, ch_ctx, idContexts.get(ch_ctx)));
			}
		}		
	}
	/**
	 * Get the number of <i>Context ID</i> of the <i>Spot</i>.
	 * @return <b>int</b> Return the number of <i>Context ID</i>.
	 * @see SpotObserver
	 */
	public int nbIdContext() { return(idContexts.size()); }
	/**
	 * To unassign a <i>Spot Observer</i> of a <i>Spot</i>.
	 * @param obs The <i>Spot Observer</i> that you want to unassign.
	 * @return <b>boolean</b> Return <b>true</b> if the remove operation correctly happened.
	 * @see SpotObserver
	 */
	public boolean removeSpotObserver(SpotObserver obs){ synchronized(observers) { return(observers.remove(obs)); }	}

	
	//[les identifiants
	/**
	 * To get the <i>Target ID</i> of the <i>Spot</i>.
	 * @return <b>String</b> Return the <i>String</i> that contains the <i>Target ID</i>.
	 */
	public String getIdTarget() {	//id target du spot
		if(idTarget==null) idTarget = getIp()+'#'+ctarget;
		return(idTarget); 
	}
	/**
	 * To get the <i>Local ID</i> of the <i>Spot</i>.
	 * @return <b>int</b> Return the <i>integer</i> variable that contains the <i>Local ID</i>.
	 */
	public int getIdLocal() { return(ctarget); }		//id locale du spot
	/**
	 * To get the <i>Global ID</i> of the <i>Spot</i>.
	 * @return <b>String</b> Return the <i>String</i> that contains the <i>Global ID</i>.
	 */
	public String getIdGlobal(){ return(global_id); } 	//id globale du spot (null si cet identifant n'est pas encore disponible, il ne l'est qu'à partir de la première notification updateSMD(...)) 
	/**
	 * To get the <i>Context ID</i> of the <i>Spot</i>.
	 * @return <b>String</b> Return the <i>String</i> that contains the <i>Context ID</i>.
	 */
	public String getIdContext(String chaine_context){
		return(idContexts.get(chaine_context));
	}
	//]
	
	 //[lecture des méta-donnée SMD (Small Méta-Data)
	
	/**
	 * To know if the SMD of the <i>Spot</i> is available.
	 * @return <b>boolean</b> Return <b>true</b> if SMD of the <i>Spot</i> is available.
	 */
	public abstract boolean readySMD();								//	-> retourne true ssi les SMD du spot sont disponibles (première notification updateSMD(...) déjà réalisé)
	/**
	 * Get the number of occurrence for a given code of <i>SMD</i>.
	 * @param code The code of the SMD you want to count.
	 * @return <b>int</b> The number of occurrence for a code of <i>SMD</i>.
	 */
	public abstract int count(int code);							//	-> retourne le nombre d'occurence existante pour le code SMD indiqué
	
	/**
	 * Get the length of the <i>SMD</i> value for the corresponding code.
	 * @param code The SMD code whose length you want to know>.
	 * @param rang TODO
	 * @return <b>int</b> Return the length of the code.
	 */
	public abstract int length(int code, int rang);					//	-> retourne le nombre d'octet de la valeur du code SMD indiqué
	/**
	 * Get the <i>int</i> value of the <i>SMD</i> for the corresponding code.
	 * @param code The SMD code whose length you want to know.
	 * @param rang TODO
	 * @return <b>int</b> Return the value of the SMD for the corresponding code. Return 0 if there is no value, -1 if itsn't numerical value.
	 */
	public abstract int getInt(int code, int rang);					//	-> lecture de l'octet numéro 0 de la valeur du code SMD indiqué, 0 si non trouvé (-1 si type non numérique pour le localspot seulement)
	/**
	 * Verify if the <i>SMD</i>, in the given code, has the given value.
	 * @param code The SMD code whose value you want to verify.
	 * @param value The value that you want to verify.
	 * @return <b>boolean</b> Return <b>true</b> if the SMD code contains the given value.
	 */
	public abstract boolean haveInt(int code , int value);			//	-> true ssi le code SMD indiqué possède la valeur value, false si non trouvé
	/**
	 * Get the <i>String</i> value of the <i>SMD</i> for the corresponding code.
	 * @param code The SMD code whose length you want to know.
	 * @param rang TODO
	 * @return <b>String</b> Return the value of the SMD for the corresponding code. Return <i>null</i> if there is no value or itsn't <i>String</i> value.
	 */
	public abstract String getString(int code , int rang);			//	-> lecture de la valeur du code SMD indiqué sous la forme d'une chaine de caractère, null si non trouvé (null si type non String pour le localspot seulement)
	/**
	 * Get the <i>byte</i> buffer of the <i>SMD</i> for the corresponding code.
	 * @param code The SMD code whose length you want to know.
	 * @param rang TODO
	 * @return <b>byte</b> Return the buffer of the SMD for the corresponding code. Return <i>null</i> if there is no value or itsn't numerical value.
	 */
	public abstract byte[] getBuff(int code , int rang);			//	-> lecture du buffer de la valeur du code SMD indiqué, null si non trouvé (null si type non buffer pour le localspot seulement)
	/**
	 * Get the <i>int</i> value at the given buffer position of the <i>SMD</i> for the corresponding code.
	 * @param code The SMD code whose length you want to know.
	 * @param pos The position that you want to read in the buffer.
	 * @param rang TODO
	 * @return <b>int</b> Return the value of the SMD for the corresponding code. Return 0 if there is no value, -1 if itsn't numerical value.
	 */
	public abstract int readBuff(int code, int pos , int rang); 	//	-> lecture de l'octet pos de la valeur du code SMD indiqué, -1 si non trouvé (-1 si type non buffer pour le localspot seulement)
	/**
	 * Get the length of the <i>SMD</i> value for the corresponding code.
	 * @param code The SMD code whose length you want to know>.
	 * @return <b>int</b> Return the length of the code.
	 */
	public int length(int code){ return(length(code, 0));}
	/**
	 * Get the <i>int</i> value of the <i>SMD</i> for the corresponding code.
	 * @param code The SMD code whose length you want to know.
	 * @return <b>int</b> Return the value of the SMD for the corresponding code. Return 0 if there is no value, -1 if itsn't numerical value.
	 */
	public int getInt(int code){ return(getInt(code, 0));}
	/**
	 * Get the <i>boolean</i> value of the <i>SMD</i> for the corresponding code.
	 * @param code The SMD code whose length you want to know.
	 * @return <b>boolean</b> Return the value of the SMD for the corresponding code.
	 */
	public boolean getBool(int code){ return(getInt(code, 0)!=0);}
	/**
	 * Get the <i>String</i> value of the <i>SMD</i> for the corresponding code.
	 * @param code The SMD code whose length you want to know.
	 * @return <b>String</b> Return the value of the SMD for the corresponding code. Return <i>null</i> if there is no value or itsn't <i>String</i> value.
	 */
	public String getString(int code){ return(getString(code, 0));}
	/**
	 * Get the <i>byte</i> buffer of the <i>SMD</i> for the corresponding code.
	 * @param code The SMD code whose length you want to know.
	 * @return <b>byte</b> Return the buffer of the SMD for the corresponding code. Return <i>null</i> if there is no value or itsn't numerical value.
	 */
	public byte[] getBuff(int code){ return(getBuff(code, 0));}
	/**
	 * Get the <i>int</i> value at the given buffer position of the <i>SMD</i> for the corresponding code.
	 * @param code The SMD code whose length you want to know.
	 * @param pos The position that you want to read in the buffer.
	 * @return <b>int</b> Return the value of the SMD for the corresponding code. Return 0 if there is no value, -1 if itsn't numerical value.
	 */
	public int readBuff(int code, int pos){ return(readBuff(code, pos, 0));}
	//]
	
 	//lecture des méta-donnée TAG (méta-données structurées)
	/**
	 * Get and read the <i>tag</i> with the given name and TagReader.
	 * @param name The name of the <i>tag</i> that you want to read.
	 * @param callback The TagReader that you will use.
	 */
	public abstract void getTag(String name, TagReader callback);
 	
 	//[envoie de message vers ce spot
	/**
	 * Send a message, with <i>ack</i> to the <i>Spot</i>, in the given <i>channel</i>.
	 * @param channel The <i>channel</i> used to send the message.
	 * @param content The content you want to send in your message.
	 * @return <b>boolean</b> Return <b>true</b> if the message was correctly delivered.
	 */
	public boolean send(int channel, DataReader content) { return(send(channel, content, true, true)); }	//-> emission du message, avec acquitement et content sera disposé aprés émission
	/**
	 * Send a message, with <i>ack</i> to the <i>Spot</i>, in the given <i>channel</i>.
	 * @param channel The <i>channel</i> used to send the message.
	 * @param content The content you want to send in your message.
	 * @param need_ack The boolean to adjust if you need (or not) ack for your message.
	 * @param auto_dispose <i>dispose()</i> function will be called, directly after sending the message.
	 * @return <b>boolean</b> Return <b>true</b> if the message was correctly delivered.
	 */
	public abstract boolean send(int channel, DataReader content, boolean need_ack, boolean auto_dispose); 	//-> emission du message, dispose() sera appliquée à l'instance content que si auto_dispose==true

	/**
	 * Send a message, with <i>ack</i> to the <i>Spot</i>, in the given <i>channel</i>.
	 * @param channel The <i>channel</i> used to send the message.
	 * @param content The content you want to send in your message.
	 * @return <b>boolean</b> Return <b>true</b> if the message was correctly delivered.
	 */
	public boolean send(int channel, byte content){ return(send(channel, content, true)); };	//-> emission du message ne contenant qu'un byte
	/**
	 * Send a message, with <i>ack</i> to the <i>Spot</i>, in the given <i>channel</i>.
	 * @param channel The <i>channel</i> used to send the message.
	 * @param content The content you want to send in your message.
	 * @param need_ack The boolean to adjust if you need (or not) ack for your message.
	 * @return <b>boolean</b> Return <b>true</b> if the message was correctly delivered.
	 */
	public abstract boolean send(int channel, byte content, boolean need_ack);	//-> emission du message ne contenant qu'un byte
	
	/**
	 * Send a message, with <i>ack</i> to the <i>Spot</i>, in the given <i>channel</i>.
	 * @param channel The <i>channel</i> used to send the message.
	 * @param content The content you want to send in your message.
	 * @return <b>boolean</b> Return <b>true</b> if the message was correctly delivered.
	 */
	public boolean send(int channel, byte[] buff) {	return(send(channel, buff, 0, buff.length, true)); } //-> emission du buffer buff avec acquittement
	/**
	 * Send a message, with <i>ack</i> to the <i>Spot</i>, in the given <i>channel</i>.
	 * @param channel The <i>channel</i> used to send the message.
	 * @param buff The content you want to send in your message.
	 * @param offset The offset for the buffer you want to send.
	 * @param size The size of the buffer that you want to send
	 * @param need_ack The boolean to adjust if you need (or not) ack for your message.
	 * @return <b>boolean</b> Return <b>true</b> if the message was correctly delivered.
	 */
	public abstract boolean send(int channel, byte[] buff , int offset, int size, boolean need_ack);	//-> emission du buffer buff
	//]
	
	//[Pour les sous-classes
	protected void fireDown(){
		synchronized(observers) { for(SpotObserver so : observers) 
			ThreadStack.perform(() -> so.down(this)); 
		} 
	}
	protected void fireIdContext(String ch_context, String id_context){
		synchronized(observers){
			for(SpotObserver so : observers) 
				ThreadStack.perform(() -> so.idContext(this, ch_context, id_context));
		}
	}
	protected void fireUpdateSMD(boolean first_definition){ //Pour chaque méthode mmm de l'interface XXXListener
		synchronized(observers){
			for(SpotObserver so : observers) 
				ThreadStack.perform(() -> so.updateSMD(this, first_definition));
		}
	}
	//NB: le fireUp est réalisé par le HubImpl
	//]
}