/**
Formalisation pour l'implémentation interne de HubUdp:
======================================================
	- Un noeud N du réseau est un point de connexion ip:port
	- il y a un unique noeud par contexte d'exécution (soit 1 par moteur HubUdp actif)
	- ip+port permet de cibler un contexte d'exécution (un noeud). A ce niveau il y a potentiellement plusieurs spots candidats 
	  pour l'arrivé d'un message. Notamment : un spot local par hub instancié sur ce noeud, plusieurs spots proxy si l'un de
	  ces spots locaux est un spot relais ou un spot de fusion. Aussi pour déterminer précisément le spot concernée, chaque
	  spot est identifié par un code appelé ctarget. C'est un id unique au niveau de l'host.
	  ==> Sur un host donnée, chaque spot (donc tous les spots de même ip) est associé à un idenfifiant unique appelé 
	  ctarget ("code target").
	  ==> la chaine ip:ctarget suffit à cibler un spot sur l'ensemble d'un réseau : ctarget suffit à indentifier
	  le hub du spot et le port de connexion à utiliser pour le joindre (et en cas de fusion, les spot proxy ciblés
	  renverrons le message vers le spot final).
	  ==> la chaine ip:ctarget suffit à désigner un spot sans ambiguité sur l'ensemble d'un réseau sans fusion. 
	- Sur chaque host, il doit y avoir au moins un noeud utilisant le port 8266, ce noeud est le noeud maitre : NM
	  les autres noeuds de l'host sont appelé noeud slave : NS.
	- Les NM ont la charge de gérer la diffusion des messages broadcasts systèmes (ceux nécessaires au fonctionnement
	interne de HubUdp).
	- Chaque noeud gére une map, appelé hubMap, qui décrit l'ensemble des spots du réseau.
	- Chaque NM est responsable de la partie de la hubMap qui concerne son host (son ip), appelé localMap. Cette 
	  partie décrit les spots du NM et ceux de tous les NS situé sur le même host que lui.
	- Pour un NM donnée, on appel globalMap la hubMap privé de la localMap du NM.
	- Les NM construisent leur globalMap en s'échangeant les informations de leur localMap respective.
	- Les NS maintiennent une copie de la globalMap et de la localMap de leur NM local (situé sur le même host qu'eux).
	- Chaque NS doit surveiller l'existance de son NM local afin de prendre ce rôle si le NM devient inactif.
	- Les NM ont la charge d'attribuer le ctarget à chaque spot de l'host (même ceux des NS)
	  --> comme le NM est unique et obligatoirement existant sur chaque host, cela garanti l'unicité de ces identifiants au 
	  niveau de l'host et donc de la cohérence de la localMap de ce host.
	  
	 ==> pour cibler un spot précis, il faut utiliser : ip + port + ctarget , ainsi en envoyant un message à ip:port il suffit
	  que le message contiennent l'indication ctarget pour que le spot cible puisse être identifié
	
	
	Chaine identifiant des hubs : 
	hub public: id = 'nom'
	hub private: id = 'nom:hash_key'
	
	
	=== gloablMap et localMap :
	
	- globalMap = hash:{ ip => ipmap } , sauf l'host local
	
	- ipmap = { ip: , rand_instance: ipmap_version:,  hash:{  id_hub =>   [...{ctarget: port: rinstance: }...]   } }			NB: id_hub =  "nom:hash_pass"

	ipmap compilée :
		(dword: ip) (short: rand_instance)(short: ipmap_version)(short: nb_hub) $ {...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance) $ |...} |...}
			==> $ désigne une position de coupure potentielle entre les paquets
			
	localMap compilée :
	 	(short: rand_instance)(short: my_ipmap_version) {...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance) $ |...} |...}
			==> $ désigne une position de coupure potentielle entre les paquets
[
Protocole :
===========

 	**** code_conf == 85 : message de niveau noyau
	[85][cmd]{... content ...}

	**** code_conf == 58 : message de niveau applicatif, channel=0 est réservé pour le protocole interne HubUdp
	[58](short: ctarget_sender)(short: ctarget_destination)(short: ack)(short: lunack)[channel]{ ...content... } ==> 10 octets d'entête
		//nb: ctarget_destination=0 --> le spot local du hub de ctarget_sender sur la machine qui recoit le message
	
	
		- channel=0 (protocole interne avec possibilité d'acquitemment et de découpage des messages > 500 octets)
		[ACK](short: ack)		<- aquittement de réception d'un paquet reçu par l'émetteur de cet acquittement
		[PART](int : num)(int : total)[channel]{...content_part...}	<- paquet à recomposer (acquitement obligatoire)
		[BROADCAST][X][Y]{...content_part...} , si X == PROTO_SPOT ->pas de Y, sinon si X = PROTO_APP -> Y = code protocole  <- quand ce message est recu par l'un des spots d'un contexte, son contenu est envoyé à tous les spots du même hub sur ce contexte 

		- Protocole des Spot Info :
		[GET]	--> demande de définition des infos SMD et des idContext du recepteur
		(*bh*) [DEF][def_version][nb_value]   ...(pour chaque valeur)[code][value_size]{...value...}...	<- réponse à DEF ou notification de changement (-> *bh*)
		[GET_TAG](string: name)						--> demande de valeur d'un tag
		[TAG_VALUE](string: name)(int: count)...|pour chaque valeur: (value: v)|...		<-- réponse à GET_TAG 
	
		//IdContext
		*bh* [ID_CONTEXT](string: chaine_context)(string: id_context)	<- le spot émetteur signal qu'il possède ou change un idContext
		[ID_CONTEXT_LIST](int: nb_id_context)|...pour chaque id: (string: chaine_context)(string: id_context) ...| <- émis lors de la réception d'une commande GET

Cibler un spot : ip:port:ctarget
	--> le ctarget pointe sur un hub particulier et un port de routage pour pointer sur un spot

/!\ :
- pour désigner un spot il faut privilégier son id global, car son ip:port:ctarget peut être faussé s'il se trouve dérrière un portail de fusion !
--> ou mieux un idContext si c'est possible, pour supporter les down/re-up 
]
*/
package skuad.hubudp.implementation.local;

import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.Map;
import java.util.TimerTask;
import java.util.Vector;
import skuad.hubudp.*;
import skuad.util.databytes.*;
import skuad.hubudp.implementation.*;
import skuad.hubudp.implementation.portal.Linker;
import skuad.util.base.*;
import skuad.util.thread.ThreadStack;

public class UdpKernel implements Runnable, TargetFinder {
		
	HubManager hubManager;
	int randInstance;			  //valeur d'identification propre du noyau courant (sur 2 octets --> de 1 à 65000)
	DatagramSocket master_socket; //!=null si le contexte d'exécution local est un noeud maitre (NM)
	DatagramSocket socket;		  //socket réseau applicative principale, ==null si non initialisé, peut être égale à master_socket si le contexte a obtenu le port maitre dés le départ
	int listen_port;			  //port d'écoute applicatif
	String ipPort;
	
	Map<Integer, ManagedSpot> managedSpots;	//table des spots géré par le contexte d'exécution courant (spots locaux + proxy), key : ctarget  -> instance ManagedSpot  (<= utilisé pour identifier la cible des messages entrants)
	Map<String, SpotImpl> otherSpots;		//table des spots non géré par le contexte courant, qui sont connectés à un hub de localHubs (donc relié au contexte d'exécution courant), key : "ip:ctarget" -> instance SpotImpl (<= utilisé pour identifier les cibles distantes et les expéditeur distant)
	Map<String, SpotMap> globalMap;	int globalmap_version;		//hash:{ ip => ipmap } , ne contient pas l'host local
	SpotMap localMap; 											//map de l'host local
	Vector<SlaveDescriptor> slaveList; 	int slavelist_version;	//liste des noeuds slave (NS) de l'hote local	
	
	public UdpKernel(HubManager hmanager) {
		hubManager = hmanager;
		randInstance = Util.randInt(1, 65000);

		managedSpots = new HashMap<Integer, ManagedSpot>(); //NB: sera toujours modifié sur verrou de localHubs
		otherSpots = new HashMap<String, SpotImpl>();
		globalMap = new HashMap<String, SpotMap>();						globalmap_version = 0;
		localMap = new SpotMap(Network.LOCALHOST, Network.localAddress);	
		slaveList = new Vector<SlaveDescriptor>();						slavelist_version = 0;
		
		//[On va essayer d'obtenir le contrôple du port maitre
		try {
			/*if(Util.isMac()) { master_socket = new DatagramSocket(null); master_socket.bind(new InetSocketAddress("127.0.0.1", ProtocoleLocal.MASTER_PORT)); }
			else*/ master_socket=new DatagramSocket(ProtocoleLocal.MASTER_PORT); //pour Windows et linux 
			//System.out.println("*SpotInfoKernel* port maitre capturé ");
			master_socket.setBroadcast(true);
			socket = master_socket;
		} catch(Exception e){	//Le port maitre n'est pas disponible, on va prendre un port au hasard
			master_socket=null;
			//System.out.println("*SpotInfoKernel* échec d'obtention du port maitre");	
			try{
				socket=new DatagramSocket();
				//System.out.println("*SpotInfoKernel* port slave obtenu = "+socket.getLocalPort());			
				socket.setBroadcast(true); 	
			} catch(Exception e2){ socket=null; return; }	//Erreur critique, normalement impossible
		}
		listen_port = socket.getLocalPort();
		//]
		
		//On lance le processus d'écoute de la socket
		(new SocketListener(this, socket)).start();

		if(master_socket!=null) { //si on est le noeud maitre, on met à jours la slavelist
			slaveList.add(new SlaveDescriptor(randInstance, listen_port)); 
			slavelist_version++; //va prendre la valeur 1
			localMap.setRInstance(randInstance, 0); //on prend le contrôle de la localMap
			//on envoie un MASTER_PULSE en BM
			broadcastMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.MASTER_PULSE).sh(randInstance).sh(localMap.version).sh(globalmap_version).sh(slavelist_version), true);
			eco_master_pulse=false;
			//NB: on ne peut pas emettre un MASTER_PULSE en BS car on ne connait pas encore de noeud slave
		} else sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.SLAVE_PULSE).sh(randInstance));//sinon on s'enregistre auprés du noeud master via un message SLAVE_PULSE
		
		//On lance la boucle d'activité du noyau (1 activation toutes les 3 secondes)
		ThreadStack.perform( this , 50, 3000);
		
		//On engregistre le processus de détection de fin d'exécution
		Runtime.getRuntime().addShutdownHook(new Thread(){
		    public void run(){
		    	//System.out.println("... fin d'exécution de noyau ..."); 
		    	boolean down_need_send = true;
		    	if(master_socket==null) sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.DOWN).sh(randInstance)); //on est slave
		    	else if(slaveList.size()>1) { //on est master on envoie le DOWN à l'un des slaves restant
					for(int i=0; i<slaveList.size(); i++) {
						SlaveDescriptor sd = slaveList.get(i);
						if(sd.port!=listen_port) {
							if(down_need_send) {
								send(master_socket, Network.localAddress, sd.port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.DOWN).sh(randInstance), true);
								down_need_send=false;
							}
						} else {
							localMap.removeTargetOnPort(UdpKernel.this, listen_port,   randInstance, 0, true);//down de tous les spots portés localement
							slaveList.removeElementAt(i); slavelist_version=HubManager.nextVersion(slavelist_version); i--; //maj de la slavelist
							//notification de la maj de la slavelist aux NS: [85][REMOVE_SLAVE](short: master_rand_instance)(short: new_slavelist_version)(short: port)(short: rand_instance)
							broadcastSlave(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.REMOVE_SLAVE).sh(randInstance).sh(slavelist_version).sh(listen_port).sh(sd.randInstance), true,0);
						}
					}
		    	}
		    }
		});		
	}
	public String getIpPort() {		//ip:port du contexte courant
		if(ipPort==null) ipPort = Network.LOCALHOST +':'+listen_port;
		return(ipPort); 
	}
	public int getPort() { return(listen_port); }
	
	//[Pour HubLocal
	protected void registerManagedLocalSpot(int ctarget, ManagedLocalSpot mlocalspot) {
		managedSpots.put(ctarget, mlocalspot);
	}
	//]
	
	public Hub joinLocalHub(String id_hub, String name, String key) {
		if(socket==null) return(null);	//une erreur critique empèche la mise en route du middleware 
		if(master_socket!=null) {
			HubLocal hub = new HubLocal(hubManager, this, id_hub, name, key); //création du hub
			int ctarget = newCTarget();
			if(ctarget<0) return(null); //il n'est plus possible de rejoindre de hub (65000 spots existe déjà sur l'hote local)
			hub.localSpotLinked(ctarget, listen_port);
			hub.setReady();
			
			synchronized(localMap) {
				localMap.addTarget(this, id_hub, ctarget, listen_port, randInstance, 0, true); //la notification de la modification de la localMap découlera de l'appel targetInMap(...)						
			}
			
			//On doit analyser l'existence des targets de localMap et de globalMap qui sont reliée à ce hub pour instancier leur spot et les mettres dans otherSpots
			searchAllTargetForHub(this, id_hub, false);
			return(hub);
		}
		else { //on doit demander au port maitre un ctarget pour pouvoir instancier ce spot local, et donc ce hub
			synchronized(waitingJoin) {
				JoinParam jparam = getWaitingJoin(id_hub, 0); 	
				if(jparam!=null) return(jparam.hub); //on est déjà connecté en attente de connexion pour ce hub
				if(localMap.targetCount()<65000) { //il est encore possible d'ajouter un spot sur l'host local
					HubLocal hub = new HubLocal(hubManager, this, id_hub, name, key); //création du hub
					waitingJoin.add(new JoinParam(hub)); //on met en attente cette demande
				
					//on demande au NM le ctarget pour le local spot de connexion à ce hub
					sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.QUERY_PLUG).s(id_hub)); // [85][QUERY_PLUG](s: id_hub)
					//	--> l'instanciation du hub dans localHubs (et la notification aux hubListener) se fera à la réception du message [ACCEPT_PLUG]
					return(hub);
				} else return(null);
			}
		}
	}
	
	public boolean addSpotProxy(HubLocal hub, Linker linker, int vjid) { // --> l'appel linker.notifyProxyAdded(...) sera déclenché quand l'ajout sera effectif
		if(socket==null) return(false);	//une erreur critique empèche la mise en route du middleware 
		if(master_socket!=null) { //on peut enregistrer directement ce proxy
			int ctarget = newCTarget();
			if(ctarget<0) return(false); //opération impossible 65000 spots existe déjà sur l'hote local
			ManagedSpotProxy msp = new ManagedSpotProxy(hub, ctarget, listen_port, linker, vjid);
			managedSpots.put(ctarget, msp); 

			synchronized(localMap) {
				localMap.addTarget(this, hub.id_hub, ctarget, listen_port, randInstance, 0, true); //la notification de la modification de la localMap découlera de l'appel targetInMap(...)						
			}
//System.out.println("addSpotProxy ok");			
			return(true);
		} else { //on doit demander au port maitre un ctarget pour pouvoir instancier ce spot proxy dans ce hub
			synchronized(waitingJoin) {
				if(localMap.targetCount()<65000) { //il est encore possible d'ajouter un spot sur l'hote local
					waitingJoin.add(new JoinParam(hub, linker, vjid)); //on met en attente cette demande
//System.out.println("addSpotProxy waiting");			
					//on demande au NM le ctarget pour ce spot proxy
					sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.QUERY_PLUG).s(hub.id_hub)); // [85][QUERY_PLUG](s: id_hub)
					//	--> l'instanciation se fera à la réception du message [ACCEPT_PLUG]
					return(true);
				} else return(false);
			}
		}
	}
	public void removeSpotProxy(HubLocal hub, Linker linker, int vjid) { // --> l'appel linker.notifyProxyAdded(...) sera déclenché quand l'ajout sera effectif
		synchronized(managedSpots) {//on doit retrouver l'instance ManagedSpot et la supprimer de managedSpots
			for(ManagedSpot mspot : managedSpots.values()){
				if(!mspot.isLocalSpot()){
					ManagedSpotProxy mspotp = (ManagedSpotProxy) mspot;
					if(mspotp.isLinkedTo(linker) && (vjid==0 || mspotp.getJid()==vjid)) {
						mspot.getLocalHub().exit(mspotp, randInstance);
						managedSpots.remove(mspotp.getIdLocal());
						if(master_socket!=null) localMap.removeTarget(this, hub.id_hub, mspotp.getIdLocal(), 0, true); //on est master
						else { //on est slave
							 //on doit envoyer au NM : [REMOVE_LOCAL](short: rand_instance)(short: my_ipmap_new_version)(s: id_hub) (short: ctarget)(short: port)(short: rinstance)
							this.sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.REMOVE_LOCAL).sh(randInstance).sh(0).s(hub.id_hub).sh(mspotp.getIdLocal()).sh(listen_port).sh(randInstance));
						}
						if(vjid!=0) return;
					}
				}
			}
		}
	}
	public ManagedSpotProxy getSpotProxy(Linker linker, int vjid) {
		synchronized(managedSpots) {//on doit retrouver l'instance ManagedSpot et la supprimer de managedSpots
			for(ManagedSpot mspot : managedSpots.values()){
				if(!mspot.isLocalSpot()){
					ManagedSpotProxy mspotp = (ManagedSpotProxy) mspot;
					if(mspotp.isLinkedTo(linker) && (vjid==0 || mspotp.getJid()==vjid)) return(mspotp);
				}
			}
			return(null);
		}
	}
	//[Méthodes internes
	protected void performPacket(DatagramSocket socket, DatagramPacket data, byte[] buff){ //réception d'un message UDP
		InetAddress sender_adr = data.getAddress();
		//System.out.println("---paquet recu de "+sender_adr.getHostName()+":"+data.getPort()+"  commande = "+ buff[7]+" total data size = "+data.getLength());//+ "  ("+diff_time+")");
		boolean local_sender = Network.isLocalAddress(sender_adr);
		int sender_port = data.getPort();
		if(local_sender && (sender_port==listen_port || (sender_port==ProtocoleLocal.MASTER_PORT && master_socket!=null))) return; //on ignore les auto-réception
		
		String sender_ip = sender_adr.getHostAddress();
		int start = data.getOffset();
		int length =  data.getLength();
		
		int code_conf = buff[start]; //code de conformité du paquet
		DataReader mess;		
		if(code_conf==ProtocoleGlobal.PROTO_KERNEL) { 		//c'est un paquet kernel
			//[85][cmd]{... content ...}			
/*{ //##Log##
int vcmd = buff[start+1];
switch(vcmd) {
case ProtocoleLocal.MASTER_PULSE: case ProtocoleLocal.SLAVE_PULSE: break; //RAS
default : System.out.println("--message PROTO_KERNEL recu de : "+sender_ip+":"+sender_port+" , code commande = "+vcmd);
}}//#######
*/
			mess = DataBytes.reader(buff, start+1, length-1);
			if(master_socket!=null) performMasterMessage(local_sender, sender_adr, sender_ip, sender_port, mess);
			else performSlaveMessage(local_sender, sender_adr, sender_ip, sender_port, mess);
		} else if(code_conf==ProtocoleGlobal.PROTO_SPOT || code_conf==ProtocoleGlobal.PROTO_APP ) {	//c'est un paquet spot, à livrer à l'un des membre de managedSpots
			//[X][Y](short: ctarget_sender)(short: ctarget_destination)(short: ack)(short: lunack){content} , Y = protocole app. si X==PROTO_APP
			//ctarget_sender		-> ctarget du spot local au contexte d'exécution émetteur
			//ctarget_destination	-> ctarget du spot vers lequel le message est envoyé	
			mess = DataBytes.reader(buff, start+1, length-1);
			int protocole;
			if(code_conf==ProtocoleGlobal.PROTO_APP) {
				protocole=mess.readByte();
/*{ //##Log##
System.out.println("--message PROTO_APP recu de : "+sender_ip+":"+sender_port+" , code protocole = "+protocole);
}//#######				
*/			} else {
				protocole = -1;
/*{ //##Log##
int vcmd = buff[start+9];
switch(vcmd) {
case ProtocoleSpot.ACK: break; //RAS
default : System.out.println("--message PROTO_SPOT recu de : "+sender_ip+":"+sender_port+" , code commande = "+vcmd);
}}//#######
*/			}
			int ct_sender = mess.readShort();
			SpotImpl sender = otherSpots.get(sender_ip+'#'+ct_sender);
			if(sender!=null) { //on ne traite le message que si l'expéditeur est connu, dans le cas contraire le message sera ré-expédié si il est à acquittement
				if(sender instanceof SpotUdp) {
					SpotUdp u_sender = (SpotUdp) sender;
					int ct_receiver = mess.readShort();
					//System.out.println("\t ==> ct_receiver = "+ct_receiver);				
					ManagedSpot receiver = managedSpots.get(ct_receiver);
					int ack = mess.readShort();
					int lunack = mess.readShort();

					if(ack!=0) { //on acquite la réception de ce message
//if(code_conf==ProtocoleGlobal.PROTO_SPOT && mess.getByte(mess.curPos())==ProtocoleSpot.GET) System.out.println("]]]]]]]]] GET receive with ack = "+ack); int x;						
						DataWriter dw = DataBytes.writer().b(ProtocoleGlobal.PROTO_SPOT).sh(ct_receiver).sh(ct_sender).sh(0).sh(0).b(ProtocoleSpot.ACK).sh(ack).sh(u_sender.rinstance);
						if(receiver==null) dw.b(ProtocoleSpot.TARGET_NOT_EXIST);
						else dw.b(ProtocoleSpot.OK);			
						send(socket, sender_adr, sender_port, dw, true);
					}	
					u_sender.messageFrom(receiver, ack, lunack, protocole, mess); 
				} else hubManager.fireKernelLog(Hub.LOG_WARNING, "incorehent instance sender in UdpKernel.performPacket(...)");
			}
//else System.out.println("===============> expéditeur non connu !!!"); 			
		}//else erreur de code de conformité, le paquet est ignoré		
	}
	
	public void run() { //boucle d'activité du noyau (toutes les 3 secondes)
		if(master_socket!=null) { 	//comportement Noeud Master
			//[Emission des pules master
			DataWriter pulse = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.MASTER_PULSE).sh(randInstance).sh(localMap.version).sh(globalmap_version).sh(slavelist_version);
			
			//1) on pulse en broadcast master
			if(!eco_master_pulse) broadcastMaster(pulse, false);
			eco_master_pulse=false;
			
			//2) vérification et pulse des slaves
			synchronized(slaveList) {
				for(int i=0; i<slaveList.size(); i++) {
					SlaveDescriptor sd = slaveList.get(i);
					if(sd.port==listen_port) continue; //on ne traite pas le contexte courant
					sd.wd++;
//System.out.println("@ wd pour le slave sd.port => "+sd.wd);					
					if(sd.wd>2) { //echec wd du slave, on va le down --> retrait de tous ses spots
//System.out.println("@ --> echec du wd ce slave");									
						localMap.removeTargetOnPort(this, sd.port, sd.randInstance, 0, true);//down de tous les spots de ce slave
						slaveList.removeElementAt(i); slavelist_version=HubManager.nextVersion(slavelist_version); i--; //maj de la slavelist
						//notification de la maj de la slavelist aux NS: [85][REMOVE_SLAVE](short: master_rand_instance)(short: new_slavelist_version)(short: port)(short: rand_instance)
						broadcastSlave(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.REMOVE_SLAVE).sh(randInstance).sh(slavelist_version).sh(sd.port).sh(sd.randInstance), true,0);
					} else if(!eco_slave_pulse) this.send(master_socket, Network.localAddress, sd.port, pulse, false); //on pulse ce slave
				}
			}
			eco_slave_pulse=false;
			pulse.dispose();
			//]
			
			//3) vérification des wd des master
			synchronized(globalMap) {	//si échec --> down des spots du master
				Vector<String> to_remove=null;
				for(SpotMap map : globalMap.values()) {
					map.wd++;
					if(map.wd>2) { //echec wd de ce NM, on va le down --> retrait de tous ses spots
						 if(to_remove==null) to_remove=new Vector<String>(); 
						 to_remove.add(map.ip);
					}
				}
				if(to_remove!=null) {
					for(String ip_map : to_remove) {
						SpotMap map = globalMap.remove(ip_map);
						if(map!=null) { //ce sera normalement toujours le cas
							map.removeTargetOnPort(this, 0, 0, map.version, false);//down de tous les spots de cet hote, sans notification
							globalmap_version=HubManager.nextVersion(globalmap_version);							
							//notification REMOVE_MASTER aux NS : //[85][REMOVE_MASTER](short: master_rand_instance)(short: globalmap_version)(dword: ip)(short: ip_rand_instance) <- suppression d'un master ==> tous ses spots devront être retiré de la globalMap et être passé à down
							DataWriter dw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.REMOVE_MASTER).sh(randInstance).sh(globalmap_version);
							map.writeIp(dw);
							dw.sh(map.rinstance);
							broadcastSlave(dw, true,0);
						}
					}					
					to_remove=null;
				}				
			}
			
			//4) purger les buffers received_ipM inactif depuis 2 cycles
			synchronized(received_ipM) {
				Vector<String> to_remove=null;
				for(String sender_ip : received_ipM.keySet()) {
					PageRecomposer precmp = received_ipM.get(sender_ip);
					precmp.wd++;
					if(precmp.wd>2) { if(to_remove==null) to_remove=new Vector<String>(); to_remove.add(sender_ip); }
				}
				if(to_remove!=null) {
					for(String sender_ip : to_remove) {
						received_ipM.remove(sender_ip).purge(false);
					}					
					to_remove=null;
				}
			}
			
			//5) purger les buffers received_spotL inactif depuis 2 cycles
			synchronized(received_spotL) {
				Vector<Integer> to_remove=null;
				for(int port : received_spotL.keySet()) {
					PageRecomposer precmp = received_spotL.get(port);
					precmp.wd++;
					if(precmp.wd>2) { if(to_remove==null) to_remove=new Vector<Integer>(); to_remove.add(port); }
				}
				if(to_remove!=null) {
					for(int port : to_remove) {
						received_spotL.remove(port).purge(false);
					}					
					to_remove=null;
				}
			}
		} else {			//comportement Noeud Slave
			//1) on pulse le master
			sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.SLAVE_PULSE).sh(randInstance)); 	
			
			//2) vérification du wd du master
			localMap.wd++; 
			if(localMap.wd>1) becomeMaster(); //le master est en situation d'échec de pulse, on va tenter un become master

			//3) vérification des waitingJoin en attente 
			synchronized(waitingJoin) { 
				for(JoinParam hubp : waitingJoin) {
					hubp.send_count++;
					if(hubp.send_count>1) sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.QUERY_PLUG).s(hubp.hub.id_hub)); // [85][QUERY_PLUG](s: id_hub)
					if(hubp.send_count>3) hubManager.fireKernelLog(Hub.LOG_WARNING, "Local master not answer to QUERY_PLUG (count send = "+hubp.send_count+" )");
				}
			}
			
			//4) vérification des waitingReadyPlug en attente
			synchronized(waitingReadyPlug) { 
				for(JoinParam hubp : waitingReadyPlug) {
					hubp.send_count++;
					if(hubp.send_count>1) sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.READY_PLUG).sh(randInstance).s(hubp.hub.id_hub).sh(hubp.ctarget)); // [85][READY_PLUG](short: rinstance)(s: id_hub)(short: ctarget)
					if(hubp.send_count>3) hubManager.fireKernelLog(Hub.LOG_WARNING, "Local master not answer to READY_PLUG (count send = "+hubp.send_count+" )");
				}				
			}
		}
	}
	//]

	//[----------------- Codes spécifiques au rôle Master -----------------------------------------------------------
	private int next_ctarget=1; //prochain id ctarget à utiliser
	private boolean eco_master_pulse=false, eco_slave_pulse=false;
	private Vector<DataWriter> builded_globalM=new Vector<DataWriter>(); int builded_globalM_version=-1;	//cache du buffer d'envoie de la globalMap
	private Vector<DataWriter> builded_localM=new Vector<DataWriter>();	int builded_localM_version=-1;		//cache du buffer d'envoie de la localMap
	private Vector<DataWriter> builded_slaveL=new Vector<DataWriter>();	int builded_slaveL_version=-1;		//cache du buffer d'envoie de la slavelist
	private Map<String, PageRecomposer> received_ipM=new HashMap<String, PageRecomposer>(); 	//buffers des localMap recue des autres NM 
	private Map<Integer, PageRecomposer> received_spotL=new HashMap<Integer, PageRecomposer>(); //buffers des spotlist recue des autres NS 
	private SlaveDescriptor getSlave(int p) { //retourne le slave descritor ayant pour port p
		synchronized(slaveList) {
			for(SlaveDescriptor sd : slaveList) if(sd.port==p) return(sd);
		}
		return(null);
	}
	protected void broadcastMaster(DataWriter wcmd, boolean dispose){ //envoi d'un message broadcast à destination de tous les noeuds maitre du réseau
		eco_master_pulse=true;
		for(InetAddress br_addr : Network.allBroadcastAddress()) {
			try{	
				master_socket.send(wcmd.toDatagramPacket(br_addr, ProtocoleLocal.MASTER_PORT)); 
			} catch(Exception e){
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Enable to send broadcastMaster message on interface ["+br_addr.getHostAddress()+"] ("+e.getMessage()+")");
			}	
		}
		if(dispose) wcmd.dispose();
	}
	protected void broadcastSlave(DataWriter wcmd, boolean dispose, int omit){ //envoi d'un message à destination de tous les noeuds slaves locaux
		eco_slave_pulse=true;
		synchronized(slaveList) {			
			for(SlaveDescriptor sd : slaveList) {
				if(sd.port!=listen_port && (omit==0 || sd.port!=omit)) try{ 
//System.out.println("__/ envoie d'un message broadcastSlave vers le slave "+sd.port);					
					master_socket.send(wcmd.toDatagramPacket(Network.localAddress, sd.port));
				} catch(Exception e){
					hubManager.fireKernelLog(Hub.LOG_WARNING, "Enable to send broadcastSlave message to port ["+sd.port+"] ("+e.getMessage()+")");
				}	
			}
		}
		if(dispose) wcmd.dispose();
	}
	protected void performMasterMessage(boolean local_sender, InetAddress sender_adr, String sender_ip, int sender_port, DataReader mess) {
		String id_hub;
		int ctarget, rand_i, map_version;
		boolean coherent = true; //mis à false en cas d'incohérence de version sur une ip_map
		int cmd = mess.readByte();
		switch(cmd) {
		case ProtocoleLocal.SLAVE_PULSE : //[85][SLAVE_PULSE](short: rand_instance)
			if(!local_sender) break; //cette demande ne peut provenir que d'un NS local	
			rand_i = mess.readShort();
			checkSlave(sender_port, rand_i);
			break;
		case ProtocoleLocal.MASTER_PULSE : //[85][MASTER_PULSE](short: rand_instance)(short: localmap_version)(short: globalmap_version)(short: slave_list_version)
			if(local_sender) break; //cette demande ne peut pas provenir d'un NS local
			rand_i = mess.readShort();
			map_version =  mess.readShort();
			synchronized(globalMap) {
				SpotMap map = globalMap.get(sender_ip);
				if(map!=null) {
					if(map.rinstance!=rand_i) { //c'est un changement de master sur le noeud NM, on va demander une synchro map						
//System.out.println("change NM detected old ri="+map.rinstance+" new ri="+rand_i);	int x;	
						map.version=0;
						map.rinstance = rand_i;
						globalmap_version = HubManager.nextVersion(globalmap_version);
						coherent=false;
						//On notifie les slaves : *bs* [85][CHANGE_MASTER](short: master_rand_instance)(short: globalmap_version)(dword: ip)(short: new_ip_rand_instance)
						DataWriter wcmd = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.CHANGE_MASTER).sh(randInstance).sh(globalmap_version);
						map.writeIp(wcmd);
						wcmd.sh(rand_i); 
						broadcastSlave(wcmd, true, 0);
					} else if(map_version!=map.version) coherent=false; //il y a une désynchro sur cette map, on va demander au NM une maj de sa LocalMap
					map.wd=0; //reset du wd de ce NM
				} else coherent=false; //si c'est un NM non connu on va lui demander sa LocalMap
			}
			if(!coherent) { //on va demande la localmap complête du noeud emetteur: [85][GET_LOCALMAP]		
//System.out.println("'''' need map for "+sender_adr+":"+sender_port); int x;			
				send(master_socket, sender_adr, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_LOCALMAP), true); 
			}
			break;
		case ProtocoleLocal.QUERY_PLUG :	//[QUERY_PLUG](s: id_hub) <- un NS demande au NM un ctarget pour le spot local qui lui permettra de se connecter au hub indiqué
			id_hub = mess.readString();
			if(!local_sender) { //cette demande ne peut provenir que d'un NS local
				//on refuse la demande
				send(master_socket, sender_adr, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.DENY_PLUG).s(id_hub).s("Enable to perform QUERY_PLUG message from a non local node").sh(0).sh(0), true);
				break;
			}
			ctarget = newCTarget(); //génération d'un nouveau ctarget
			if(ctarget>0) send(master_socket, sender_adr, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.ACCEPT_PLUG).s(id_hub).sh(ctarget), true);
			else send(master_socket, sender_adr, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.DENY_PLUG).s(id_hub).s("Enable to produce more ctarget value").sh(0).sh(0), true);
			break;
		case ProtocoleLocal.READY_PLUG:  //[READY_PLUG](short: rinstance)(s: id_hub)(short: ctarget) <- le NS signal que son spot local est pret, c'est le signal pour le NM de prendre véritablement en compte cette existance
			rand_i = mess.readShort();			
			id_hub = mess.readString();
			ctarget = mess.readShort();
			if(!local_sender) { //cette demande ne peut provenir que d'un NS local
				//on refuse la demande
				send(master_socket, sender_adr, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.DENY_PLUG).s(id_hub).s("Enable to perform READY_PLUG message from a non local node").sh(ctarget).sh(0), true);
				break;
			}
			checkSlave(sender_port, rand_i); //on s'assure que le slave est bien dans la slave_list
			
			if(localMap.targetExist(ctarget, sender_port, rand_i)) break; //cette demande READY est déjà traité
			else if(localMap.targetExist(ctarget, 0, 0)) { //erreur de conflit sur ctarget, on doit répondre REPLUG ou DENY_PLUG à l'émetteur
				int new_ctarget = newCTarget(); //génération d'un nouveau ctarget
				if(new_ctarget>0) send(master_socket, sender_adr, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.DENY_PLUG).s("Code target allready in use").s(id_hub).sh(ctarget).sh(new_ctarget), true);
				else send(master_socket, sender_adr, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.DENY_PLUG).s(id_hub).s("Enable to produce more ctarget value").sh(ctarget).sh(0), true);				
				break;
			}//sinon tout est ok, on ajoute le spot correspondant dans les listes
			
			//maj de localMap
			synchronized(localMap) {
				localMap.addTarget(this, id_hub, ctarget, sender_port, rand_i,0, true); //la notification de la modification de la localMap découlera de l'appel targetInMap(...)		
			}
			break;
		case ProtocoleLocal.ADD_LOCAL : //*B* [85][ADD_LOCAL](short: rand_instance)(short: my_ipmap_new_version)(s: id_hub)(short: ctarget)(short: port)(short: rinstance)
			//Un host distant signal l'ajout d'un spot sur sa localmap
			if(local_sender) break; //on ignore les envoies de provenance locale (car on est le NM et c'est nous qui envoyons ADD_LOCAL pour les ajouts sur la localMap)						
			SpotMap map = globalMap.get(sender_ip);
			if(map==null) coherent=false; //on ne traite que si l'hote est connu, sinon on lui demandera sa LocalMap
			else {
				rand_i = mess.readShort();
				map_version =  mess.readShort();
				id_hub = mess.readString();
				ctarget = mess.readShort();
				int port = mess.readShort();
				int ri =  mess.readShort();
				if(rand_i==0 || rand_i!=map.rinstance) break; //on ignore le message s'il ne s'agit pas du master connu (-> on va attendre le basculement de ce master pour syncrhoniser sa map)
				if(map_version != HubManager.nextVersion(map.version) || map.targetExist(ctarget,0,0))  coherent=false; //incohérence de version, on vademander une synchro map	
				else {
					//maj de la map
					synchronized(globalMap) {
						globalmap_version = HubManager.nextVersion(globalmap_version);
						map.addTarget(this, id_hub, ctarget, port, ri, map_version, true); //la notification de la modification de la globalMap découlera de l'appel targetInMap(...)	
					}
				}
				map.wd=0; //reset du wd de ce NM
			}
			
			if(!coherent) { //on va demande la localmap complête du noeud emetteur: [85][GET_LOCALMAP]			
				send(master_socket, sender_adr, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_LOCALMAP), true); 
			}			
			break;
		case ProtocoleLocal.REMOVE_LOCAL:	//[REMOVE_LOCAL](short: rand_instance)(short: my_ipmap_new_version)(s: id_hub) (short: ctarget)(short: port)(short: rinstance) <- un NM distant signal une suppression de l'in de ses spots
			rand_i = mess.readShort();
			map_version =  mess.readShort();
			id_hub = mess.readString();
			ctarget = mess.readShort();
			int port = mess.readShort();
			int ri = mess.readShort();
			if(local_sender) { //C'est un spot local qui signal le down d'un de ses spots proxy
				if(port==sender_port) synchronized(slaveList) {
					for(int i=0; i<slaveList.size(); i++) {
						SlaveDescriptor sd = slaveList.get(i);
						if(sd.port==sender_port) {
							if(sd.randInstance==rand_i) localMap.removeTarget(this, id_hub, ctarget, 0, true);
							break; //on sort du for
						}	 
					}
				}
			} else { //c'est un master distant
				map = globalMap.get(sender_ip);
				if(map==null) coherent=false; //on ne traite que si l'hote est connu, sinon on lui demandera sa LocalMap
				else {
					if(rand_i==0 || rand_i!=map.rinstance) break; //on ignore le message s'il ne s'agit pas du master connu (-> on va attendre le basculement de ce master pour syncrhoniser sa map)

					if(map_version != HubManager.nextVersion(map.version) || !map.targetExist(ctarget,port, ri)) coherent=false; //incohérence de version, on va demander une maj de cette map	
					else map.removeTarget(this, id_hub, ctarget, map_version, true); //maj de map
				}
				if(!coherent) { //on va demande la localmap complête du noeud emetteur: [85][GET_LOCALMAP]		
					send(master_socket, sender_adr, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_LOCALMAP), true); 
				}
			}
			break;
		case ProtocoleLocal.GET_GLOBALMAP : //[85][GET_GLOBALMAP]	<- un NS a détecté une désycnrho de sa globalMap, il demande au NM sa global via ce message. La réponse sera un (ou plusieurs message) GLOBALMAP
			if(!local_sender) break; //cette demande ne peut provenir que d'un NS local
			synchronized(globalMap) {
				buildGlobalMap(); //on s'assure que le buffer d'envoie de la globalMap est à jour
				//transmission du message
				for(DataWriter dw :  builded_globalM) send(master_socket, sender_adr, sender_port, dw, false);
			}
			break;
		case ProtocoleLocal.GET_LOCALMAP : //[85][GET_LOCALMAP]	<- un NS ou un NM a détecté une désycnrho de la map du spot local, il demande au NM sa local via ce message. La réponse sera un (ou plusieurs message) LOCALMAP
			synchronized(localMap) {
				buildLocalMap(); //on s'assure que le buffer d'envoie de la localMap est à jour
				//transmission du message
				for(DataWriter dw :  builded_localM) send(master_socket, sender_adr, sender_port, dw, false);
			}
			break;
		case ProtocoleLocal.GET_SLAVELIST : //[85][GET_SLAVELIST]		<- la réponse sera un (ou plusieurs message) SLAVELIST
			synchronized(slaveList) {
				buildSlaveList(); //on s'assure que le buffer d'envoie de la slaveList est à jour
				//transmission du message
				for(DataWriter dw :  builded_slaveL) send(master_socket, sender_adr, sender_port, dw, false);
			}
			break;
		case ProtocoleLocal.LOCALMAP: //[LOCALMAP](short: rand_instance)(short: my_ipmap_version)[page][nb_page]{...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance) $ |...} |...}
			//Un NM envoie une maj de sa LocalMap, on doit recomposer les pages et l'analyser
			if(local_sender) break; //cette demande ne peut pas provenir d'un NS local
			rand_i = mess.readShort();
			map_version =  mess.readShort();
			map = globalMap.get(sender_ip);
			if(map!=null && rand_i!=map.rinstance) break; //on ignore le message si le rand_instance ne correspond pas à l'host connu (-> on va attendre le basculement de ce master pour syncrhoniser sa map)
			synchronized(received_ipM) {
				PageRecomposer precmp = received_ipM.get(sender_ip);
				if(precmp==null) { //c'est le premier paquet LOCALMAP pour cet hote
					precmp = new PageRecomposer();
					received_ipM.put(sender_ip, precmp);
				}
				if(precmp.add(rand_i, map_version, mess.cloneRest())) {
					received_ipM.remove(sender_ip);
					performReceivedIpMap(sender_ip, precmp);
				}
			}
			break;
		case ProtocoleLocal.SPOTLIST : //[SPOTLIST](short: rand_instance)[page][nb_page] {...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget) $ |...} |...}
			//Un NS envoie une maj de sa spotlist, on doit recomposer les pages et l'analyser
			if(!local_sender) break; //cette demande ne peut provenir que d'un NS local
			rand_i = mess.readShort();
			synchronized(received_spotL) {
				PageRecomposer precmp = received_spotL.get(sender_port);
				if(precmp==null) { //c'est le premier paquet SPOTLIST pour ce slave
					precmp = new PageRecomposer();
					received_spotL.put(sender_port, precmp);
				}
				if(precmp.add(rand_i, 0, mess.cloneRest())) {
					received_spotL.remove(sender_port);
					performReceivedSpotList(sender_port, rand_i, precmp);
				}
			}
			break;
		case ProtocoleLocal.DOWN : //[DOWN](sh: randi) <- le noeud émetteur signale la fin de son exécution
			if(!local_sender) break; //cette demande ne peut provenir que d'un NS local	
			rand_i = mess.readShort();
			synchronized(slaveList) {
				for(int i=0; i<slaveList.size(); i++) {
					SlaveDescriptor sd = slaveList.get(i);
					if(sd.port==sender_port) {
						if(sd.randInstance==rand_i) {
							localMap.removeTargetOnPort(this, sd.port, rand_i, 0, true);//down de tous les spots de ce slave
							slaveList.removeElementAt(i); slavelist_version=HubManager.nextVersion(slavelist_version); i--; //maj de la slavelist
							//notification de la maj de la slavelist aux NS: [85][REMOVE_SLAVE](short: master_rand_instance)(short: new_slavelist_version)(short: port)(short: rand_instance)
							broadcastSlave(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.REMOVE_SLAVE).sh(randInstance).sh(slavelist_version).sh(sd.port).sh(sd.randInstance), true,0);
						}
						break; //on sort du for
					}	 
				}
			}
			break;
		default :
			hubManager.fireKernelLog(Hub.LOG_WARNING, "Unknow command kernel receive for master mode ["+cmd+"]");
		}
		mess.dispose();
	}
	private void checkSlave(int sender_port, int rand_i) {
//System.out.println("== analyse checkSlave pour le slave : "+sender_port);		
		synchronized(slaveList) {
			SlaveDescriptor sd = null;
			for(int i=0; i<slaveList.size(); i++) {
				SlaveDescriptor cur = slaveList.get(i);
				if(cur.port==sender_port) {
					if(cur.randInstance!=rand_i) { //l'instance de ce slave a changé, on va down son ancienne version => down de tous ses spots	
						localMap.removeTargetOnPort(this, cur.port, cur.randInstance, 0, true);//down de tous les spots de ce slave					
						slaveList.removeElementAt(i); slavelist_version=HubManager.nextVersion(slavelist_version); //maj de la slavelist
						//notification de la maj de la slavelist aux NS: [85][REMOVE_SLAVE](short: master_rand_instance)(short: new_slavelist_version)(short: port)(short: rand_instance)
						broadcastSlave(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.REMOVE_SLAVE).sh(randInstance).sh(slavelist_version).sh(cur.port).sh(cur.randInstance), true,0);
					} else sd=cur;
					break; //sort du for
				}
			}
			if(sd==null) { //c'est un nouveau slave
//System.out.println("== détection d'un nouveau slave sur le port #"+sender_port+"#  ==");
				sd = new SlaveDescriptor(rand_i, sender_port);
				slaveList.add(sd);
				slavelist_version = HubManager.nextVersion(slavelist_version);
				
				//on envoie une MASTER_PULSE à ce slave pour qu'il prenne en compte la bonne valeur de randinstance du master
				send(master_socket, Network.localAddress, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.MASTER_PULSE).sh(randInstance).sh(localMap.version).sh(globalmap_version).sh(slavelist_version), true);
				
				//on notifie le up_slave de ce slaves aux autres slaves --> [85][ADD_SLAVE](short: master_rand_instance)(short: new_slavelist_version)(short: port)(short: rand_instance)
				broadcastSlave(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.ADD_SLAVE).sh(randInstance).sh(slavelist_version).sh(sender_port).sh(rand_i), true, sender_port);

				//on demande au slave sa spot_list
				send(master_socket, Network.localAddress, sender_port,  DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_SPOTLIST), true);
			}
			sd.wd = 0; //on reset le compteur wd de ce slave
		}
	}
	private void  performReceivedIpMap(String map_ip, PageRecomposer precomp) { //on a tous les paquets pour analyser la localMap de l'hote map_ip
		//precomp contient tous les paquets de description de la localMap du master, dans l'ordre des pages : 
		//	chaque paquet = [page][nb_page]{...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance)  $ |...} |...}
		//--> doit déclencher un ADD_MASTER vers les NS pour le cas d'un nouveau NM dans la globalMap
		if(precomp.lock()) { 			
			synchronized(globalMap){
				SpotMap map = globalMap.get(map_ip);
				if(map==null){ //c'est une nouvelle ip_map
					map = new SpotMap(map_ip, null);
					map.setRInstance(precomp.rinstance, 0);
					globalMap.put(map_ip, map);
					globalmap_version = HubManager.nextVersion(globalmap_version);

					//[ADD_MASTER](short: master_rand_instance)(short: globalmap_version)(dword: ip)(short: ip_rand_instance)
					DataWriter dw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.ADD_MASTER).sh(randInstance).sh(globalmap_version);
					map.writeIp(dw);
					dw.sh(map.rinstance);
					this.broadcastSlave(dw, true,0);
				} else {
					map.setRInstance(precomp.rinstance, map.version);
				}

				//analyse de cette map, NB: les notifications des changements vers les NS sera réalisé au prochain MASTER_PULSE quand les NS se rendrons compte du changement de version de la global_map
				if(performReceivedMap(precomp, map, -1, precomp.version)){ //au moin un changement a été appliqué
					globalmap_version = HubManager.nextVersion(globalmap_version);
				}
			}
			precomp.purge(true);
		}
	}
	private void  performReceivedSpotList(int slave_ri, int port, PageRecomposer precomp) { //on a tous les paquets pour analyser la spotlist du slave
		//precomp contient tous les paquets de description de la spotlist du slave, dans l'ordre des pages : 
		//	chaque paquet = [page][nb_page] {...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget) $ |...} |...}
		if(precomp.lock()) { 
			synchronized(localMap) {
				Vector<Target> presents = localMap.targetOnPort(port);			
				DataReader dr = precomp.curPage();
				
				String id_hub=null; //hub en cours de lecture
				int nb_target=0;
				
				while(dr!=null){
					if(dr.canRead()) {
						if(id_hub==null) { //on doit lire: (s: id_hub)(i:nb_target) 
							id_hub = dr.readString();
							nb_target = dr.readInt();
						} else if(nb_target>0) { //on doit lire (short: ctarget)
							int ctarget = dr.readShort();
							nb_target--;
							boolean is_new=true;
							
							//on regarde si cette cible est bien présente
							for(int i=0; i<presents.size(); i++){
								Target cur = presents.get(i);
								if(cur.ctarget==ctarget){ //cible trouvé
									if(cur.id_hub.equals(id_hub) && cur.rinstance==slave_ri) is_new=false; //tout est ok pour cette cible
									else localMap.removeTarget(this, cur.id_hub, ctarget, 0, true); //cette cible n'est pas dans le bon hub ou n'est pas sur le bon rinstance, on la supprime
									
									presents.removeElementAt(i); 
									break; //on sort du for
								}
							}
							if(is_new) {
								if(localMap.targetExist(ctarget, 0, 0)){
									//le code ctarget entre en conflit avec un code déjà existant, on envoie un message DENY_PLUG au contexte porteur
									int new_ctarget = newCTarget();
									if(new_ctarget<0) new_ctarget=0;
									//[85][DENY_PLUG](s: id_hub)(s: mess)(short: ctarget)(short: possible_ne_ctarget)
									this.send(master_socket, Network.localAddress, port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.DENY_PLUG).s("Code target allready in use").sh(ctarget).sh(new_ctarget), true);			
								} else localMap.addTarget(this, id_hub, ctarget, port, slave_ri, 0, true);  //c'est une nouvelle cible
							}
						} else id_hub = null; //fin d'analyse de ce hub
					} else dr = precomp.next_page(); //on doit changer de page
				}
				
				//fin d'analyse, les éléments encore dans presents doivent être supprimé
				for(int i=0; i<presents.size(); i++){
					Target cur = presents.get(i);
					localMap.removeTarget(this, cur.id_hub, cur.ctarget, 0, true);
				}
			}
			precomp.purge(true);
		}
	}
	private synchronized int newCTarget() { //-1 si impossible
		if(next_ctarget>65500) next_ctarget=1;
		if(localMap.targetCount()>=65000) return(-1); //nombre maximal atteind
		while(localMap.targetExist(next_ctarget, 0, 0)) { //on calcul le prochain code ctarget disponible
			next_ctarget++;
			if(next_ctarget>65500) next_ctarget=1;
		}
		return(next_ctarget++);
	}
	private void buildLocalMap() {
		synchronized(localMap) { if(localMap.version!=builded_localM_version) { //on doit reconstruire le buffer d'envoie de la localMap
			for(DataWriter dw : builded_localM) dw.dispose();
			builded_localM.removeAllElements();

			//[85][LOCALMAP](short: rand_instance)(short: my_ipmap_version)[page][nb_page]{...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance) $ |...} |...}
			//	==> $ désigne une position de coupure potentielle entre les paquets
			if(localMap.targetCount()==0) { //cas particulier de la localMap vide
				builded_localM.addElement(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.LOCALMAP).sh(randInstance).sh(localMap.version).b(0).b(0));
			} else {
				int page=1; //numéro de la page courante
				DataWriter cdw=null; 	//paquet en cours d'écriture
				cdw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.LOCALMAP).sh(randInstance).sh(localMap.version).b(page);
				cdw.b(0); //[7] nb_page sera fixé en (1)
				builded_localM.add(cdw);
				
				page = localMap.buildSpotMap(ProtocoleLocal.LOCALMAP, randInstance, localMap.version, builded_localM, page, cdw);
				
				//(1) on fixe les valeurs nb_page sur l'ensemble des paquets
				for(DataWriter dw : builded_localM) dw.setByte(7, page);
			}
			builded_localM_version = localMap.version;
		}}
	}
	private void buildSlaveList() {
		synchronized(slaveList) { if(slavelist_version!=builded_slaveL_version) {
			for(DataWriter dw : builded_slaveL) dw.dispose();
			builded_slaveL.removeAllElements();

			// [85][SLAVELIST](short: rand_instance)(short: slavelist_version)[page][nb_page]{...| (short: rand_instance)(short: port) $ |...}
			//	==> $ désigne une position de coupure potentielle entre les paquets
			if(slaveList.size()==0) { //cas particulier de la slaveList vide
				builded_slaveL.addElement(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.SLAVELIST).sh(randInstance).sh(slavelist_version).b(0).b(0));
			} else {
				int page=1; //numéro de la page courante
				DataWriter cdw=null; 	//paquet en cours d'écriture
				cdw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.SLAVELIST).sh(randInstance).sh(slavelist_version).b(page);
				cdw.b(0); //[7] nb_page sera fixé en (1)
				builded_slaveL.add(cdw);
				
				for(SlaveDescriptor sd : slaveList) {
					if(cdw.size()+4>ProtocoleLocal.MAX_UDP_BUFFER_SIZE) { //on doit opérer sur un nouveau paquet
						page++;
						cdw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.SLAVELIST).sh(randInstance).sh(slavelist_version).b(page);
						cdw.b(0); //[7] nb_page sera fixé en (1)
						builded_slaveL.add(cdw);
					}
					cdw.sh(sd.randInstance).sh(sd.port);	
				}
				
				//(1) on fixe les valeurs nb_page sur l'ensemble des paquets
				for(DataWriter dw : builded_slaveL) dw.setByte(7, page);
			}
			builded_slaveL_version = slavelist_version;			
		} }
	}
	private void buildGlobalMap() {
		synchronized(globalMap) { if(globalmap_version!=builded_globalM_version) { //on doit reconstruire le buffer d'envoie de la globalMap
			for(DataWriter dw : builded_globalM) dw.dispose();
			builded_globalM.removeAllElements();

			//-> [85][GLOBALMAP](short: rand_instance)(short: globalmap_version)[page][nb_page]
			//{...| 
			//		(dword: ip)(short: ip_rand_instance)(short: ipmap_version)(short: nb_hub) $ {...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance) $ |...} |...} 
			//|...}
			//	==> $ désigne une position de coupure potentielle entre les paquets
			
			if(globalMap.size()==0) { //cas particulier de la globalMap vide
				builded_globalM.addElement(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GLOBALMAP).sh(randInstance).sh(globalmap_version).b(0).b(0));
			} else {
				int page=1; //numéro de la page courante
				DataWriter cdw=null; 	//paquet en cours d'écriture
				cdw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GLOBALMAP).sh(randInstance).sh(globalmap_version).b(page);
				cdw.b(0); //[7] nb_page sera fixé en (1)
				builded_globalM.add(cdw);
				
				for(SpotMap map : globalMap.values()) {
					if(cdw.size()+10>ProtocoleLocal.MAX_UDP_BUFFER_SIZE) { //on doit opérer sur un nouveau paquet
						page++;
						cdw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GLOBALMAP).sh(randInstance).sh(globalmap_version).b(page);
						cdw.b(0); //[7] nb_page sera fixé en (1)
						builded_globalM.add(cdw);
					}	
					map.writeIp(cdw);
					cdw.sh(map.rinstance).sh(map.version).sh(map.hubs.size());
					
					page = map.buildSpotMap(ProtocoleLocal.GLOBALMAP, randInstance, globalmap_version, builded_globalM, page, cdw);
					cdw = builded_globalM.lastElement(); //on récupère le dernier paquet
				}
				
				//(1) on fixe les valeurs nb_page sur l'ensemble des paquets
				for(DataWriter dw : builded_globalM) dw.setByte(7, page);
			}
			builded_globalM_version = globalmap_version;
		}}
	} 	
	//-----------------------------------------------------------]
	
	//[----------------- Codes spécifiques au rôle Slave -----------------------------------------------------------
	Vector<JoinParam> waitingJoin = new Vector<JoinParam>(); 		//table des demandes joinHub en attente
	Vector<JoinParam>  waitingReadyPlug = new Vector<JoinParam>(); 	//table des signalements readyPlug en attente
	PageRecomposer received_globalM=null;	//cache du buffer de réception de la gloobalMap
	PageRecomposer received_localM=null;  	//cache du buffer de réception de la localMap
	PageRecomposer received_slaveL=null;  	//cache du buffer de réception de la slaveList

	private JoinParam getWaitingJoin(String id_hub, int jid){
		synchronized(waitingJoin) {
			for(JoinParam jp : waitingJoin) if(jp.jid==jid && jp.hub.id_hub.equals(id_hub)) return(jp);
		}
		return(null);
	}
	private JoinParam removeOneWaitingJoin(String id_hub) {
		synchronized(waitingJoin) {
			for(int i=0; i<waitingJoin.size(); i++) {
				JoinParam jp = waitingJoin.get(i); 
				if(jp.hub.id_hub.equals(id_hub)) {
					waitingJoin.removeElementAt(i);
					return(jp);
				}
			}
		}
		return(null);
	}
	private int removeAllWaitingJoin(String id_hub) {
		int nb=0;
		synchronized(waitingJoin) {
			for(int i=0; i<waitingJoin.size(); i++) {
				if(waitingJoin.get(i).hub.id_hub.equals(id_hub)) {
					waitingJoin.removeElementAt(i); i--;
					nb++;
				}
			}
		}
		return(nb);
	}
	private JoinParam getWaitingReadyPlug(int ctarget){
		synchronized(waitingReadyPlug) {
			for(JoinParam jp : waitingReadyPlug) if(jp.ctarget==ctarget) return(jp);
		}
		return(null);
	}
	private JoinParam removeReadyPlug(int ctarget) {
		synchronized(waitingReadyPlug) {
			for(int i=0; i<waitingReadyPlug.size(); i++) if(waitingReadyPlug.get(i).ctarget==ctarget) {
				return(waitingReadyPlug.remove(i));
			}
		}
		return(null);
	}
	
	protected synchronized void validJoinHub(String id_hub, int ctarget) {	//met en oeuvre la connexion au hub en attente sous l'id id_hub (on a reçu la commande ACCEPT_PLUG du NM)
		synchronized(waitingJoin) {
			JoinParam hparam = removeOneWaitingJoin(id_hub);
			if(hparam!=null) { //on s'assure que la demande est bien en attente
				hparam.send_count=0; hparam.ctarget=ctarget;
				synchronized(waitingReadyPlug) { waitingReadyPlug.add(hparam); }
				if(hparam.linker==null) hparam.hub.localSpotLinked(ctarget, listen_port); //c'est un spot local
				else { //c'est un spot proxy
					ManagedSpotProxy msp = new ManagedSpotProxy(hparam.hub, ctarget, listen_port, hparam.linker, hparam.jid);
					managedSpots.put(ctarget, msp);
				}
				//on doit notifier le NM, il se chargera alors de notifier le reste du réseau (dont les autres NS)
				sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.READY_PLUG).sh(randInstance).s(id_hub).sh(ctarget)); // [85][READY_PLUG](short: rinstance)(s: id_hub)(short: ctarget)
				//on va attendre sa réponse ADD pour ce spot local pour notifier que le hub est ready
			}
		}
	}
	protected void performSlaveMessage(boolean local_sender, InetAddress sender_adr , String sender_ip, int sender_port, DataReader mess) {
		String id_hub;
		int ctarget, rand_i, version, port, ri;
		boolean coherent = true; //mis à false en cas d'incohérence de version sur la localMap ou de globalMap
		if(local_sender && sender_port==ProtocoleLocal.MASTER_PORT) localMap.wd=0;	//on reset le wd du master local
		int cmd = mess.readByte();
		switch(cmd) {
		case ProtocoleLocal.MASTER_PULSE : //[85][MASTER_PULSE](short: rand_instance)(short: localmap_version)(short: globalmap_version)(short: slave_list_version)
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive MASTER_PULSE from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}	
			rand_i = mess.readShort();
			if(rand_i!=localMap.rinstance) { //le master local vient d'être découvert, ou il a changé, on va mettre à jour toutes les listes
				localMap.setRInstance(rand_i, localMap.version);
				sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_SLAVELIST));
				sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_LOCALMAP));
				sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_GLOBALMAP));
			} else {
				if(mess.readShort() != localMap.version) sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_LOCALMAP));
				if(mess.readShort() != globalmap_version) sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_GLOBALMAP));
				if(mess.readShort() != slavelist_version) sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_SLAVELIST));
			}
			break;
		case ProtocoleLocal.ACCEPT_PLUG:	//[ACCEPT_PLUG](s: id_hub)(i: ctarget)	<- le NM accepte une demande QUERY_PLUG et propose la valeur ctarget à utiliser
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive ACCEPT_PLUG from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			id_hub = mess.readString();
			ctarget = mess.readShort();
			validJoinHub(id_hub, ctarget);		
			break;
		case ProtocoleLocal.ADD_SLAVE: //[ADD_SLAVE](short: master_rand_instance)(short: new_slavelist_version)(short: port)(short: rand_instance)	
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive ADD_SLAVE from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();
			port = mess.readShort();
			rand_i = mess.readShort();

			//maj de la slaveList
			synchronized(slaveList) {
				if(version != HubManager.nextVersion(slavelist_version)) sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_SLAVELIST)); //incohérence de version, on va demander une maj de la slaveList		
				else {
					slaveList.add( new SlaveDescriptor(rand_i , port));
					slavelist_version = version;
				}
			}
			break;
		case ProtocoleLocal.REMOVE_SLAVE: //[REMOVE_SLAVE](short: master_rand_instance)(short: new_slavelist_version)(short: port)(short: rand_instance)
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive REMOVE_SLAVE from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();			
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();
			port = mess.readShort();
			rand_i = mess.readShort();

			//maj de la slaveList
			synchronized(slaveList) {
				coherent = false;
				if(version == HubManager.nextVersion(slavelist_version)){
					for(int i=0; i<slaveList.size(); i++) {
						SlaveDescriptor sd = slaveList.get(i); 
						if(sd.port==port && sd.randInstance==rand_i) {
							slaveList.removeElementAt(i); 
							slavelist_version = version;
							coherent = true;
							break; //on sort du for
						}
					}
				}
			}	
			if(!coherent) sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_SLAVELIST)); //incohérence de version, on va demander une maj de la slaveList
			break;
		case ProtocoleLocal.GET_SPOTLIST: // [85][GET_SPOTLIST]		<- la réponse sera un (ou plusieurs message) SPOTLIST
			//--> [85][SPOTLIST](short: rand_instance)[page][nb_page] {...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget) $ |...} |...}
			sendSpotList(sender_adr, sender_port);
			break;
		case ProtocoleLocal.ADD_LOCAL:	//[ADD_LOCAL](short: rand_instance)(short: new_loacalmap_version)(s: id_hub)(short: ctarget)(short: port)(short: rinstance)  <-- le masteur indique une mise à jour de la localmap
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive ADD_LOCAL from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();
			id_hub = mess.readString();
			ctarget = mess.readShort();
			port = mess.readShort();
			ri = mess.readShort();
			//maj de localMap
			synchronized(localMap) {
				if(version != HubManager.nextVersion(localMap.version) || localMap.targetExist(ctarget,0, 0)) coherent=false; //incohérence de version, on va demander une maj de la localMap
				else {
//System.out.println("---> application du ADD_LOCAL dans la localmap");					
					localMap.addTarget(this, id_hub, ctarget, port, ri, version, false);
				}
			}
			if(!coherent) sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_LOCALMAP)); //on demande une maj de la localMap
			break;
		case ProtocoleLocal.REMOVE_LOCAL:	//[REMOVE_LOCAL](short: rand_instance)(short: my_ipmap_new_version)(s: id_hub) (short: ctarget)(short: port)(short: rinstance)
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive REMOVE_LOCAL from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();
			id_hub = mess.readString();
			ctarget = mess.readShort();
			port = mess.readShort();
			ri = mess.readShort();
			
			//maj de localMap
			synchronized(localMap) {
				if(version != HubManager.nextVersion(localMap.version) || !localMap.targetExist(ctarget,port, ri)) coherent=false; //incohérence de version, on va demander une maj de la localMap		
				else localMap.removeTarget(this, id_hub, ctarget, version, false);
			}			
			if(!coherent) sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_LOCALMAP)); //on demande une maj de la localMap
			break;
		case ProtocoleLocal.ADD_DISTANT:	//[ADD_DISTANT](short: master_rand_instance)(short: globalmap_version)(dword: ip)(short: ip_rand_instance)(short: ipmap_version)(s: id_hub) (short: ctarget)(short: port)(short: rinstance) <-- le masteur indique une mise à jour de la globalmap
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive ADD_DISTANT from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();

			//maj de globalMap
			synchronized(globalMap) {
				if(version != HubManager.nextVersion(globalmap_version)) coherent = false; //incohérence détecté, on va demander une synchro globalmap
				else {
					String map_ip= mess.readStringIp();
					rand_i = mess.readShort();
					int map_version =  mess.readShort();

					SpotMap map = globalMap.get(map_ip);
					if(map==null || rand_i!=map.rinstance || map_version!=HubManager.nextVersion(map.version)) coherent = false; //incohérence détecté, on va demander une synchro globalmap
					else {
						id_hub = mess.readString();
						ctarget = mess.readShort();
						port = mess.readShort();
						ri = mess.readShort();
						if(map.targetExist(ctarget,0, 0)) coherent = false; //incohérence détecté, on va demander une synchro globalmap
						else { //tout est ok, on met à jour
							map.addTarget(this, id_hub, ctarget, port, ri, map_version, false);			
							globalmap_version = version;
						}
					}
				}
			}
			if(!coherent) { //on va demander une synchro globalmap au NM
				//[85][GET_GLOBALMAP]		<- la réponse sera un (ou plusieurs message) GLOBALMAP
				sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_GLOBALMAP));
			}
			break;	
		case ProtocoleLocal.REMOVE_DISTANT:	//[REMOVE_DISTANT](short: master_rand_instance)(short: globalmap_version)(dword: ip)(short: ip_rand_instance)(short: ipmap_version)(s: id_hub) (short: ctarget)(short: port)(short: rinstance)
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive REMOVE_DISTANT from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();

			//maj de globalMap
			synchronized(globalMap) {
				if(version != HubManager.nextVersion(globalmap_version)) {
//System.out.println("/// version globalmap non cohérent : curren = "+globalmap_version+" , given : "+version);					
					coherent = false; //incohérence détecté, on va demander une synchro globalmap
				} else {
					String map_ip= mess.readStringIp();
					rand_i = mess.readShort();
					int map_version =  mess.readShort();

					SpotMap map = globalMap.get(map_ip);
					if(map==null || rand_i!=map.rinstance || map_version!=HubManager.nextVersion(map.version)) coherent = false; //incohérence détecté, on va demander une synchro globalmap
					else {
						id_hub = mess.readString();
						ctarget = mess.readShort();
						port = mess.readShort();
						ri = mess.readShort();
						if(!map.targetExist(ctarget, port, ri)) coherent = false; //incohérence détecté, on va demander une synchro globalmap
						else { //tout est ok, on met à jour
							map.removeTarget(this, id_hub, ctarget, map_version, false);			
							globalmap_version = version;
						}
					}
				}
			}
			if(!coherent) { //on va demander une synchro globalmap au NM
				//[85][GET_GLOBALMAP]		<- la réponse sera un (ou plusieurs message) GLOBALMAP
				sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_GLOBALMAP));
			}			
			break;
		case ProtocoleLocal.ADD_MASTER: //[ADD_MASTER](short: master_rand_instance)(short: globalmap_version)(dword: ip)(short: ip_rand_instance)
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive ADD_MASTER from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();

			//maj de globalMap
			synchronized(globalMap) {
				if(version != HubManager.nextVersion(globalmap_version)) coherent = false; //incohérence détecté, on va demander une synchro globalmap
				else {
					String map_ip= mess.readStringIp();
					rand_i = mess.readShort();
	
					SpotMap map = globalMap.get(map_ip);
					if(map!=null) coherent = false; //incohérence détecté, on va demander une synchro globalmap
					else {
						map = new SpotMap(map_ip, null);
						map.setRInstance(rand_i, 0);
						globalMap.put(map_ip, map);
						globalmap_version = version;
					}
				}
			}
			if(!coherent) { //on va demander une synchro globalmap au NM
				//[85][GET_GLOBALMAP]		<- la réponse sera un (ou plusieurs message) GLOBALMAP
				sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_GLOBALMAP));
			}			
			break;
		case ProtocoleLocal.CHANGE_MASTER: //*bs* [85][CHANGE_MASTER](short: master_rand_instance)(short: globalmap_version)(dword: ip)(short: new_ip_rand_instance) <- changement du rand_instance d'un NM
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive CHANGE_MASTER from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();

			//maj de globalMap
			synchronized(globalMap) {
				if(version != HubManager.nextVersion(globalmap_version)) coherent = false; //incohérence détecté, on va demander une synchro globalmap
				else {
					String map_ip= mess.readStringIp();
					rand_i = mess.readShort();
	
					SpotMap map = globalMap.get(map_ip);
					if(map==null) coherent = false; //incohérence détecté, on va demander une synchro globalmap
					else {
						map.version = 0;
						map.rinstance = rand_i;
						globalmap_version = version;
					}
				}
			}
			if(!coherent) { //on va demander une synchro globalmap au NM
				//[85][GET_GLOBALMAP]		<- la réponse sera un (ou plusieurs message) GLOBALMAP
				sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_GLOBALMAP));
			}			
			break;
		case ProtocoleLocal.REMOVE_MASTER: //[85][REMOVE_MASTER](short: master_rand_instance)(short: globalmap_version)(dword: ip)(short: ip_rand_instance) <- suppression d'un master ==> tous ses spots devront être retiré de la globalMap et être passé à down
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive REMOVE_MASTER from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();

			//maj de globalMap
			synchronized(globalMap) {
				if(version != HubManager.nextVersion(globalmap_version)) coherent = false; //incohérence détecté, on va demander une synchro globalmap
				else {
					String map_ip= mess.readStringIp();
					rand_i = mess.readShort();
	
					SpotMap map = globalMap.get(map_ip);
					if(map==null || map.rinstance!=rand_i) coherent = false; //incohérence détecté, on va demander une synchro globalmap
					else {
						globalMap.remove(map_ip);
						map.removeTargetOnPort(this, 0, 0, map.version, false);//down de tous les spots de cet hote, sans notification
						globalmap_version = version;
					}
				}
			}
			if(!coherent) { //on va demander une synchro globalmap au NM
				//[85][GET_GLOBALMAP]		<- la réponse sera un (ou plusieurs message) GLOBALMAP
				sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.GET_GLOBALMAP));
			}			
			break;
		case ProtocoleLocal.DENY_PLUG: //[DENY_PLUG](s: id_hub)(s: mess)(short: ctarget)(short: possible_new_ctarget)	<- (ctarget ==0 si c'est suite à un QUERY_PLUG, et >0 si c'est suite à un READY_PLUG ou une re-syncrho) 
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive REPLUG from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			id_hub = mess.readString();
			String message = mess.readString();
			ctarget = mess.readShort();
			int new_ctarget = mess.readShort();
			
			if(ctarget==0){ //c'est une demande de QUERY_PLUG
				synchronized(waitingJoin) { 
					if(removeAllWaitingJoin(id_hub)>0) hubManager.fireKernelLog(Hub.LOG_ERROR, "Enable to obtain ctarget to plug to hub ["+id_hub+"], local master says : "+message);
				}
			} else { //c'est une demande de READY_PLUG
				synchronized(waitingReadyPlug) {
					ManagedSpot mspot = managedSpots.remove(ctarget);
					if(mspot!=null) { //c'est un spot connu...
						if(new_ctarget==0){
							hubManager.fireKernelLog(Hub.LOG_ERROR, "Enable to obtain ctarget to plug to hub ["+id_hub+"], local master says : "+message);
							removeReadyPlug(ctarget);
						} else {
							JoinParam jparam = getWaitingReadyPlug(ctarget);
							if(jparam==null) break; //cette situation ne devra pas être possible...
							jparam.ctarget = new_ctarget; //on procède au changement du ctarget
							mspot.changeCTargetAndPort(new_ctarget, listen_port);
							managedSpots.put(new_ctarget, mspot);

							//on doit notifier le NM, il se chargera alors de notifier le reste du réseau (dont les autres NS)
							sendToLocalMaster(DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.READY_PLUG).s(id_hub).sh(new_ctarget)); // [85][READY_PLUG](s: id_hub)(short: ctarget)
						}
					}
				}
			}
			break;			
		case ProtocoleLocal.LOCALMAP: //[LOCALMAP](short: rand_instance)(short: my_ipmap_version)[page][nb_page]{...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance) $ |...} |...}
			//Le NM envoie un maj de la LocalMap
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive LOCALMAP from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
//System.out.println("reception de la localmap....");			
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();
			if(received_localM==null) received_localM=new PageRecomposer();
			if(received_localM.add(version, mess.cloneRest())) performReceivedLocalMap(); //on a tous les paquets pour analyser la localMap recue
			break;
		case ProtocoleLocal.GLOBALMAP: //[GLOBALMAP](short: rand_instance)(short: globalmap_version)[page][nb_page]
			//{...| 
			//		(dword: ip)(short: ip_rand_instance)(short: ipmap_version)(short: nb_hub) $ {...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance) $ |...} |...} 
			//|...}
			//Le NM envoie un maj de la GloabalMap
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive GLOBALMAP from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la map)
			version = mess.readShort();				
			if(received_globalM==null) received_globalM=new PageRecomposer();
			if(received_globalM.add(version, mess.cloneRest())) performReceivedGlobalMap(); //on a tous les paquets pour analyser la globalMap recue
			break;	
		case ProtocoleLocal.SLAVELIST : // [85][SLAVELIST](short: rand_instance)(short: slavelist_version)[page][nb_page]{...| (short: rand_instance)(short: port) $ |...}
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive SLAVELIST from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break;
			}
			rand_i = mess.readShort();
			if(rand_i==0 || rand_i!=localMap.rinstance) break; //on ignore le message s'il ne s'agit pas du master local courant (-> on va attendre le basculement de master pour syncrhoniser la slavelist)
			version = mess.readShort();				
			if(received_slaveL==null) received_slaveL=new PageRecomposer();
			if(received_slaveL.add(version, mess.cloneRest())) performReceivedSlaveList(); //on a tous les paquets pour analyser la slaveList recue
			break;	
		case ProtocoleLocal.DOWN : //[DOWN](sh: randi) <- le noeud master émetteur signale la fin de son exécution
			rand_i = mess.readShort();
			if(!local_sender || sender_port!=ProtocoleLocal.MASTER_PORT) {
				hubManager.fireKernelLog(Hub.LOG_WARNING, "Receive DOWN from a non local master (sender: "+sender_ip+":"+sender_port+")");
				break; //cette demande ne peut provenir que du master local				
			}
			if(rand_i == localMap.rinstance) ThreadStack.perform(() -> this.becomeMaster(), 400); //on va prendre le contrôle du rôle maitre dans 400ms
			break;			
		default :
			hubManager.fireKernelLog(Hub.LOG_WARNING, "Unknow command kernel receive for slave mode ["+cmd+"]");			
		}
		mess.dispose();
	}
	protected void sendToLocalMaster(DataWriter dw) { //dw sera disposé aprés transmission
		try {
			socket.send(dw.toDatagramPacket(Network.localAddress,ProtocoleLocal.MASTER_PORT));
		}catch(Exception e) {
			hubManager.fireKernelLog(Hub.LOG_WARNING, "Enable to send message to local master ("+e.getMessage()+")");
		}
		dw.dispose();
	}
	private void sendSpotList(InetAddress sender_adr, int sender_port) {
		Map<String, HubLocal> localHubs = hubManager.getAllLocalHubs();
		synchronized(localHubs) {
			if(localHubs.size()==0) { //cas particulier de la spotlist vide
				send(socket, sender_adr, sender_port, DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.SPOTLIST).sh(randInstance).b(0).b(0), true);
			} else {
				int page=1; //numéro de la page courante
				DataWriter cdw=null; 	//paquet en cours d'écriture
				Vector<DataWriter> packet_list=new Vector<DataWriter>();
				cdw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.SPOTLIST).sh(randInstance).b(page);
				cdw.b(0); //[5] nb_page sera fixé en fin de traitement
				packet_list.add(cdw);

				for(HubLocal ihub : localHubs.values()) { if(ihub.getState()==Hub.HUB_READY) {
					int old_pos = cdw.size(); 
					cdw.s(ihub.getHubId()).i(1); //OXO: voir à modifier ici en cas de présence de spot proxy
					if(cdw.size()>ProtocoleLocal.MAX_UDP_BUFFER_SIZE) { //on doit opérer sur un nouveau paquet
						cdw.setSize(old_pos);
						page++;
						cdw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.SPOTLIST).sh(randInstance).b(page);
						cdw.b(0); //[5] nb_page sera fixé par le code appelant
						packet_list.add(cdw);
						cdw.s(ihub.getHubId()).i(1); //OXO: voir à modifier ici en cas de présence de spot proxy
					}
					if(cdw.size()+2>ProtocoleLocal.MAX_UDP_BUFFER_SIZE) { //on doit opérer sur un nouveau paquet
						page++;
						cdw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.SPOTLIST).sh(randInstance).b(page);
						cdw.b(0); //[5] nb_page sera fixé en (1)
						packet_list.add(cdw);
					}	
					cdw.sh(ihub.getLocalSpotIdLocal());
				}}

				//on fixe les valeurs nb_page sur l'ensemble des paquets et on les transmets au demandeur
				for(DataWriter dw : packet_list) {
					dw.setByte(5, page);
					send(socket, sender_adr, sender_port, dw, true);
				}
				packet_list=null;
			}
		}
	}
	
	private void performReceivedLocalMap() { //on a tous les paquets pour analyser la localMap recue (received_localM + received_localM_version)
//System.out.println("... on essais de lock ...");		
		if(received_localM.lock()) { //received_localM contient tous les paquets de description de la localMap du master, dans l'ordre des pages : 
			//	chaque paquet = [page][nb_page]{...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance)  $ |...} |...}
//System.out.println("... traitement de la localmap");		
			performReceivedMap(received_localM, localMap, -1, received_localM.version);

			received_localM.purge(true);
			received_localM=null;
		}
	}
	private void performReceivedGlobalMap() { //on a tous les paquets pour analyser la globalMap recue (received_globalM + received_globalM_version)
		if(received_globalM.lock()) { //received_globalM contient tous les paquets de description de la globalMap du master, dans l'ordre des pages : 
			//	chaque paquet = [page][nb_page]{...| (dword: ip)(short: ip_rand_instance)(short: ipmap_version)(short: nb_hub) $ {...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance)  $ |...} |...} |...}
			synchronized(globalMap) {
				Vector<String> know_map = Util.copyStringIterable(globalMap.keySet()); //capture de la liste des ip des maps connus, on y retira l'ip de tous les maps présentes dans le message (en fin d'analyse il restera donc celles supprimées)
				
				DataReader dr = received_globalM.curPage();
				while(dr!=null){
					if(dr.canRead()) {
						//on doit lire: (dword: ip)(short: ip_rand_instance)(short: ipmap_version)(short: nb_hub) 
						String ip_map = dr.readStringIp(); 
						int ip_ri = dr.readShort();
						int version = dr.readShort();
						int nb_hub = dr.readShort();
							
						SpotMap map = globalMap.get(ip_map);
						if(map==null){ //c'est une nouvelle ip_map
							map = new SpotMap(ip_map, null);
							map.setRInstance(ip_ri, version);
							globalMap.put(ip_map, map);
						} else {
							know_map.remove(ip_map);
							map.setRInstance(ip_ri, version);
						}

						//analyse de cette map
						performReceivedMap(received_globalM, map, nb_hub, version);
						
					} else dr = received_globalM.next_page(); //on change de page
				}
	
				//on a fini l'analyse
				if(know_map.size()>0) { //on regarde si on a des map en plus dans l'ancienne version de la globalMap...
					for(String ip_map : know_map) { //... oui, on doit tous les supprimer
						SpotMap map = globalMap.remove(ip_map);
						if(map!=null){
							map.removeTargetOnPort(this, 0, 0, map.version, false);//down de tous les spots de cet hote, sans notification
						}
					}
				}
				globalmap_version = received_globalM.version;
			}
			
			received_globalM.purge(true);
			received_globalM=null;
		}
	}
	private void performReceivedSlaveList() { //on a tous les paquets pour analyser la slaveList recue
		if(received_slaveL.lock()) { //received_slaveL contient tous les paquets de description de la slaveList du master, dans l'ordre des pages : 
			//	chaque paquet = [page][nb_page]{...| (short: rand_instance)(short: port) $ |...}, received_slaveL.version à la valeur de la slavelist_version décrit
			slaveList.removeAllElements();
			synchronized(slaveList) {
				for(int i=0; i<received_slaveL.stack.size(); i++) {
					DataReader dr = received_slaveL.stack.get(i);
					dr.skip(2); //on saute  [page][nb_page]
					while(dr.canRead()) {
						int ri = dr.readShort();
						int port = dr.readShort();
						slaveList.add(new SlaveDescriptor(ri, port)); 
					}
				}
			}
			slavelist_version = received_slaveL.version;
			received_slaveL.purge(true);
			received_slaveL=null;
		}
	}
	private synchronized void becomeMaster() {
		if(master_socket==null) { //si ce n'est pas le cas, c'est que le rôle master à déjà été obtenu par l'instance
			//[On va essayer d'obtenir le contrôle du port maitre
			try {
				/*if(Util.isMac()) { master_socket = new DatagramSocket(null); master_socket.bind(new InetSocketAddress("127.0.0.1", ProtocoleLocal.MASTER_PORT)); }
				else*/ master_socket=new DatagramSocket(ProtocoleLocal.MASTER_PORT); //pour Windows et linux 
				master_socket.setBroadcast(true); 
				hubManager.fireKernelLog(Hub.LOG_INFO, "Master role obtained");
			} catch(Exception e){ master_socket=null; return; } //Le port maitre n'est pas disponible
			//]
			localMap.setRInstance(randInstance, localMap.version); //on prend le contrôle de la localMap
			newCTarget(); next_ctarget+=99; //on réalise un décalage de 100 unité dans next_ctarget afin de limiter les conflits sur les attributions ctarget en attente
			if(next_ctarget>65500) next_ctarget=50;
			
			//On lance le processus d'écoute de la socket master
			(new SocketListener(this, master_socket)).start();

			synchronized(slaveList) {	//on s'assure d'être présent dans la slaveList
				SlaveDescriptor me = getSlave(listen_port);
				if(me==null || me.randInstance!=randInstance) {
					if(me!=null) { //on doit down ce slave
						localMap.removeTargetOnPort(this, me.port, me.randInstance, 0, false);//down de tous les spots de ce slave					
						slaveList.remove(me); //maj de la slavelist
					}
					slaveList.add(new SlaveDescriptor(randInstance, listen_port)); 
					slavelist_version = HubManager.nextVersion(slavelist_version);		
					//la notification découlera de la signalisation de l'obtention du rôle maitre effectuée à la fin de cette méthode
				}
			}	
			
			synchronized(managedSpots) { //vérification des éléments waitingJoin et waitingReadyPlug en attente 
				synchronized(waitingJoin) { 
					for(JoinParam hubp : waitingJoin) {
						if(hubp.linker==null) {
							int ctarget = newCTarget();
							if(ctarget>0) { 
								hubp.hub.localSpotLinked(ctarget, listen_port);
								hubp.hub.setReady();

								synchronized(localMap) {
									localMap.addTarget(this, hubp.hub.id_hub, ctarget, listen_port, randInstance, 0, false);						
								}

								//On doit analyser l'existence des targets de localMap et de globalMap qui sont reliée à ce hub pour instancier leur spot et les mettres dans otherSpots
								searchAllTargetForHub(this, hubp.hub.id_hub, false);
							}
						} else {
							HubLocal hub = hubp.hub;
							int ctarget = newCTarget();
							if(ctarget>0) {
								ManagedSpotProxy msp = new ManagedSpotProxy(hub, ctarget, listen_port, hubp.linker, hubp.jid);
								managedSpots.put(ctarget, msp); 

								synchronized(localMap) {
									localMap.addTarget(this, hub.id_hub, ctarget, listen_port, randInstance, 0, false); //la notification de la modification de la localMap découlera de l'appel targetInMap(...)						
								}
							}
						}
					}
					waitingJoin.clear();
				}
				synchronized(waitingReadyPlug) { 
					for(JoinParam hubp : waitingReadyPlug) {					
						HubLocal hub = hubp.hub;
						if(localMap.targetExist(hubp.ctarget, listen_port, randInstance)) continue; //cette demande READY est déjà traité 
						else if(localMap.targetExist(hubp.ctarget, 0, 0)) { //erreur de conflit sur ctarget, on doit changer de ctarget ou analyser le cas DENY_PLUG
							int new_ctarget = newCTarget(); //génération d'un nouveau ctarget
							if(new_ctarget>0) {
								ManagedSpot mspot = managedSpots.remove(hubp.ctarget);
								if(mspot!=null) { //c'est un spot connu...
									mspot.changeCTargetAndPort(new_ctarget, listen_port);
									managedSpots.put(new_ctarget, mspot);
									hubp.ctarget = new_ctarget;
								} else continue; //spot non connu, ce cas de figure ne devrait pas être possible
							} else {
								hub.setError("Enable to produce more ctarget to plug to hub ["+hubp.hub.id_hub+"]");
								continue;
							}
						}//sinon tout est ok, on ajoute le spot correspondant dans les listes

						//maj de localMap
						synchronized(localMap) {
							localMap.addTarget(this, hubp.hub.id_hub, hubp.ctarget, listen_port, randInstance, 0, false);
						}
						//NB: comme on est plus slave les traitement ci-dessous ne seront pas fait dans targetInMap, on les fait donc maintenant						
						if(hubp.linker!=null) hub.enter((SpotImpl) managedSpots.get(hubp.ctarget), randInstance); //c'est un spot proxy
						else {
							hub.setReady(); 		//c'est maintenant qu'on fixe le hub ready
						
							//On doit aussi charger les spots de localMap et global de ce hub et les mettres dans otherSpots et notifier leur existence
							searchAllTargetForHub(this, hub.getHubId(), false);
						}
					}
					waitingReadyPlug.clear();
				}

				//vérification de la présence des spots locaux dans la localMap
				Vector<Target> presents = localMap.targetOnPort(listen_port);
				//on regarde si tous les présents dans la localMap existent bien
				for(int i=0; i<presents.size(); i++) {
					Target cur = presents.get(i);
					if(!managedSpots.containsKey(cur.ctarget)) { //détection d'un spot fantôme dans la localMap -> on le supprime
						localMap.removeTarget(this, cur.id_hub, cur.ctarget, 0, false);
						presents.removeElementAt(i); i--;
					}
				}
				if(presents.size()!=managedSpots.size()) { //il manque au moins un spots géré dans la localMap
					Vector<Integer> to_remove=null;
					Vector<ManagedSpot> to_add=null;

					for(ManagedSpot mspot : managedSpots.values()) {
						int ctarget = mspot.getIdLocal();
						if(!localMap.targetExist(ctarget, listen_port, randInstance)) { //détection d'un spot géré non présent dans la localMap -> on doit l'ajouter
							HubLocal hub = mspot.getLocalHub();
							if(localMap.targetExist(ctarget, 0, 0)) { //le code ctarget est déjà exploité !!!
								hubManager.fireKernelLog(Hub.LOG_ERROR, "Internal code ctarget already use to plug to hub ["+hub.getHubId()+"]");
								//on doit le changer
								int new_ctarget = newCTarget(); //génération d'un nouveau ctarget
								if(new_ctarget>0) {
									mspot.changeCTargetAndPort(new_ctarget, listen_port);
									if(to_remove==null) to_remove=new Vector<Integer>();
									if(to_add==null) to_add=new Vector<ManagedSpot>();
									to_remove.add(ctarget);
									to_add.add(mspot);
								} else {
									hubManager.fireKernelLog(Hub.LOG_ERROR, "Enable to produce more ctarget to plug to hub ["+hub.getHubId()+"]");
									continue;
								}
							}
							localMap.addTarget(this, hub.getHubId(), ctarget, listen_port, randInstance, 0, false);
						}
					}

					if(to_remove!=null) for(int i=0; i<to_remove.size(); i++) managedSpots.remove(to_remove.get(i));
					if(to_add!=null) for(ManagedSpot cur : to_add) managedSpots.put(cur.getIdLocal(), cur);
				}
			}
			
			//purge des buffers received_globalM, received_localM et received_slaveL
			if(received_globalM!=null && received_globalM.purge(false)) received_globalM=null;  
			if(received_localM!=null && received_localM.purge(false)) received_localM=null; 
			if(received_slaveL!=null && received_slaveL.purge(false)) received_slaveL=null; 
			
			//On notifie l'acquisition du rôle master //[MASTER_PULSE](short: rand_instance)(short: localmap_version)(short: globalmap_version)(short: slave_list_version)
			DataWriter dw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.MASTER_PULSE).sh(randInstance).sh(localMap.version).sh(globalmap_version).sh(slavelist_version);
			broadcastMaster(dw, false);
			broadcastSlave(dw, true,0); 
		}
	}
	//-----------------------------------------------------------]
	
	//[Méthode pour les instances de SpotMap
		//interface TargetFinder :
	public void targetInMap(SpotMap map, Target target, boolean notify_change) { //indique que la target est présent dans la map, notify_change==true ==>  on doit notifier la modification de cette map sur le réseau
//System.out.println("=== targetInMap ===> ctarget = "+target.ctarget+" port = "+target.port);	
		if(map!=localMap || target.port!=listen_port) { //ce n'est pas une cible du contexte courant... -> on doit créer un SpotUDP
			HubLocal hub = hubManager.getLocalHub(target.id_hub);
			if(hub!=null) {
				String idtarget=map.ip+'#'+target.ctarget;				
				synchronized(otherSpots) { if(!otherSpots.containsKey(idtarget)) {
//System.out.println("enregistrement d'un spot non local sous l'idTarget = "+idtarget);					
					SpotImpl spot = new SpotUdp(hub, target.ctarget, map.ip, map.getIadr(), target.port, target.rinstance); 
					otherSpots.put(idtarget, spot);
//System.out.println("... ===> injection dans otherSpots sous l'id : "+idtarget);						
					hub.enter(spot, target.rinstance);
				}}
			}
		} else if(master_socket==null) {//sinon, si on est slave on doit regarder si c'est un spot en attente ReadyPlug
//System.out.println("=== analyse de waitingReadyPlug pour targetInMap ...");	
			HubImpl hub = hubManager.getLocalHub(target.id_hub);
			if(hub!=null) synchronized(waitingReadyPlug) {
				JoinParam jparam = removeReadyPlug(target.ctarget);
				if(jparam!=null) {
					ManagedSpot mspot = managedSpots.get(target.ctarget);
					if(jparam.linker!=null) hub.enter((SpotImpl) mspot.getSpot(), randInstance);
					else { //c'est un spot local 
//System.out.println("\t ... waitingReadyPlug trouvé !!!");	
						hub.setReady(); 		//c'est maintenant qu'on active ce hub

						//On doit aussi charger les spots de localMap et global de ce hub et les mettres dans otherSpots et notifier leur existence
						searchAllTargetForHub(this, target.id_hub, false);
					}
				}
//else System.out.println("\t ... waitingReadyPlug NON trouvé !!! ctarget = "+target.ctarget);
			}		
		} else { //on est master, le spot est déjà créé dans managedSpots
			HubLocal hub = hubManager.getLocalHub(target.id_hub);
			if(hub!=null) {
				ManagedSpot mspot = managedSpots.get(target.ctarget);
				if(!mspot.isLocalSpot()) hub.enter((SpotImpl) mspot.getSpot(), randInstance); //c'est un spot proxy
			}
		}
		
		if(notify_change && master_socket!=null) { //on est le noeud master, et on doit notifier la modification de cette map sur le réseau
			if(map==localMap) { //c'est une modification de la localMap
				DataWriter wcmd = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.ADD_LOCAL).sh(randInstance).sh(localMap.version).s(target.id_hub).sh(target.ctarget).sh(target.port).sh(target.rinstance); 
				//on notifie les autres NM avec [ADD_LOCAL] pour actualiser leur globalMap
				broadcastMaster(wcmd, false); 	//[85][ADD_LOCAL](short: rand_instance)(short: my_ipmap_new_version)(s:id_hub)(short: ctarget)(short: port)(short: rinstance)
				//et les NS pour actualiser leur localMap
				broadcastSlave(wcmd, true, 0);		//[85][ADD_LOCAL](short: rand_instance)(short: new_loacalmap_version)(s: id_hub)(short: ctarget)(short: port)(short: rinstance)				
			} else { //cela concerne un hote distant
				//Notification ADD_DISTANT aux NS locaux
				// *bs* [85][ADD_DISTANT](short: master_rand_instance)(short: globalmap_version)(dword: ip)(short: ip_rand_instance)(short: ipmap_version)(s: id_hub) (short: ctarget)(short: port)(short: rinstance) 
				DataWriter wcmd = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.ADD_DISTANT).sh(randInstance).sh(globalmap_version);
				map.writeIp(wcmd);
				wcmd.sh(map.rinstance).sh(map.version).s(target.id_hub).sh(target.ctarget).sh(target.port).sh(target.rinstance); 
				broadcastSlave(wcmd, true, 0);		//[85][ADD](short: rand_instance)(short: new_loacalmap_version)(s: id_hub)(short: ctarget)(short: port)
			}
		}
	}
	protected void targetOutMap(SpotMap map, Target target, boolean notify_change) { //indique que la target n'est plus présente dans la map, notify_change==true ==>  on doit notifier la modification de cette map sur le réseau
//ystem.out.println("=== targetOutMap pour targetId = "+(map.ip+"#"+target.ctarget)); int x;	
		if(map==localMap && target.port==listen_port) { //c'est un spot local qui vient d'être retiré de la localMap !
			if(master_socket==null) { //cette situation découle d'un ordre de suppression du master (normalement ce sera toujours le cas)
				ManagedSpot mspot = managedSpots.get(target.ctarget);
				if(mspot!=null) {
					if(mspot.isUp()) sendSpotList(Network.localAddress, ProtocoleLocal.MASTER_PORT); //on va envoyer une SPOT_LIST au master pour le forcer à réintégrer ce spot
					//else if(!mspot.getSpot().isLocalSpot()) mspot.getLocalHub().exit((SpotImpl) mspot);
				}
			}
		} else synchronized(otherSpots) { 		
			SpotImpl spot = otherSpots.remove(map.ip+"#"+target.ctarget);
			if(spot!=null) {
//System.out.println("    ===> cible trouvé.");					
				spot.getHub().exit(spot, target.rinstance); //c'est un spot non géré et instancié, on le down
			}
//else System.out.println("    ===> cible NON trouvé."); 			
		}
		
		if(notify_change && master_socket!=null) { //on est le noeud master, et on doit notifier la modification sur le réseau
			if(map==localMap) { //cette suppression concerne la localMap, on notify en broadcast total (NM + NS)
				//[85][REMOVE_LOCAL](short: rand_instance)(short: my_ipmap_new_version)(s: id_hub) (short: ctarget)(short: port)(short: rinstance)
				DataWriter wcmd = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.REMOVE_LOCAL).sh(randInstance).sh(localMap.version).s(target.id_hub).sh(target.ctarget).sh(target.port).sh(target.rinstance); 
				//on notifie les autres NM pour actualiser leur globalMap
				broadcastMaster(wcmd, false); 
				//et les NS pour actualiser leur localMap
				broadcastSlave(wcmd, true, 0);
			} else { //la suppression concerne une map distante, on notify en broadcast local (NS)
				//[85][REMOVE_DISTANT](short: master_rand_instance)(short: globalmap_version)(dword: ip)(short: ip_rand_instance)(short: ipmap_version)(s: id_hub) (short: ctarget)(short: port)(short: rinstance)
				globalmap_version = HubManager.nextVersion(globalmap_version);
				DataWriter dw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(ProtocoleLocal.REMOVE_DISTANT).sh(randInstance).sh(globalmap_version);
				map.writeIp(dw);
				dw.sh(map.rinstance).sh(map.version).s(target.id_hub).sh(target.ctarget).sh(target.port).sh(target.rinstance); 
				broadcastSlave(dw, true, 0);
			}
		}
	}
	//]
	
	public void spotSend(SpotUdp spot_t, DataWriter dw, int ack) { //dw sera disposé
		DatagramPacket packet = dw.toDatagramPacket(spot_t.address, spot_t.getPort());
		try { socket.send(packet); }
		catch(Exception e) { hubManager.fireKernelLog(Hub.LOG_WARNING, "Enable to send message ("+e.getMessage()+")"); }
		
		if(ack!=0) manageAck(new WaitAckPacket(spot_t, dw, packet, ack)); //on va gérer l'acquittement de ce message
		else dw.dispose(); //on peut disposer le paquet
	}
	protected void searchAllTargetForHub(TargetFinder finder, String id_hub, boolean flag) {
		localMap.searchTargetForHub(finder, id_hub, flag);	//les spots de la localMap (on n'instancira pas les spot géré localement car eux sont déjà dans managedSpots)

		//les spots de la globalMap
		synchronized(globalMap) {
			for(SpotMap lm : globalMap.values()) lm.searchTargetForHub(finder, id_hub, flag);
		}
	}
	public Spot getSpotByIdTarget(String idTarget) {
//System.out.println("recherche de la cible  = "+idTarget);		
		Spot spot = otherSpots.get(idTarget);
		if(spot==null) {
//System.out.println("ce n'est pas un otherSpots !");				
			//il faut testé si spot n'est pas un spot géré localement
			int ddotp = idTarget.indexOf("#");
			if(ddotp>0) {
				String ip = idTarget.substring(0, ddotp);
				int ct=0;
				try { ct = Integer.parseInt(idTarget.substring(ddotp+1)); }
				catch(Exception e) { return(null); }				
				return(getSpot(ip, ct));
			}
		}
		return(spot);
	}
	public Spot getSpot(String ip, int ctarget) {
//System.out.println("recherche de la cible ip = "+ip+" ctarget = "+ctarget);		
		Spot spot=null;
		if(Network.isLocalAddress(ip)) {
			ip=Network.LOCALHOST;
			ManagedSpot mspot = managedSpots.get(ctarget);
			if(mspot!=null) spot=mspot.getSpot();
		} 
		if(spot==null) spot = otherSpots.get(ip+"#"+ctarget);
		return(spot);
	}
	public Spot getSpotByIdLocal(int local_id){
		Spot spot = null;
		//on regarde si la cible est dans le contexte d'exécution local
		ManagedSpot mspot = managedSpots.get(local_id);
		if(mspot!=null) spot=mspot.getSpot();
		if(spot==null) spot = otherSpots.get(Network.LOCALHOST+'#'+local_id); //on regarde si la cible est dans un autre contexte de la machine hote
		return(spot);
	}
	
	//[Gestion des ré-emssion pour les paquets à acquittements
	Runnable resender=null; 		//code de ré-emission des paquets non acquités
	TimerTask timer_resender=null;	//processus de ré-emission
	Vector<WaitAckPacket> pwaits = new Vector<WaitAckPacket>();
	public boolean performAck(SpotUdp producer, int num_ack){ //acte la réception du paquet de numéro indiqué par le receiver
		synchronized(pwaits){ for(int i=0; i<pwaits.size(); i++){
			WaitAckPacket cur = pwaits.get(i);
			if(cur.target==producer && cur.ack==num_ack) {
				cur.content.dispose();
				pwaits.removeElementAt(i);
				if(pwaits.size()==0){ //arret du processus de ré-emission des paquets non acquitté
					timer_resender.cancel(); 
					timer_resender=null;
				}
				producer.econtext.notifyAck(num_ack);
				return(true);
			}
		}}
		return(false);
	}
	protected void manageAck(WaitAckPacket wap){ //ajoute le paquet indiqué dans la pile de traitement des acquittements
		synchronized(pwaits){
			pwaits.add(wap);
			if(pwaits.size()==1){
				//démarrage du processus de ré-emission des paquets non acquitté
				if(resender==null){ 
					resender = new Runnable(){ //code de ré-emission des paquets non acquités
						public void run(){
							synchronized(pwaits){ 
								 for(int i=0; i<pwaits.size(); i++){
										WaitAckPacket cur = pwaits.get(i);
										if(cur.target.isUp() && cur.count<ProtocoleLocal.MAX_RESEND) try{ 
//System.out.println("--> recend no ack");		
											cur.count++;
											if(cur.count>1) socket.send(cur.packet); //on ne ré-emet pas sur le 1er cycle
										} catch(Exception e){} 	
										else {
											if(cur.count>=ProtocoleLocal.MAX_RESEND) hubManager.fireKernelLog(Hub.LOG_WARNING, "No acked message for target : "+cur.target.getIpPort());
											cur.content.dispose();
											pwaits.removeElementAt(i); i--;
											if(pwaits.size()==0){ //arret du processus de ré-emission des paquets non acquitté
												timer_resender.cancel(); 
												timer_resender=null;
												return;
											}
										}
								 }
							}
						}
					};
				}
				timer_resender = ThreadStack.perform(resender, 200, 200);
			}
		}
	}	
	//]
	
	//[Méthodes utiles
	@SuppressWarnings("unchecked")
	private boolean performReceivedMap(PageRecomposer prcomp, SpotMap map, int nb_max_hub, int new_version) { //on a tous les paquets pour analyser une map		
		//	chaque paquet = [page][nb_page]{...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance) $ |...} |...}
		//si nb_hub==-1 on analyse tant qu'une donnée est lisible, sinon on va extraire nb_hub lors de l'analyse
		//retourne true si des changements sont apporté à la map
		synchronized(map) {
//System.out.println("performReceivedMap sur map_ip="+map.ip);			
			int old_version=map.version;
			int nb_hub=0; 						//nombre de hub analysé
			String id_hub=null; 				//id de l'hub en cours d'analyse
			Vector<Target> know_targets = null;	//target connu pour l'hub en cours d'analyse
			int nb_target = 0; 					//nb d'info target restant à lire pour l'hub en cours d'analyse
			int ctarget, port, ri;
			Vector<String> know_hub = Util.copyStringIterable(map.hubs.keySet()); //capture de la liste des id des hubs connus, on y retira l'id de tous les hub présent dans le message (en fin d'analyse il restera donc ceux supprimé)
//System.out.println("--- Nb hub connu (1) = "+know_hub.size()+"  version = "+new_version);
			DataReader dr = prcomp.curPage();
			while(dr!=null && (nb_max_hub==-1 || nb_hub<nb_max_hub)){
				if(dr.canRead()) {
					if(id_hub==null) { //on doit lire: (s: id_hub)(i:nb_target) 
						id_hub = dr.readString();
						nb_target = dr.readInt();
						know_targets = map.hubs.get(id_hub);
						if(know_targets!=null) know_targets=(Vector<Target>) know_targets.clone();
//System.out.println("--- analyse du hub : "+id_hub+" , nbr de cible annoncé = "+nb_target+" , nbre de cible connu = "+(know_targets==null ? 0 : know_targets.size()));						
						know_hub.remove(id_hub);
//System.out.println("--- Nb hub connu (2) = "+know_hub.size());						
					} else if(nb_target>0) { //on doit lire (short: ctarget)(short: port)(short: rinstance)
						ctarget = dr.readShort();
//System.out.println("--- analyse de la cible, ctarget = "+ctarget);						
						port = dr.readShort();
						ri = dr.readShort();
						nb_target--;

						if(know_targets==null) {
//System.out.println("--- nouvelle cible target dans la map (nouveau hub), ctarget = "+ctarget);							
							map.addTarget(this, id_hub, ctarget, port, ri, 0, false); //c'est un nouveau hub, on est en insertion directe
						} else {
							boolean is_new=true;
							//on regarde si cette cible est connu
							for(int j=0; j<know_targets.size(); j++) {
								Target cur = know_targets.get(j);
								if(cur.ctarget==ctarget) { //la cible est connu
									if(cur.port!=port || cur.rinstance!=ri) {
//System.out.println("---+++ remplacement de la cible ctarget = "+ctarget);											
										map.removeTarget(this, id_hub, ctarget, 0, false); //mais elle n'est pas sur le bon port ou randinstance, on fait une suppression
									}
									else is_new=false; //sinon tout est ok pour cette target
									know_targets.removeElementAt(j);
									break; //on sort du for
								}
							}
							if(is_new) {
//System.out.println("--- nouvelle cible target dans la map, ctarget = "+ctarget);								
								map.addTarget(this, id_hub, ctarget, port, ri, 0, false); 
							}
						}
					} else { //on a fini l'analyse du hub d'id id_hub, zone [Z1]
//System.out.println("--- fin d'analyse (1) du hub : "+id_hub+" , nbre de cible connu non retrouvé = "+(know_targets==null ? 0 : know_targets.size()));						
						if(know_targets!=null && know_targets.size()>0) { //on regarde si on a des target en plus dans l'ancienne version de la map...
							for(int j=0; j<know_targets.size(); j++) { //... oui, on doit tous les supprimer
//System.out.println(">>>>> suppression de : "+ know_targets.get(j).ctarget);								
								map.removeTarget(this, id_hub, know_targets.get(j).ctarget, 0, false);
							}
							know_targets=null;
						}
						id_hub=null; //fin d'analyse pour cet id_hub
						nb_hub++;
					}
				} else dr = prcomp.next_page(); //on change de page
			}
			if(id_hub!=null) { //NB: pour le dernier hub analysé on ne rentre pas dans la zone [Z1], on doit donc potentiellement faire son analyse ici
//System.out.println("--- fin d'analyse (2) du hub : "+id_hub+" , nbre de cible connu non retrouvé = "+(know_targets==null ? 0 : know_targets.size()));						
				if(know_targets!=null && know_targets.size()>0) { //on regarde si on a des target en plus dans l'ancienne version de la map...
					for(int j=0; j<know_targets.size(); j++) { //... oui, on doit tous les supprimer
//System.out.println(">>>>> suppression de : "+ know_targets.get(j).ctarget);								
						map.removeTarget(this, id_hub, know_targets.get(j).ctarget, 0, false);
					}
					know_targets=null;
				}		
				id_hub=null; //fin d'analyse pour cet id_hub
				nb_hub++;
			}
//System.out.println("--- Nb hub connu (3) = "+know_hub.size());	
			if(know_hub.size()>0) { //fin d'analyse, on regarde si on a des hubs en plus
				for(String hub_id : know_hub) {
//System.out.println(">>>>> suppression du hub : "+ hub_id);						
					map.removeHub(this, hub_id, 0, false); 
				}
			}
			boolean change = (old_version != map.version) || (old_version!=new_version);
			map.version=new_version;
			
			return(change);
		}
	}
	protected void send(DatagramSocket soc, InetAddress sender_adr, int port, DataWriter dw, boolean dispose) { //dw sera disposé aprés transmission
		try {
			soc.send(dw.toDatagramPacket(sender_adr,port));
		}catch(Exception e) {
			hubManager.fireKernelLog(Hub.LOG_WARNING, "Enable to send message ("+e.getMessage()+")");
		}
		if(dispose) dw.dispose();
	}
}

class JoinParam {
	HubLocal hub;
	int ctarget;
	int jid;
	int send_count;	//compteur d'envoie
	Linker linker;
	
	public JoinParam(HubLocal vhub) { //vjid==0 pour un spot local, et !=0 pour un spot proxy
		hub = vhub;
		ctarget=-1;
		send_count=0;
		jid = 0;
	}
	public JoinParam(HubLocal vhub, Linker l, int vjid) { //vjid==0 pour un spot local, et !=0 pour un spot proxy
		hub = vhub;
		ctarget=-1;
		send_count=0;
		linker = l;
		jid = vjid;
	}
}

class SlaveDescriptor {
	int port;
	int randInstance;
	int wd;
	
	public SlaveDescriptor(int r, int p) {
		randInstance = r;
		port = p;
		wd=0;
	}
}

class PageRecomposer { //classe utilisé pour recomposer les messages envoyés en plusieurs pages fractionné
	//les paquets doivent être de la forme : [page][nb_page]{... content ...}
	 Vector<DataReader> stack;
	 int version, rinstance, nb_page;
	 int wd;
	 boolean locked;
	 
	 public PageRecomposer() {
		 stack = null;
		 wd=0;
		 locked = false;
	 }
	 
	 public boolean add(int vversion, DataReader dr) { //retourne true si la pile est complête
		 return(add(rinstance, vversion, dr));
	 }
	 
	 public synchronized boolean lock() {
		 if(locked || (nb_page>0 && nb_page<stack.size())) return(false);
		 locked=true;
		 return(true);
	 }
	 
	 public synchronized boolean purge(boolean force_mode) {
		 if(locked && !force_mode) return(false); 
		 for(DataReader odr : stack) odr.dispose();
		 stack.removeAllElements();
		 wd=0;
		 locked = false;
         cur_pos_page=0;
         cur_page=null;
		 return(true);
	 }
	 
	 public synchronized boolean add(int vrinstance, int vversion, DataReader dr) { //retourne true si la pile est complête
//System.out.println("PageRecomposer : locked = "+locked);		 
		 if(locked) return(false);
		 wd=0;
		 int dr_nbpage =  dr.getByte(dr.curPos()+1);
//System.out.println("PageRecomposer : dr_nbpage = "+dr_nbpage);		 
		 if(stack ==null || version!=vversion || nb_page!=dr_nbpage || vrinstance!=rinstance) { //on doit réinitialiser le buffer de réception
			 version = vversion;
			 rinstance = vrinstance;
			 nb_page = dr_nbpage;
			 if(stack == null) stack = new Vector<DataReader>();
			 else purge(true);
			 stack.add(dr);
		 } else {
			 int page =  dr.getByte(0);
			 boolean insered = false;
			 for(int p=0; p<stack.size(); p++) {
				 int cur_page = stack.get(p).getByte(0);
				 if(cur_page==page) { dr.dispose(); insered = true; break; } //on a déjà recu cette page
				 else if(cur_page>page) { //on doit insérer à cette position	
					 stack.insertElementAt(dr, p);
					 insered = true; 
					 break;
				 }
			 }				
			 if(!insered) stack.add(dr); //on insère à la fin
		 }
//System.out.println("PageRecomposer : "+stack.size()+" / "+nb_page);		 
		 return(nb_page==0 || nb_page==stack.size());
	 }
	 
	 
	 int cur_pos_page=0; 		//position de la page courante
	 DataReader cur_page=null;	//page courante
	 public DataReader curPage(){
		 if(cur_page==null && cur_pos_page<stack.size()){
			 cur_page = stack.get(cur_pos_page);
			 cur_page.skip(2); //on saute l'entête [page][nb_page]
		 }
		 return(cur_page);
	 }
	 public DataReader next_page(){ //on passe à la page suivante
		 cur_pos_page++;
		 cur_page=null;
		 return(curPage());
	 }
}