package skuad.sim.implementation;

import skuad.sim.*;
import java.util.*;
import skuad.sim.simop.*;
import skuad.util.time.temporality.*;
import skuad.util.time.Pulse;

public abstract class SimContext implements InitSimOperation {
	//NB: quand la simulation subit un reset l'ensemble des informations de cette classe est ré-initialisé !
	
	public final static long DEFAULT_PING = 40; //ping par défaut (en ms)
	
	//[Constante d'état possible pour le contexte de la simulation
	// SimPilot.OFF
	// SimPilot.INITIALIZING
	// SimPilot.IN_RUN
	// SimPilot.STOPPED
	// SimPilot.ENDED
	//NB: dans le contexte de simulation les états PAUSED/PAUSE_PLANNED et RESUMED ne sont pas visible
	//]
	
	//Les composants simulés
	int state;
	boolean component_set_locked, component_started;
	
	//NB: les ajoutes dans ces listes sont automatique started si state==SimContext.IN_RUN
	//Leur manipulation add/remove doivent être synchronisé sur l'instance contexte
	// si component_set_locked==true les add/remove doivent être bufferisés
	Vector<VirtualEnv> virtual_envs;
	Vector<InternalPulsable> pulsables;
	
	Vector<WaitingComponent> wainting_components;
	
	TemporalityScheduler scheduler;	//moteur à temporalité
	SimProcess sim_process;
	
	public SimContext() {
		state = SimPilot.OFF;
		component_set_locked = false;
		component_started = false;
		
		virtual_envs = new Vector<VirtualEnv>();
		pulsables = new Vector<InternalPulsable>();
		wainting_components = new Vector<WaitingComponent>();
		
		sim_process=null;
		scheduler=new TemporalityScheduler();
	}

	public void log(String mess) {
		sim_process.log(scheduler.curTime(), mess);
	}
	
	//[Les simop
	public InitSimOperation initSimOperation() { return(this); }

	public abstract ReadSimOperation readSimOperation();
	public abstract StartSimOperation startSimOperation();
	public abstract RunSimOperation runSimOperation();
	public abstract void startAllAgent();//on demarre les agents qui ont besoin de l'être
	public abstract void stopAllAgent();//on demarre les agents qui ont besoin de l'être
	//]
	
	//Traitement de l'exécution de la simulation
	public void run(SimProcess vsim_process) {
		sim_process = vsim_process;
		boolean continue_run=true;
		while(continue_run) { //ici on a forcément state==OFF 
			state=SimPilot.INITIALIZING; 
			scheduler.reset();
			scheduler.setPing(DEFAULT_PING);

			
			if(sim_process.initSim()) {	
				//on start les composants simulés
				lockComponentSets();
				component_started = true;
				//les environnements simulé
				for(VirtualEnv vens : virtual_envs) vens.start(sim_process.sim_pilot, startSimOperation());
				//les objets pulsables
				for(InternalPulsable ips : pulsables) ips.start();
				state=SimPilot.IN_RUN;
				unlockComponentSets();

				startAllAgent(); //on demarre les agents qui ont besoin de l'être
				
				long previous_tc=-1;
				long next_tc=scheduler.goStartTime(); //fixe scheduler.curTime() sur la valeur startTime

				while(sim_process.preCycle(next_tc)) { //ici on a : next_tc == scheduler.curTime()
					lockComponentSets();
					boolean env_need_other_cycle=false;
					for(VirtualEnv vens : virtual_envs) { //préparation des environnements virtuels
						env_need_other_cycle = vens.initializeNextTime(sim_process.sim_pilot, startSimOperation(), runSimOperation(), previous_tc, next_tc) || env_need_other_cycle;
						if(!sim_process.performWaitAction(SimPilot.IN_RUN, true, false)) { continue_run=false; break; }
					}
					unlockComponentSets();
					if(!continue_run) break;

					if(scheduler.terminated() || (!env_need_other_cycle && scheduler.empty()) ) { //on a atteind la fin de la simulation fini par dépassement de temps ou absence d'activité, ou d'une simulation infini par absence d'activité  (pas d'activité pulsable, et pas de demande de nouveau cycle aprés netx_tc des environnements virtuels)
						state=SimPilot.ENDED;
						break;
					}

					//Déclenchement du prochain cycle sur le timecode next_tc (sans effet si aucun pulse sur ce time_code)
					scheduler.runCurTime();

					if(!sim_process.endCycle(next_tc)) break;

					//détermination du TC pour le prochain cycle
					previous_tc = next_tc;
					next_tc = scheduler.nextTC(); //next_tc <=0 si aucun et simulation infini				
//System.out.println("next_tc = "+next_tc);
					//on va consulter les environnements virtuels sur cette proposition
					lockComponentSets(); 
					for(VirtualEnv vens : virtual_envs) {
						long need_tc=vens.candidateNextTime(sim_process.sim_pilot, startSimOperation(), runSimOperation(), previous_tc, next_tc);
						if(!sim_process.performWaitAction(SimPilot.IN_RUN, true, false)) { continue_run=false; break; }
						if(need_tc>0 && (next_tc<=0 || need_tc<next_tc)) next_tc=need_tc;
					}
					unlockComponentSets();
					if(!continue_run) break;
//System.out.println("need_tc = "+next_tc);
					//Pour le cas ou les environnements virtuels aurait impacté les temporalités
					long revised_tc = scheduler.nextTC(); //next_tc <=0 si aucun et simulation infini	
					if(revised_tc>0 && (next_tc<=0 || revised_tc<next_tc)) next_tc=revised_tc;
//System.out.println("revised_tc = "+next_tc);					
					if(next_tc<=0) { //on a atteind la fin de la simulation pour une simulation infini par absence de demande de next_tc (pas d'activité pulsable, et pas de candidat netx_tc des environnements virtuels)
						state=SimPilot.ENDED;
						break;
					}
					
					scheduler.setCurTime(next_tc); //Préparation du prochain cycle sur next_tc
					scheduler.setPing(0);
				}	
			}
			if(state!=SimPilot.ENDED) state=SimPilot.STOPPED; 	
			
			sim_process.endSim(state, scheduler.curTime()); //Traitement de l'état de fin de simulation
			lockComponentSets();
			if(component_started) {
				component_started = false;
				stopAllAgent(); //les agents	
				//les objets pulsables
				for(InternalPulsable ips : pulsables) ips.stop(true);
				//les environnements simulé
				for(VirtualEnv vens : virtual_envs) vens.stop(sim_process.sim_pilot);
			}
			unlockComponentSets();
			continue_run=sim_process.performWaitAction(SimPilot.INITIALIZING, false, true);
		}
	}
	//]
	
	//[Pour les créations des pulses de InternalPulsable
	public Pulse newPulse(PulseTarget target, String label, long delay) { //ponctuel
		if(state==SimPilot.INITIALIZING || state==SimPilot.IN_RUN) {
			return(new InternalTemporalRule(scheduler, target, label, delay, 0));
		}
		return(null);
	}
	public Pulse newPulse(PulseTarget target, String label, long delay ,long periode) { //périodique
		return(new InternalTemporalRule(scheduler, target, label, delay, periode));
	}
	//]
	
	//[Les composants de la simulation
	protected void lockComponentSets() { //--> les manipulations sur les listes virtual_envs//pulsables//agents sont buffrésié
		synchronized(wainting_components) {
			component_set_locked=true;
		}
	}
	protected void unlockComponentSets() { 
		synchronized(wainting_components) {
			component_set_locked=false;
			//on débufferise les manipulations sur les listes virtual_envs//pulsables//agents 
			for(WaitingComponent wc : wainting_components) {
				if(wc.venv!=null) {
					if(wc.add) {
						virtual_envs.add(wc.venv);
						if(component_started) wc.venv.start(sim_process.sim_pilot, startSimOperation());
					} else if(virtual_envs.remove(wc.venv) && component_started) wc.venv.stop(sim_process.sim_pilot);
				} else {
					if(wc.add) {
						pulsables.add(wc.ipulsable);
						if(component_started) wc.ipulsable.start();
					} else if(pulsables.remove(wc.ipulsable)) wc.ipulsable.stop(component_started);					
				}
			}
			wainting_components.removeAllElements();
		}
	}
	
	private InternalPulsable getIPulsable(Pulsable pulsable) {
		for(InternalPulsable ipulsable : pulsables) {
			if(ipulsable.pulsable==pulsable) return(ipulsable);
		}
		return(null);
	}
	//]
	
	//[Pour SimProcess
	protected synchronized void reset() { //on annule toutes les sims opérations !
		state = SimPilot.OFF;
		lockComponentSets();
		if(component_started) {
			component_started = false;
			stopAllAgent(); //les agents
			//les objets pulsables
			for(InternalPulsable ips : pulsables) ips.stop(true);
			//les environnements simulé
			for(VirtualEnv vens : virtual_envs) vens.stop(sim_process.sim_pilot);
		}
		unlockComponentSets();
		virtual_envs.removeAllElements();
		pulsables.removeAllElements();
		wainting_components.removeAllElements();
		scheduler.reset();
	}	
    //]
	
	//[InitSimOperation
	public boolean setTimes(long time_start, long vping, long time_stop) {
		if(state==SimPilot.INITIALIZING) {
			scheduler.reset(time_start, time_stop);
			scheduler.setPing(vping);
			return(true);
		}
		return false;
	}
	public boolean setMinimalStep(long mstep) {		//fixe le pas de temps minimum, valeur par défaut 0
		if(state==SimPilot.INITIALIZING) {
			scheduler.setMinimalStep(mstep);
			return(true);
		}
		return false;		
	}
	//]

	//[SimOp de base
	public int getState() { return(state); }
	public long curTime() { return(scheduler.curTime()); }
	public long getStartTime() { return(scheduler.timeStart()); }
	public long getPingTime() { return(scheduler.getPing()); }
	public long getEndTime() { return(scheduler.timeEnd()); }
	
	public void addVirtualEnv(VirtualEnv venv) {
		synchronized(wainting_components) {
			if(component_set_locked) wainting_components.add(new WaitingComponent(true, venv));
			else {
				virtual_envs.add(venv);
				if(component_started) venv.start(sim_process.sim_pilot, startSimOperation()); //si la simulation est démmaré il faut appliquer un start(...)
			}
		}
	}	
	public boolean removeVirtualEnv(VirtualEnv venv) {
		synchronized(wainting_components) {
			if(component_set_locked) {
				if(virtual_envs.contains(venv)) {
					wainting_components.add(new WaitingComponent(false, venv));
					return(true);
				} 
			} else {
				if(virtual_envs.remove(venv)) {
					if(component_started) venv.stop(sim_process.sim_pilot);
					return(true);
				} 				
			}
			return(false);
		}
	}
	public void addPulsable(Pulsable pulsable) { // pas de vérification de double ajout
		synchronized(wainting_components) {
			InternalPulsable ipulsable=new InternalPulsable(sim_process.sim_pilot, this, pulsable);
			if(component_set_locked) wainting_components.add(new WaitingComponent(true, ipulsable));
			else {
				pulsables.add(ipulsable);
				if(component_started) ipulsable.start(); //si la simulation est démmaré il faut appliquer un start(...)
			}
		}	
	}
	public boolean removePulsable(Pulsable pulsable) {
		synchronized(wainting_components) {
			InternalPulsable ipulsable = getIPulsable(pulsable);
			if(ipulsable==null) return(false);
			if(component_set_locked) wainting_components.add(new WaitingComponent(false, ipulsable));
			else {
				pulsables.remove(ipulsable);
				ipulsable.stop(component_started);
			}
			return(true);
		}
	}
	//]
}

class WaitingComponent {
	boolean add;
	VirtualEnv venv;
	InternalPulsable ipulsable;
	
	public WaitingComponent(boolean vadd, VirtualEnv vvenv) {
		add = vadd;
		venv = vvenv;
		ipulsable=null;
	}
	
	public WaitingComponent(boolean vadd, InternalPulsable vipulsable) {
		add = vadd;
		venv = null;
		ipulsable=vipulsable;
	}
}