/**
 # Les composants simulés :
 Il y a trois sortes de composants simulés :
 	- les agents : ce sont des instances d'AgentU
 	- les Pulsable : ce sont des composants qui ont une rythmique comportemetale sans pour autant être des agents
 	- les VirtaulEnv : ce sont des environnements virtuel qui peuvent influer sur la progression du temps
 
 --> c'est composants ont tous une méthode start(...) et stop(...):
 	- start(...) : invoqué à l'activation du composant
 			- si le composant existe avec le 1er cycle de simulation :
 				--> la méthode est déclenché aprés la phase d'initialisation de la simu et juste avant le 1er cycle
 			- si le composant est créé aprés le 1er cycle et que la simulation est en cours :
 				--> la méthode est déclenché dès l'enregistrement du composant dans le simulateur
 
 	- stop(...) : invoqué à l'arret du composant
 			-> soit quand le composant est désenregistré de la simu quand celle-ci est en cours
 			-> soit quand la simu à fini sont exécution (ENDED ou STOPPED)
 
 # Fonctionnement du pilotage du processus de simulation :
	- si verrou wait_actions ouvert :
		- si process_actif :
			==> les actions doivents être empilées 
			si state = PAUSED --> le processus est en pause
 	 			==> l'état doit passer à RESUMED
 	 			==> le processus doit être réveillé
 			
		- si process_non_actif : --> le processus n'est pas démarré
			- state = OFF --> seul l'action start est éligible ->  état doit passer à INITIALIZING + on lance le processus
			sinon on a forcément : state = ENDED || STOPPED --> seul les actions RESET et START sont éligibles
				- reset --> internalReset();
				- start --> internalReset() + l'état doit passer à INITIALIZING + on lance le processus
 
 */
package skuad.sim.implementation;

import java.util.Vector;
import skuad.sim.*;
import skuad.util.thread.ThreadStack;
import skuad.sim.simop.*;

public class SimProcess implements Runnable {
	
	//[Constante d'état possible pour le process de la simulation
	// SimPilot.OFF
	// SimPilot.INITIALIZING
	// SimPilot.IN_RUN
	// SimPilot.PAUSE_PLANNED
	// SimPilot.PAUSED
	// SimPilot.RESUMED
	// SimPilot.STOPPED
	// SimPilot.ENDED
	//]
	
	//[Constante wait_action
	public final static int NONE=0;
	public final static int TO_START=1;
	public final static int TO_STOP=2;
	public final static int TO_RESET=3;
	public final static int TO_PAUSE=4;
	//]
	
	SimPilot sim_pilot;
	SimContext sim_context;
	int state;
	Vector<Integer> wait_actions;			//Pile FIFO des actions de pilotage en attente
	Vector<SimInitializer> sim_initializer;	//les initialiseur de la simulation
	Vector<SimObserver> sim_observers;		//les observers de la simulation
	Vector<ResetListener> reset_listeners;		//les observers de la simulation
	boolean process_actif;
	
	public SimProcess(SimPilot vsim_pilot, SimContext vsim_cnontext) {
		sim_pilot = vsim_pilot;
		state = SimPilot.OFF; 
		wait_actions=new Vector<Integer>();
		sim_initializer = new Vector<SimInitializer>();
		sim_observers = new Vector<SimObserver>();
		reset_listeners = new Vector<ResetListener>();
		sim_context = vsim_cnontext;
		process_actif=false;
	}
	
	public int getProcessState() { return(state); }
	
	//[Interface Runnabele
	public void run() { sim_context.run(this); }
	//]
	
	//[pour SimAU
	public void addInitializer(SimInitializer simInit) {synchronized(sim_initializer) {  
		if(!sim_initializer.contains(simInit)) sim_initializer.add(simInit);
	}}
	public boolean removeInitializer(SimInitializer simInit) { synchronized(sim_initializer) { return(sim_initializer.remove(simInit)); } }
	public void addSimObserver(SimObserver simo) {synchronized(sim_observers) {  
		if(!sim_observers.contains(simo)) sim_observers.add(simo);
	}}
	public boolean removeSimObserver(SimObserver simo) { synchronized(sim_observers) { return(sim_observers.remove(simo)); } }
	public void addResetListener(ResetListener rlisten) {synchronized(reset_listeners) {  
		if(!reset_listeners.contains(rlisten)) reset_listeners.add(rlisten);
	}}
	public boolean removeResetListener(ResetListener rlisten) { synchronized(reset_listeners) { return(reset_listeners.remove(rlisten)); } }	
	//]
	
	//[Pilotage de la simulation
	public boolean cmd(int action) {
		synchronized(wait_actions) {
			if(process_actif) {
				addWaitAction(action); //on met l'action en attente
				if(state==SimPilot.PAUSED) { //on doit réveiller le processus de traitement !
					state=SimPilot.IN_RUN;
					wait_actions.notify();
				}				
			} else if(state==SimPilot.OFF) { 
				if(action==TO_START) { //on doit lancer le processus de simulation
					state=SimPilot.INITIALIZING; 
					process_actif=true;
					ThreadStack.perform(this);
				} else return(false); //seul l'action TO_START est possible pour l'état OFF
			} else { //ici state == ENDED || STOPPED 
				if(action==TO_RESET) internalReset();
				else if(action==TO_START) {
					internalReset();
					return(cmd(TO_START));
				} else return(false); //seul les actions TO_START ou RESET sont possibles ici
			}
			return(true);
		}
	}
	//]
	
	//[Pour SimContext
	protected boolean initSim() { //false si la simu doit s'arréter
		fireSimEvent(SimPilot.SIM_INIT, -1);
	
		if(performWaitAction(SimPilot.INITIALIZING, false, false)) { //on lance les traitements des initialiseurs
			synchronized(sim_initializer) {
				for(SimInitializer siminit : sim_initializer) {
					siminit.initSim(sim_pilot, sim_context.initSimOperation(), sim_context.startSimOperation());
					if(!performWaitAction(SimPilot.INITIALIZING, false, false)) return(false);
				}
			}
			state=SimPilot.IN_RUN;
			fireSimEvent(SimPilot.SIM_START, 0); //Démmarrage effectif de la simulation
			return(performWaitAction(SimPilot.IN_RUN, false, false));
		}
		return(false);
	}
	protected boolean preCycle(long cur_time) { //false si la simu doit s'arréter
		fireSimEvent(SimPilot.BEGIN_CYCLE, cur_time);
		return(performWaitAction(SimPilot.IN_RUN, true, false));
	}
	protected boolean endCycle(long cur_time) {
		fireSimEvent(SimPilot.END_CYCLE, cur_time);
		return(performWaitAction(SimPilot.IN_RUN, true, false));
	}
	protected void endSim(int sim_state, long cur_time) { //Traitement de l'état de fin de simulation, ici: sim_state=ENDED || STOPPED
		synchronized(wait_actions) {
			state=sim_state;
			if(state == SimPilot.ENDED) fireSimEvent(SimPilot.SIM_END, cur_time);
			else fireSimEvent(SimPilot.SIM_STOP, cur_time);
		}
	}
	//]
	
	//[Méthodes interne
	protected void addWaitAction(int val) {
		if(wait_actions.isEmpty()) wait_actions.add(val);
		else if(wait_actions.lastElement().intValue()!=val) wait_actions.add(val);
	}
	protected void pushHeadAction(int val) {
		if(wait_actions.isEmpty()) wait_actions.add(val);
		else if(wait_actions.firstElement().intValue()!=val) wait_actions.insertElementAt(val, 0);
	}
	protected int popAction() {
		if(wait_actions.isEmpty()) return(NONE);
		else return(wait_actions.remove(0));
	}
	protected void fireSimEvent(int type, long time_code) {
		synchronized(sim_observers) {
			for(SimObserver simo : sim_observers) simo.simEvent(sim_pilot, type, sim_context.readSimOperation(), time_code);
		}
	}
	protected void fireResetSim() {
		synchronized(reset_listeners) {
			for(ResetListener rlisten : reset_listeners) rlisten.resetSim(sim_pilot);
		}
	}
	//]
	
	protected void log(long time, String mess) {
		synchronized(sim_observers) {
			for(SimObserver simo : sim_observers) simo.simLog(sim_pilot, time, mess);
		}
	}
	
	protected boolean performWaitAction(int start_state, boolean perform_pause, boolean simulation_ended) { //true si le processus de simulation doit continuer son traitement
		//si  perform_pause===true, et les pauses seront appliquées (on aura alors start_state==IN_RUN), sinon elles seront mis en si  attente via state=PAUSE_PLANNED
		//NB: on a forcément start_state==INITIALIZING || IN_RUN 
		synchronized(wait_actions) {
			int next_action;
			while( (next_action=popAction())!=NONE ) { //on a forcément :
				//si simulation_ended==false : state== INITIALIZING || IN_RUN || PAUSE_PLANNED
				//si simulation_ended==true : state== OFF || STOPPED || ENDED
				switch(next_action){
				case TO_PAUSE :  
					if(!simulation_ended && state!=SimPilot.PAUSED) state=SimPilot.PAUSE_PLANNED; 
					//sinon action non éligible
					break;
				case TO_STOP :   
					if(!simulation_ended) { state=SimPilot.STOPPED; return(false); }
					//sinon action non éligible
					break;
				case TO_START :  
					if(!simulation_ended) {
						if(state==SimPilot.PAUSED) {
							state=start_state;
							ThreadStack.perform(()->fireSimEvent(SimPilot.SIM_RESUME, sim_context.curTime()));
						} else state=start_state;
					} else {
						if(state==SimPilot.STOPPED || state==SimPilot.ENDED) internalReset();
						state=SimPilot.INITIALIZING;
						return(true); //on relance la simulation
					}
					break;
				case TO_RESET : 
					if(!simulation_ended) {
						pushHeadAction(TO_RESET);
						state=SimPilot.STOPPED; 
						return(false);
					} else internalReset();
					break;
				}
			}
			
			if(state==SimPilot.PAUSE_PLANNED && perform_pause) {
				state=SimPilot.PAUSED; 
				ThreadStack.perform(()->fireSimEvent(SimPilot.SIM_PAUSE, sim_context.curTime()));
			}
			
			if(state==SimPilot.PAUSED) { //on doit mettre le processus en pause
				try{ wait_actions.wait(); } catch(Exception e) { e.printStackTrace(); }
				fireSimEvent(SimPilot.SIM_RESUME, sim_context.curTime());
				return(performWaitAction(SimPilot.IN_RUN, perform_pause, simulation_ended));
			}
			if(simulation_ended) {
				process_actif=false;
				return(false);
			} else return(true);
		}
	}
	private void internalReset() {
		if(state!=SimPilot.OFF) {
			ThreadStack.perform(()->fireSimEvent(SimPilot.SIM_RESET, 0)); 
			fireResetSim();
			sim_context.reset(); //réinitialise complètement le contexte de simulation
			state=SimPilot.OFF;
		}
	}
	//]
}