package skuad.ubiquity;

import skuad.ubiquity.implementation.*;
import skuad.ubiquity.space.*;
import java.util.HashMap;
import java.util.Map;

import skuad.hubudp.*;

/**
 * <h3>Point d'entrée proncipal de l'API Ubiquity.</h3>
 * Cette classe est une classe statique (elle ne s’instancie pas), elle est donc ainsi accessible directement en tout point du code.<br/>
 * Cette classe permet d’entreprendre les deux tâches principales offertes par la plateforme :<ul>
 * <li>Créer un participant dans un space (pour interagir avec des unités de code locales ou distantes) : via la méthode {@link #newInteractor(String, String, String)} </li>
 * <li>Obtenir le descripteur d’un space (pour analyser et observer ce qui se passe dans ce space): via la méthode {@link #getSpace(String, String)} </li>
 * </ul><br/>
 * Cette classe contient également quelques méthodes dédiées à la configuration et au monitoring du site local, mais les propriétés d’auto-configuration de la 
 * plateforme font que ces méthodes ne seront utiles que dans de très rares cas très spécifiques. Vous pouvez donc les ignorer.
 * <br/>
 * Enfin, cette classe contient les définitions de toutes les constantes utilisés comme valeur d’argument, ou retourné en résultat d’appel, des différentes 
 * méthodes de la librairie.
 * */
public class Ubiquity { //Ubiquity
		
	private final static Map<Hub, UbiquityManager> managers = new HashMap<Hub, UbiquityManager>();	
	private synchronized static UbiquityManager getManager(Hub hub) {
		if(hub==null) hub = HubUdp.joinLocalHub();
		UbiquityManager mng = managers.get(hub);
		if(mng==null) { //instanciation du manager Ubiquity pour cet hub
			mng = new UbiquityManager(hub);
			managers.put(hub, mng);
		}
		return(mng);
	}
	
	/**
	 * Obtention du descripteur d'un space.<br/>
	 * <br/>
	 * Cette méthode permet d'obtenir le descripteur d'un space dont l'identité {nom, mot de passe} est spécifié en argument.<br/>
	 * Ce descripteur permettra d'ajouter des participants dans le space (méthode {@link Ubiquity.space.Space#newInteractor(String, byte, String)}
	 * et d'observer ce qui se passe au sein du space.
	 * @param name Le nom du space. Une valeur <b>null</b>, ou une chaine vide, peuvent être utilisées pour obtenir un space anonyme.<br/>
	 * La valeur spéciale <b>"#new"</b> peut aussi être utilisé pour créer un nouveau space avec un nom généré aléatoirement.
	 * @param password Le mot de passe du space. Une valeur <b>null</b>, ou une chaine vide, peuvent être utilisées pour le cas d'un space publique 
	 * (à accés libre).
	 * @return Le descripteur du space correspondant à l'identité du space spécifié en argument.
	 */
	public static Space getSpace(Hub hub, String name){
		UbiquityManager umanager = getManager(hub);
		return(umanager.getSpace(name, null));
	}
	public static Space getSpace(Hub hub, String name, String password){
		UbiquityManager umanager = getManager(hub);
		return(umanager.getSpace(name, password));
	}
	public static Space getLocalSpace(Spot spot, String name, String password){ // //space purement local virtuellement logé sur le spot "spot" (non distribué sur le réseau)
		return(new ImplSpace(spot, name, password));
	}
	//]
	
	private Ubiquity(){} //cette classe ne s'instancie pas
	
	//[Réponse générale d=0
	/**
	 * Valeur exprimant un spot ou un agent inconnu.<br/>
	 * <br/>
	 * Cette valeur est celle retournée quand on demande la valeur du statut d'un agent ou d'un spot qui n'existe pas.
	 */
	public final static byte NOT_EXIST=0;	//le spot ou l'agent est inconnu
	
	/**
	 * Constante pour qualifier un succés.<br/>
	 * <br/>
	 * Cette valeur exprime le succès d'une opération, ou la bonne prise en compte d'une requête.<br/>
	 */
	public final static byte OK=1;

	/**
	 * Constante pour qualifier une alerte.<br/>
	 * <br/>
	 * Cette valeur exprime généralement qu'un message signale un problème non critique.
	 */	
	public final static byte WARNING=2;
	
	/**
	 * Constante pour qualifier un niveau critique.<br/>
	 * <br/>
	 * Cette valeur exprime l'échec d'une oprétation, ou un message signalant un problème d'importance critique. 
	 */
	public final static byte ERROR=3;
	

	/**
	 * Valeur exprimant un état non actif.<br/>
	 * <br/>
	 * Cette valeur est celle retournée quand on essaie de faire des modifications sur les facettes d'un agent qui a été retiré du space.
	 * */	
	public final static byte NOT_ACTIF=4; 	//l'agent n'est pas actif
	
	/**
	 * Valeur exprimant un message vide.<br/>
	 */
	public final static byte EMPTY_MESS=5;
	
	/**
	 * Valeur de code d'un rapport de transmission de message indiquant que le message a bien été transmis.<br/>
	 * <br/>
	 * On obtient cette valeur comme code de transmission dans l'appel {@link Ubiquity.FullParticipantAdpator.FullBodyListener#sendingReport(Ubiquity.agent.AHandle, Ubiquity.agent.identity.RIdentity, String, byte)}
	 * quand le message a bien été transmis à son destinataire.
	 */
	public final static byte DELIVRED=6;	//le message à été placé dans la boite de réception de son destinataire
	public final static byte PERFORMED=7; 	//pour indiquer que le message a été consumé par son destinataire avec le status traité
	public final static byte IGNORED=8; 	//pour indiquer que le message a été consumé par son destinataire avec le status ignoré
	public final static byte TARGET_NOT_EXIST=9; //le destinataire du message n'existe plus
	
	//[Valeurs de statut des sites et des spots (d=10)
	/**
	 * Statut des spots : le spot est désactivé, il ne fait partie d'aucun space.
	 */		
	public final static byte DOWN=10;	
	/**
	 * Statut des spots : le spot est actif.
	 */	
	public final static byte UP=11;
	/**
	 * Statut des spots : le spot est marqué Ghost.<br/>
	 * <br/>
	 * Cet état traduit une latence réseau à destination de ce spot. Les messages qui doivent être transmis vers cette destination sont pour l'instant
	 * bufférisés en attendant que la latence se résorbe.<br/>
	 * Quand un spot se déconnecte brutalement du réseau, son statut bascule provisoirement sur cet état <b>GHOST</b> avant de passer à <b>DOWN</b>
	 */		
	public final static byte GHOST=12; 
	/**
	 * Statut des spots : statut transitoire lors de l'insertion du spot dans le space.<br/>
	 * <br/>
	 * Cet état transitoire traduit le fait que le spot n'a pas encore confirmé qu'il a pris connaissance de l'existence du spot local.
	 */		
	public final static byte SYNCHRONISATION=13;
	/**
	 * Statut des spots : statut transitoire lors de l'insertion du spot dans le space.<br/>
	 * <br/>
	 * Cet état transitoire traduit le fait que le spot local est en attente des informations nécessaires à l'initialisation du spot distant.
	 * 
	 */		
	public final static byte INITIALISATION=14;
	
	public final static byte UNKNOW=14;
	//]	
	
	//[Valeurs de statut des agents (d=20)
	//UNKNOW et NOT_EXIST et INITIALISATION (état transitoire avant disponibilité de l'agent).	
	/**
	 * Statut de l'agent : valeur indiquant que l'agent est supprimé du space.<br/>
	 */
	public final static byte REMOVED=20;
	
	/**
	 * Statut de l'agent : valeur indiquant que l'agent est actif. 
	 */
	public final static byte ACTIF=21;	
	//]
	
	//[Constante pour les activités d=30
	public final static byte NOT_RUN=40;
	public final static byte WAITING=41;
	public final static byte RUN=42;
	public final static byte CANCELED=43;
	public final static byte ONCE=44;
	public final static byte CHAIN=45;
	public final static byte REPEAT=46;
	public final static byte NONE=47;
	//]
}