package skuad.ubiquity.space;

import skuad.ubiquity.space.mailbox.*;
import skuad.ubiquity.space.state.*;
import skuad.ubiquity.space.skill.*;
import skuad.ubiquity.space.alias.*;
import skuad.ubiquity.space.activity.*;

/**
 * Interface des interacteurs.<br/>
 * <br/>
 * Un objet de ce type peut être obtenue : soit sur le point d'entrée SKUAD : {@link ubiquity.Ubiquity#newInteractor(String,String,String)}, 
 * soit sur le descripteur d'un space en particulier {@link Ubiquity.space.Space#newInteractor(String, byte, String)}.<br/>
 * <br/>
 * Un interacteur permet à une unité de code de manipuler un participant dans un space. Pour cela, l'interacteur donne accés
 * en écriture aux différentes facettes du participant :<ul>
 * <li> la méthode {@link #myIdentity()} : retourne le gestionnaire d'identité de l'agent (type {@link Ubiquity.agent.identity.WIdentity}).</li>
 * <li> la méthode {@link #myState()} : retourne le gestionnaire d'état de l'agent (type {@link Ubiquity.agent.state.WState}).</li>
 * <li> la méthode {@link #myFonctionnality()} : retourne le gestionnaire de compétence de l'agent (type {@link Ubiquity.agent.fonctionnality.WFonctionnality}).</li>
 * <li> la méthode {@link #activity()} : retourne le gestionnaire d'activité de l'agent (type {@link Ubiquity.agent.activity.Scheduler}).</li>   
 * </ul><br/>
*/
public interface Avatar {
	
	public Space getSpace(); //null si le space est protégé par un mot de passe, il faut alors utilisé getSpace(String)
	public Space getSpace(String password);
	
	/**
	 * Obtention du descripteur du participant associé à l'interacteur.<br/>
	 * <br/>
	 * @return le descripteur du participant associé à l'interacteur.
	 */
	public Participant participant();	
	
	/**
	 * Teste si l'interacteur est toujours actif, c'est à dire si le participant associé n'a pas été retiré du space.<br/>
	 * <br/>
	 * Cette méthode permet de tester rapidement si le participant est toujours présent dans le space.<br/>
	 * Elle équivaut à l'appel de la méthode {@link Ubiquity.space.Participant#exist()} sur le participant lié à l'interacteur (qu'on peut obtenir par 
	 * la méthode {@link #participant()}).
	 * 
	 * @return <b>true</b> si l'interacteur est actif, <b>false</b> sinon.
	 */
	public boolean exist();
		
	/**
	 * Supprime l'interacteur, et donc le participant qui lui est associé dans le space.<br/>
	 * <br/>
	 * Cette méthode permet de supprimer le participant associé à l'interacteur. A la suite de son déclenchement l'agent sera
	 * supprimé et son statut passera à la valeur {@link ubiquity.Ubiquity#REMOVED}.<br/>
	 * L'agent ne sera donc plus actif, l'ensemble de ses activités en cours seront annulées et il ne pourra exercer
	 * aucune action sur ses 4 facettes. 
	 * @return <ul>
	 * 	<li>{@link ubiquity.Ubiquity#OK} si tout est ok.</li>
	 * 	<li>{@link ubiquity.Ubiquity#NOT_ACTIF} si l'agent n'est pas actif (en particuier s'il a déjà été supprimé du space).</li>
	 * </ul>
	 */
	public byte remove();
	
	/**
	 * 
	 * @return
	 */
	public MailBox mailBox();
	
	//[Les 4 facettes : Alias/State/Skill/Activity
	/**
	 * Retourne le gestionnaire (objet permettant la modification) de la facette Alias de l'agent.<br/>
	 *  
	 * @return Le gestionnaire de la facette Alias de l'agent.
	 */
	public WAlias alias();	
	
	/**
	 * Retourne le gestionnaire (objet permettant la modification) de la facette State de l'agent.<br/>
	 *  
	 * @return Le gestionnaire de la facette State de l'agent.
	 */
	public WState state();
	
	/**
	 * Retourne le gestionnaire (objet permettant la modification) de la facette Fonctionnality de l'agent.<br/>
	 *  
	 * @return Le gestionnaire de la facette Fonctionnality de l'agent.
	 */
	public WSkill skill();
	
	/**
	 * Retourne le gestionnaire (objet permettant la modification) de la facette Activity de l'agent.<br/>
	 *  
	 * @return Le gestionnaire de la facette Activity de l'agent.
	 */	
	public WActivity activity();
	//]
}
