package skuad.ubiquity.space;

import skuad.ubiquity.space.skill.*;
import skuad.ubiquity.space.state.*;
import skuad.ubiquity.space.alias.*;
import skuad.ubiquity.space.activity.*;
import skuad.hubudp.*;

/**
 * Descripteur des participants à un space.<br/>
 * <br/>
 * Ce descripteur permet de consulter les informtations concernant un participant.
 */
public interface Participant {
	
	public Space getSpace(); //null si le space est protégé par un mot de passe, il faut alors utilisé getSpace(String)
	public Space getSpace(String password);
	
	/**
	 * Test si ce participant est toujours présent dans le space.<br/>
	 * Un participant est retiré d'un space dans deux situations :<ul>
	 * <li> soit parce que le spot du participant n'est plus dans le space (la machine correspondant à ce spot n'est plus accessible sur le réseau).</li>
	 * <li> soit parce que l'unité de code qui pilote ce participant l'a supprimé via un appel à la méthode {@link Ubiquity.Interactor#delete()}.</li>
	 * </ul>
	 * @return
	 * <ul>
	 * 	<li><b>true</b> si le participant est encore présent dans le space.</li>
	 * 	<li><b>false</b> si le participant n'est plus présent dans le space.</li>
	 * </ul>
	 */
	public boolean exist();
	
	public boolean ordoredBefore(Participant ia);
	public boolean ordoredAfter(Participant ia);
	
	/**
	 * Retourne l'identifiant local de l'agent sur son spot porteur.<br/>
	 * <br/>
	 * Cette valeur est celle attribuée à l'agent au moment de son enregistrement sur son spot local. C'est une valeur qui est unique sur un noeud donnée (deux agents situés
	 * sur un même spot ne peuvent avoir la même valeur d'identifiant local).<br/>
	 * 
	 * @return Valeur de la chaîne identifiant l'agent localement sur son spot.
	 */
	public String localId();	
 
	/**
	 * L'identifiant global du participant dans le Space.<br/>
	 * <br/>
	 * Cette chaine est produite de la façon suivante :  <i>id local de l'agent</i>@<i>id du global de l'hote qui porte le participant</i>.<br/>
	 * Ainsi, par construction, cette chaine permet de désigner un participant unique sur l'ensemble du space.
	 * 
	 * @return l'identifiant global du participant dans le Space.
	 */
	public String fullId();
	
	/**
	 * Test si ce participant est local ou distant.<br/>
	 * <br/>
	 * @return <ul>
	 * 	<li><b>true</b> si le participant est présent physiquement dans le contexte local.</li>
	 * 	<li><b>false</b> si le participant est piloté par un code situé dans un contexte distant.</li>
	 * </ul>
	 */
	public boolean isLocal();

	/**
	 * Renvoie le descripteur de l'hote sur lequel le participant se trouve.<br/>
	 * 
	 * @return le descripteur du spot qui porte l'agent.
	 */
	public Spot spot(); //spot sur lequel se trouve le participant		

	//[Lecture des 4 facettes : Alias/State/Skill/activity
	/**
	 * Retourne le descripteur en lecture de la facette fonctionnalité de l'agent.<br/>
	 * 
	 * @return le descripteur en lecture de la facette fonctionnalité de l'agent, ou <b>null</b> s'il s'agit d'un agent dont la visibilité ne permet son accés public depuis le site local.
	 */
	public RSkill getSkill();
	
	/**
	 * Retourne le descripteur en lecture de la facette state de l'agent.<br/>
	 * 
	 * @return le descripteur en lecture de la facette state de l'agent, ou <b>null</b> s'il s'agit d'un agent dont la visibilité ne permet son accés public depuis le site local.
	 */	
	public RState getState();
	
	/**
	 * Retourne le descripteur en lecture de la facette alias de l'agent.<br/>
	 * 
	 * @return le descripteur en lecture de la facette alias de l'agent, ou <b>null</b> s'il s'agit d'un agent dont la visibilité ne permet son accés public depuis le site local.
	 */	
	public RAlias getAlias();
	
	/**
	 * Retourne le descripteur en lecture de la facette activity de l'agent.<br/>
	 * 
	 * @return le descripteur en lecture de la facette activity de l'agent, ou <b>null</b> s'il s'agit d'un agent dont la visibilité ne permet son accés public depuis le site local.
	 */	
	public RActivity getActivity();
	//]
	
	/**
	 * Ajout d'un écouteur de suppression locale à cet agent.<br/>
	 * L'écouteur sera ici notifié que pour la suppression de cet agent, pour une écoute plus globale (portant sur tous les agents du space),
	 * il faut passer par l'objet monitor et utiliser la méthode {@link Ubiquity.Monitor#addAgentListener(AgentListener)}.<br/>
	 * <br/>
	 * <b>Attention :</b> cette méthode est sans effet si le niveaud de visibilité de l'agent le rend innaccessible depuis le site physique local.
	 * 
	 * @param sl Objet à notifier lorsque l'agent sera supprimé du space.
	 */
	public void addListener(ParticipantListener sl);
	
	/**
	 * Désenregistrement d'un écouteur de suppression locale à cet agent.<br/>
	 * 
	 * @param sl Objet spécifié sur l'appel {@link #addListener(AgentListener)} qu'il faut annuler ici.
	 */
	public void removeListener(ParticipantListener sl);	
}