package skuad.ubiquity.space;

import skuad.ubiquity.space.skill.*;
import skuad.ubiquity.space.state.*;
import skuad.ubiquity.space.alias.*;
import skuad.ubiquity.space.activity.*;

/**
 * Descripteur des spaces.<br/>
 * <br/>
 * Ce descripteur décrit un space et permet d'y ajouter un corps. Il peut s'obtenir de 2 manières : <ul>
 * <li>Soit via la classe centrale {@link ubiquity.Ubiquity} et sa méthode {@link ubiquity.Ubiquity#getSpace(String, String)}</li>
 * <li>Soit via le descripteur d'un spot : {@link Ubiquity.spot.Spot} et sa méthode {@link Ubiquity.spot.Spot#getSpace()}</li>
 * </ul>
 */
public interface Space {
	
	//[Information général sur le Space auquel ce spot est rataché
	/**
	 * Consultation du nom du space.
	 * @return le nom du space.
	 */
	public String nameSpace();
	
	/**
	 * Indique si le space est protégé par un mot de passe.<br/>
	 * Notez qu'il n'y a aucun moyen de consulter la valeur du mot de passe. Cette méthode ne sert qu'a déterminé si le space et publique 
	 * (sans mot de passe) ou privé (avec mot de passe)
	 * @return <b>true</b> si le space est protégé par un mot de passe, <b>false</b> sinon.
	 */
	public boolean havePassword(); //l'espace dispose d'un mot de passe ou non
	//]	
		
	/**
	 * Crée un nouveau participant dans le space.<br/>
	 * <br/>
	 * Cette méthode crée un nouveau participant dans le space et retourne l'avatar permettant de controler le participant créé.<br/>
	 *  
	 * @param local_id identifiant local à attribuer au participant. On peut utiliser la valeur <i>null</i>, <b>""</b> ou <b>"#new"</b> pour que l'identifiant 
	 * soit généré automatiquement. Le caractère @ est interdit, et l'id choisi ne peut pas commencer par les caractères #, !, & et *
	 * 
	 * @return <ul> 
	 * 		<li><i>null</i> si l'identifiant mentionné est déjà utilisé ou contient un symbole interdit.</li>	
	 * 		<li>L'interacteur permettant de contrôler le nouveau participant créé.</li>
	 * 	</ul>
	 * @see  Ubiquity.agent.Interactor
	 */
	public Avatar newAvatar();
	public Avatar newAvatar(String local_id);
	
	//[lecture des informations sur les agents
	/**
	 * Obtention du descripteur d'identité d'un agent.<br/>
	 * <br/>
	 * La chaine d'identifiant spécifant l'agent cible peut être de 4 formes :<ul>
	 * 	<li>un identifiant local (localId) : une simple chaine sans symbole <b>@</b> et qui ne commence pas par <b>#</b> ou <b>&amp;</b>. Cette forme exprime l'identifiant local de l'agent. L'agent cible ne sera
	 *  donc recherché que sur le spot local</li>
	 *  <li>un identifiant complet (fullId): une chaine de la forme <b>localId@siteId</b>. Cette forme exprime l'identifiant global de l'agent. Ce dernier sera donc recherché sur le spot correspondant
	 *  à l'identifiant de site spécifié.</li>
	 *  <li>un ciblage indirecte fort : une chaine de la forme <b>#alias</b>. Cette forme exprime un agent qui a le focus sur un alias. L'agent cible sera donc
	 *  recherché sur l'ensemble des agents du space, et c'est celui qui aura le focus sur l'alias indiqué qui sera l'agent retourné.</li>
	 *  <li>un ciblage indirecte faible : une chaine de la forme <b>&alias</b>. Cette forme exprime un agent qui possède un alias. L'agent cible sera donc
	 *  recherché sur l'ensemble des agents du space, et l'un de ceux ayant pour alias la valeur indiqué sera l'agent retourné.</li>
	 * </ul>
	 * 
	 * @param id identifiant de l'agent ciblé.
	 * @return Le descripteur d'identité de l'agent, ou <b>null</b> si aucun agent ne correspont à l'identifiant indiqué.
	 */
	public Participant getParticipant(String id);
	//]
	
	//[ecoute des modifications subit par les agents du space.
	/**
	 * Enregistre un écouteur d'ajout/suppression d'agents du space.
	 * 
	 * @param ial Objet à notifier quand un agent apparait, disparait.
	 */
	public void addParticipantListener(ParticipantListener ial);
	
	/**
	 * Désenregistre un écouteur  d'ajout/suppression d'agents du space
	 * @param ial Objet écouteur à désenregistrer.
	 * 
	 * @return <i>true</i> si l'objet a bien été désenregistré, ou <i>false</i> si cet objet n'est pas un écouteur connu.
	 */
	public void removeParticipantListener(ParticipantListener ial);
	
	/**
	 * Enregistre un écouteur de facette Identity des agents du space.
	 * 
	 * @param ial Objet à notifier quand un agent apparait, disparait ou modifie sa facette Identity.
	 */
	public void addAliasListener(AliasListener ial);
	
	/**
	 * Désenregistre un écouteur de facette Identity des agents du space.
	 * @param ial Objet écouteur à désenregistrer.
	 * 
	 * @return <i>true</i> si l'objet a bien été désenregistré, ou <i>false</i> si cet objet n'est pas un écouteur connu.
	 */
	public void removeAliasListener(AliasListener ial);
	
	/**
	 * Enregistre un écouteur de facette Fonctionnality des agents du space.
	 * 
	 * @param fl Objet à notifier quand un agent modifie sa facette Fonctionnality.
	 */
	public void addSkillListener(SkillListener fl);
	/**
	 * Désenregistre un écouteur de facette Fonctionnality des agents du space.
	 * @param fl Objet écouteur à désenregistrer.
	 * 
	 * @return <i>true</i> si l'objet a bien été désenregistré, ou <i>false</i> si cet objet n'est pas un écouteur connu.
	 */
	public void removeSkillListener(SkillListener fl);
	
	/**
	 * Enregistre un écouteur de facette State des agents du space.
	 * 
	 * @param fl Objet à notifier quand un agent modifie sa facette State.
	 */
	public void addStateListener(StateListener fl);
	/**
	 * Désenregistre un écouteur de facette State des agents du space.
	 * @param fl Objet écouteur à désenregistrer.
	 * 
	 * @return <i>true</i> si l'objet a bien été désenregistré, ou <i>false</i> si cet objet n'est pas un écouteur connu.
	 */
	public void removeStateListener(StateListener fl);
	
	/**
	 * Enregistre un écouteur de facette Activity des agents du space.
	 * 
	 * @param ial Objet à notifier quand un agent modifie sa facette Activity.
	 */
	public void addActivityListener(ActivityListener ial);
	
	/**
	 * Désenregistre un écouteur de facette Activity des agents du space.
	 * @param ial Objet écouteur à désenregistrer.
	 * 
	 * @return <i>true</i> si l'objet a bien été désenregistré, ou <i>false</i> si cet objet n'est pas un écouteur connu.
	 */
	public void removeActivityListener(ActivityListener ial);
	//]
}
