package skuad.ubiquity.space.alias;

import skuad.ubiquity.space.*;

/**
 * Interface d'écoute des facettes Identity des agents.<br/>
 * <br/>
 * Cette interface est celle que doivent implémenter les objets que l'on enregistre sur l'écoute des changements subis par les facettes Identity des agents. Par cette 
 * écoute on reçoit donc une notification quand l'agent
 * ajoute, retire ou modifie la priorité d'un de ses alias dans sa facette Identity. On est aussi informé des changements
 * de focus (perte/acquisition) sur les alias des agents (voir {@link #aliasPriorityChange(RIdentity, String, short, short)}).<br/>
 * NB : on dit qu'un agent a le focus sur un alias quand l'agent devient celui qui a la plus haute priorité pour cet alias. Il est alors celui qui recevra les messages émis
 * avec une désignation du destinataire de la forme "#<i>alias</i> (voir la section d'envoi de de message de l'interface {@link Ubiquity.agent.Interactor}).<br/>
 * <br/>
 * D'autre part, puisque l'identité primaire d'un agent (son identifiant local reçu au moment de son insertion dans le space) est aussi l'une des caractéristiques de sa
 * facette identité : avec l'écoute de cette facette on est aussi informé de l'apparition et la disparition des agents dans le space.<br/>
 * <br/>
 * Il y a deux niveaux d'écoute possibles :<br/>
 *  - l'écoute locale, centrée sur un agent spécifique : cette écoute se met en oeuvre par l'enregistrement de l'écouteur directement sur le descripteur 
 * {@link Ubiquity.agent.identity.RIdentity} de l'agent ciblé, via la méthode {@link Ubiquity.agent.identity.RIdentity#addListener(IdentityListener)}.<br/>
 *  - l'écoute globale au space : à ce niveau on écoute les évolutions d'identité de tous les agents, et pas seulement d'un seul d'entre eux comme précédemment. Pour se faire
 * l'écouteur doit cette fois être enregistré sur le monitor via la méthode {@link Ubiquity.Monitor#addIdentityAgentListener(IdentityListener)}.<br/>
 * 
 * @see Ubiquity.agent.identity.RIdentity
 * @see Ubiquity.Monitor
 */
public interface AliasListener {	
	/**
	 * Notifie l'ajout d'un alias sur un agent.<br/>
	 * @param identity descripteur d'indentité de l'agent concerné
	 * @param newAlias nom du nouvel alias acquis par l'agent
	 * @param priority valeur de la priorité fixée sur cet alias
	 */
	public void aliasAdded(Participant identity, String newAlias, int priority);
	
	/**
	 * Notifie la suppression d'un alias sur un agent.<br/>
	 * <br/>
	 * Notez qu'un alias est aussi supprimé quand sa priorité est fixé à la valeur 0.
	 * @param identity desripteur d'indentité de l'agent concerné.
	 * @param lostAlias nom de l'alias perdu par l'agent.
	 * @param oldPriority ancienne valeur de priorité de cet alias avant la suppression (ou sa mise à 0).
	 */
	public void aliasRemoved(Participant identity, String lostAlias, int oldPriority);
	
	/**
	 * Notifie le changement de valeur de priorité subie par l'alias d'un agent.<br/>
	 * 
	 * @param identity descripteur d'indentité de l'agent concerné
	 * @param alias nom de l'alias concerné
	 * @param oldPriority ancienne valeur de priorité de cet alias
	 * @param newPriority nouvelle valeur de priorité de cet alias
	 */
	public void aliasPriorityChange(Participant identity, String alias, int oldPriority, int newPriority);
	
	/**
	 * Notifie le changement de focus pour un alias.<br/>
	 * <br/>
	 * A un instant <b>t</b> un seul et unique agent ne peut disposer du focus sur un alias dans l'ensemble d'un space. Cet agent prioritaire est déterminé comme tel en fonction de 3 critères : la 
	 * valeur de priorité associée à son alias, la valeur de la chaîne identifiant de son spot (unique pour un space donnée), son identifiant local (unique sur un spot donnée).<br/>
	 * La règle interne est la suivante : un agent est prioritaire sur un alias <a>a</a> si cet agent est celui qui a la priorité la plus forte pour cet alias, et en cas d'égalité de priorité, celui
	 * qui se trouve sur le spot dont la chaine identitiante a le plus petit ordre lexicographique, et en cas où les agents candidats sont sur le même spot, celui qui a la chaine identifiante locale
	 * de plus petit ordre lexicographique.<br/>
	 * Ainsi, par définition cette règle est déterministe et conduit à l'élection d'un unique agent sur le focus d'un alias. Et un changement de focus pour occurer de diverses façons : soit lors d'un changement
	 * de priorité exprimé par un agent sur son alias, soit suite à l'ajout ou la suppression d'un agent, soit suite à l'insertion ou le retrait d'un spot.<br/>
	 * <br/>
	 * Quand il s'agit d'un focus lié à un nouvel alias, ou quand la dernière occurence d'un alias disparait du space : une seule notification de ce type est déclenchée.<br/>
	 * Par contre, quand il s'agit d'une passation de pouvoir (un agent perd le focus au profit d'un autre agent), alors deux notifications sont produites : l'une pour indiquer la perte du focus
	 * par le premier agent, et l'autre son acquision par le second.<br/>
	 * 
	 * @param identity descripteur d'identité de l'agent concerné par cette notification de changement de focus
	 * @param alias nom de l'alias concerné
	 * @param oldFocus ancien état du focus (valeur boolénne indiquant si oui ou non l'agent disposé préalablement du focus)
	 * @param newFocus nouvel état du focus  (valeur boolénne indiquant si oui ou non l'agent dispose dorénavant du focus)
	 */
	public void aliasFocusChange(Participant identity, String alias, boolean oldFocus, boolean newFocus);
	
}