package skuad.ubiquity.space.mailbox;

import skuad.ubiquity.space.*;
import skuad.util.value.Value;

/**
 * * L'handle permet aussi d'envoyer des messages à destination des autres agents. Un message peut être envoyé de 2 façons : avec ou sans notification de l'état 
 * de transmision. Et d'autre part il existe 2 façons de désigner le destinatire : par une simple chaine de caractères, ou par sa facette identifiante 
 * {@link Ubiquity.agent.identity.RIdentity}.<br>
 * Cela fait donc en tout 4 formes possibles pour les méthodes d'envoi de message :
 * - {@link #send(RIdentity, String, Value)} : c'est la forme la plus simple, le destinataire est désigné par son identité et aucune notification
 * 	n'est exigée.<br>
 * - {@link #send(RIdentity, String, Value, String)} : toujours la même forme de désignation de la cible, via l'identité de l'agent distant, mais cette fois avec
 * l'indication d'une clé de notification de transmission. Cette dernière va donner lieu à un notification de rapport de transmission, par l'invocation de la méthode 
 * {@link Ubiquity.agent.FullParticipantAdpator#sendingReport(AHandle, RIdentity, String, byte)} afin d'indiquer à l'agent si la transmission du message a bien pu être réalisée.<br/>
 * - {@link #send(String, String, Value)} : cette forme de transmission est un peu plus ouverte dans le sens où le destinataire est désigné ici d'une façon plus souple,
 * par une chaine de caractères appelée <b>chaine de ciblage</b>. C'est la syntaxe de cette chaine qui va déterminer le, ou les, destinataires du message. Cette syntaxe est 
 * décrite ci-dessous.<br/>
 * - {@link #send(String, String, Value, String)} : dans cette dernière forme la cible est désignée par une chaine de ciblage et une clé de notification de transmission
 * est spécifiée. Cela va donc conduire au déclenchement d'une notification de rapport de transmission pour chacun des agents qui auront été interprétés comme destinataires
 * par la chaine de ciblage.<br/>  
 * <br/>
 * L'utilisation de la chaine de ciblage permet de désigner plus directement le ou les agents qui devront recevoir le message. La syntaxe de cette chaine peut prendre
 * 4 formes :<br/>
 * - une chaine qui ne commence pas par # ou * et qui ne contient pas de symbole @ : écrite de cette façon la chaine de ciblage représente la chaine d'identifiant
 * local d'un agent (celle que l'agent obtient au moment de son incription sur le spot local). Le destinataire du message sera donc l'agent local (sur le même spot
 * que l'expéditeur) qui possède la chaine d'identifiant ayant la même valeur que celle de la chaine de ciblage.<br/>
 * - une chaine qui ne commence pas par # ou * et qui contient le symbole @ : la chaine est alors interprété sous la forme id_local@id_spot et correspond à la chaine
 * d'identifiant complète d'un agent. La position du premier symbole @ va être 
 * utilisée pour couper la chaine en 2 afin d'isoler l'identifiant local de l'agent (la première partie de la chaine) et l'identifiant du spot sur lequel il se trouve (la
 * deuxième partie de la chaine). Un unique agent est donc désigné par cette forme de ciblage, et ce sera lui (s'il existe) qui recevra le message.<br/>
 * - une chaine qui commence par le symbole # : dans cette forme la désignation s'effectue sur la base des alias et non plus des chaines d'identifiants. La chaine de ciblage
 * est interprétée comme correspondant à #alias. Le texte qui suit le symbole # est donc isolé puis utilisé pour identifier l'agent cible qui sera l'agent du space
 * possédant le focus sur l'alias (voir {@link Ubiquity.agent.identity.WIdentity#setAlias(String, short)}).<br/>
 * - une chaine qui commence par le symbole * : c'est un variante de la forme précédente. Ici la désignation s'effectue également sur la base de l'alias isolé après 
 * le symbole *, mais cette fois on ne cible pas l'unique agent qui a le focus sur l'alias mais <b>tous</b> les agents qui possède cet alias. Ainsi, cette forme
 * de ciblage peut conduire à ce que plusieurs destinataires soient identifiés et ils recevront tous le message.<br/>
 * <br/>
 */
public interface MailBox {
	/**
	 * Descripteur du participant auquel appartient cette boite au lettre.
	 * @return descripteur du participant auquel appartient cette boite au lettre.
	 */
	public Participant participant();
	
	public boolean haveMessage();
	public int nbMessage();
	public Message get(int num);
	public Message pop();
	
	public void addListener(MessageListener sl);
	public void removeListener(MessageListener sl);	

	public void addReporter(SendReportListener sl);
	public void removeReporter(SendReportListener sl);	

	//[Envoie de messages
	/**
	 * Envoi d'un message via une désignation de la cible par identifiant.<br/>
	 * <br/>
	 * Voir l'aide générale de cette classe pour plus de détails.
	 * 
	 * @param target Identifiant de l'agent destinataire du message.
	 * @param skill_context Contexte de compétence relatif à ce message.
	 * @param message Contenu du message
	 * @return <ul>
	 * 	<li>{@link Ubiquity.SKUAD#OK} si tout est ok.</li>
	 *  <li>{@link Ubiquity.SKUAD#NOT_ACTIF} si l'agent expéditeur est supprimé du space.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_EXIST}  si l'agent ciblé n'est plus actif.</li> 
	 *  <li>{@link Ubiquity.SKUAD#EMPTY_MESSAGE} si le mssage spécié est vide.</li> 
	 * </ul>
	 */
	public byte send(Participant target, String skill_context, Value message);

	/**
	 * Envoi d'un message via une désignation de la cible par identifiant et un demande de rapport de transmission.<br/>
	 * <br/>
	 * Voir l'aide générale de cette classe pour plus de détails.
	 * 
	 * @param target Identifiant de l'agent destinataire du message.
	 * @param skill_context Contexte de compétence relatif à ce message.
	 * @param message Contenu du message
	 * @param idreport Clé d'identification pour le rapport de transmission.
	 * @return <ul>
	 * 	<li>{@link Ubiquity.SKUAD#OK} si tout est ok.</li>
	 *  <li>{@link Ubiquity.SKUAD#NOT_ACTIF} si l'agent expéditeur est supprimé du space.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_EXIST}  si l'agent ciblé n'est plus actif.</li> 
	 *  <li>{@link Ubiquity.SKUAD#EMPTY_MESSAGE} si le mssage spécié est vide.</li> 
	 * </ul>
	 * 
	 * @see Ubiquity.agent.FullParticipantAdpator#sendingReport(AHandle, RIdentity, String, byte)
	 */
	public byte send(Participant target, String skill_context, Value message, String idreport);
	
	/**
	 * Envoi d'un message via une désignation de la cible par une chaine de ciblage.<br/>
	 * <br/>
	 * Voir l'aide générale de cette classe pour plus de détails.
	 * 
	 * @param target Chaine de ciblage indiquant le ou les destinataires du message.
	 * @param skill_context Contexte de compétence relatif à ce message.
	 * @param message Contenu du message
	 * @return <ul>
	 * 	<li>{@link Ubiquity.SKUAD#OK} si tout est ok.</li>
	 *  <li>{@link Ubiquity.SKUAD#NOT_ACTIF} si l'agent expéditeur est supprimé du space.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_EXIST} si la chaine de ciblage ne désigne aucun agent actif.</li> 
	 *  <li>{@link Ubiquity.SKUAD#EMPTY_MESSAGE} si le mssage spécié est vide.</li> 
	 * </ul>
	 */	
	public byte send(String target, String skill_context, Value message);
	
	/**
	 * Envoi d'un message via une désignation de la cible par une chaine de ciblage et une demande de rapport de transmission.<br/>
	 * <br/>
	 * Voir l'aide générale de cette classe pour plus de détails.
	 * 
	 * @param target Chaine de ciblage indiquant le ou les destinataires du message.
	 * @param skill_context Contexte de compétence relatif à ce message.
	 * @param message Contenu du message
	 * @param idreport Clé d'identification pour le rapport de transmission.
	 * @return <ul>
	 * 	<li>{@link Ubiquity.SKUAD#OK} si tout est ok.</li>
	 *  <li>{@link Ubiquity.SKUAD#NOT_ACTIF} si l'agent expéditeur est supprimé du space.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_EXIST} si la chaine de ciblage ne désigne aucun agent actif.</li> 
	 *  <li>{@link Ubiquity.SKUAD#EMPTY_MESSAGE} si le mssage spécié est vide.</li> 
	 * </ul>
	 * 
	 * @see Ubiquity.agent.FullParticipantAdpator#sendingReport(AHandle, RIdentity, String, byte)
	 */
	public byte send(String target, String skill_context, Value message, String idreport);
	//]
}
