/**
 * Cette classe spécifie le UDA-LLP : "UDA - Low Level Protocol" afin de constituer une base d'intéropérabilité 
 * commune avec les implémentations d'UDA réalisées dans les différents langages de programmation.
 * 
 * Les valeurs des constantes sont des entiers positifs sur 1 à 4 octets.
 * Chaque ensemble (Set1, Set2, ...) contiennent des valeurs de constante sans redondance (pas 2 constantes de même valeur)
 
 Pour un codage/décodage binaires standardisés des interactions Device<->Monitor il faut utiliser les interfaces DataReader et UdaWriter ainsi que les méthodes UDA.decode(...) // UDA.encodeXXX(...)
 ou la méthode Device.perform(...) sur les appareils ayant des interactions Device<->Monitor synchronisées (comme les ArduinoLike)
 
 * Légende de l'encodage des données préconisé :
 
 		LABEL			-> le LABEL désigne une constante UDA
 		[label] 		-> un octet pour encoder l'élément désigné par label  
 		{X: label} 		-> X octets pour encoder l'élément désigné par label, X peut être : un nombre, label spécifié en amont, ou le symbole ? pour indiquer une valeur variable
 		(name: label)	-> label encodé selon le format de nom name
 
 * Syntaxe de l'encodage des données préconisé : 
 		- octet : 	[label]
 		- entier : 	(int: label) = [type_int]{ ? : valeur} , type_int peut être : ZERO (?=0), ONE (?=1), BYTE (?=1), NBYTE (?=1), SHORT (?=2), NSHORT (?=2), WORD (?=4), NWORD (?=4)
 																						-> sens des octets : poids fort vers poids faible.
 		- décimal : (float: label) = { 4 : valeur }
 		- tableau d'octet: (tab: label) = (int: size){size : contenu} 
 		- chaine de caractère: (string: label) = (int: size){size : contenu}
  		- valeur quelconque typée : (val : valeur) = [type]{ ? : valeur }, type peut être un entier (ZERO, ONE, BYTE, NBYTE, SHORT, NSHORT, WORD, NWORD) ou : FLOAT, TBYTES, STRING
 */
package skuad.uda;

import skuad.util.databytes.*;

public class UDA {
	
	public final static String SOURCE_VERSION="0.0.10";
	
	public final static int H_VERSION = 1, L_VERSION = 0;	//numéro de version du protocole protocole UDA : 0.1  (X.Y, X et Y chacun sur 1 byte : X -> Major version, Y -> Minor version)
			//NB: 2 versions du protocole UDA sont concidéré comme compatible si et seulement si leur H_VERSION sont identique

	//Commun à tous les Set
	public final static int NOT_EXIST = 0;		
	public final static int UNKNOW = 1;
	
//[Set 1 : Code de commande {1 octet}
	//Code spéciaux (2 à 9)
	public final static int VERSION = 2; 	//-> indication de version du protocole	
	public final static int CLOSE = 9;    	//-> fin d'échange
	//Code commande non utilisé (10 à 19)   //-> réservé à usage libre (non utilisé, et aucune signification, au niveau du protocole UDA lui même)
		
	//Code des méthodes Device (20 à 39) :
		//action
	public final static int LOG = 20; 		//-> log(...)
	public final static int MOVE = 21; 		//-> move(...)
	public final static int ACTION = 22; 	//-> writeActuator(...)
		//lecture
	public final static int GET_PARAMS = 23; 			//-> getParams(Monitor mon);
	public final static int GET_MOVE_CAPACITY = 24;		//-> getMoveCapacity(Monitor mon);
	public final static int GET_SENSOR_PARAMS = 25; 	//-> getSensorParam(Monitor mon, int sensor_id);
	public final static int GET_ACTUATOR_PARAMS = 26;	//-> getActuatorParam(Monitor mon, int actuator_id);
	public final static int READ_SENSOR = 27; 			//-> readSensor(Monitor mon, int sensor_id);
	public final static int GET_STATE = 28; 			//-> getState(Monitor mon);
	
	//Code des méthodes Monitor (40 à 69) :
		//Réponse aux requêtes
	public final static int PARAMS = 40; 			//-> params(...)
	public final static int MOVE_CAPACITY = 41;		//-> moveCapacity(...)
	public final static int SENSOR_PARAMS = 42; 	//-> sensorParam(...)
	public final static int ACTUATOR_PARAMS = 43;	//-> actuatorParam(...)
	public final static int STATE = 47; 			//-> state(...)
	
		//Ecoute
	public final static int SENSOR 	= 44; 			//-> sensorValue(...)
	public final static int MOVED2D = 45; 			//-> moved2D(...)
	public final static int MOVED3D = 46; 			//-> moved3D(...)
	//déjà défini : public final static int LOG = 20; 	//-> log(...)

	//Code des commande de mouvement (0 + 70 à 99) :
	public final static int STOP  		=  0;  
	public final static int FORWARD  	=  70;       //avant
	public final static int BACK     	=  71;       //arrière
	public final static int TURN_RIGHT	=  72;       //rotation à droite
	public final static int TURN_LEFT 	=  73; 		 //rotation à gauche
//]
	
//[Set 3 : Code des types de mouvement (movement_type) {1 octet}
	public final static int NO_MOVE = 2;   	//-> Device dépourvu de capacité de mouvement
	public final static int MOVE2D = 3;   	//-> Device capable de se déplacer dans un espace à 2 dimension
//]
	
//[Set 4 : Code des types de Device {int32}
	public final static int LOGGER = 2;		//-> Device de log  
		//#log
	public final static int LIGHT = 3;		//-> Device actuator de type Lumière
		//actuator 0 : bool state	-->	 0 pour éteindre, 1 pour allumer 
	public final static int BUTTON = 4;		//-> Device sensor de type Bouton
		//sensor 0 : bool state	(OBS) -->	 0 / 1 
	public final static int TEMPERATURE = 5;	//-> Device sensor de type Bouton
		//sensor 0 : float temperatue	--> temperature en °C
	public final static int GPS = 6;	//-> Device sensor de type GPS : valeur de type byte[] dont le 1er élément donne le code du format d'encodage
		//sensor 0 : byte[] coords	--> coordonnée dans le format spécifié par coords[0]
	public final static int ROVER = 100;	//-> Device rover
	public final static int MOBILE = 101;	//-> Device mobile
		//#move
		//sensor 0 : float speed	--> vitesse du rover en cm/s
		//sensor 1 : float angle	--> angle du rover angle en radiant
//]
	 
//[Set 5 : Code des types d'Actuator {int32}
	public final static int ALIGHT = 2;		//-> lumière
	public final static int ASELECTOR = 3;	//-> sélecteur
	public final static int ASPEED = 4;
	public final static int AANGLE = 5;
//]	
	
//[Set 6 : Code des types de Sensor {int32}
	public final static int SBUTTON = 2;		//-> bouton
	public final static int STEMPERATURE = 3;
	public final static int SSPEED = 4;
	public final static int SANGLE = 5;
	public final static int SGPS = 6;
	public final static int SVISION = 7;
//]	
	
//[Set 7 : Code des erreurs {1 octets}
	public final static int OK = 2;			//-> tout est OK, pas d'erreur
	public final static int BAD_TYPE = 3;	//-> erreur de typage
	public final static int BAD_SYNTAX = 4;	//-> erreur de syntaxe
	public final static int CMD_UNKNOW = 5;	//-> commande inconnu
	public final static int RETRY = 6;		//-> on doit essayer de réappliquer le traitement
	public final static int NOT_POSSIBLE = 7;	//-> opération impossible
//]
		
//[Set 7 : Format d'encodage TBYTES {1 octets}
	public final static int GPS_INT_FORMAT 		= 10;
	public final static int GPS_FLOAT_FORMAT 	= 11;
//]			

	//[Fonctions utiles
	public static String labelMove(int cmd){ //retourne le label du mouvement correspondant à la commande de mouvement indiquée, ou "-unknow-" si la commande n'est pas connue
		switch(cmd){
		case STOP : return("STOP");
		case FORWARD : return("FORWARD");
		case BACK  : return("BACK");
		case TURN_RIGHT : return("TURN_RIGHT");
		case TURN_LEFT : return("TURN_LEFT");
		default :  return("-unknow-");
		}
	}
	public static boolean isCompatibleVersion(int byte_hight, int byte_low){
		return(H_VERSION==byte_hight);
	}
	//]
	
	//[encoder/decoder Uda
	public static boolean needMonitor(int cmd){ //true ssi le traitement de la commande cmd nécessite d'invoquer un Monitor
		switch(cmd){
		case GET_STATE: case GET_PARAMS: case  GET_MOVE_CAPACITY: case  GET_SENSOR_PARAMS: case  GET_ACTUATOR_PARAMS: case READ_SENSOR: return(true);
		}
		return(false);
	}
		//décodeur sur Device
	public static int decodeDevice(Device device, int cmd, IReader reader, Monitor mon){ //traite la commande encoder dans l'objet reader et l'applique au device, si un moniteur est nécessaire et si mon!=null il sera utilisé
		//retourne UDA.OK si la commande a été traité avec succés, ou l'un des codes d'erreur des constantes UDA du set 7.
		int actuator_id, sensor_id;
		switch(cmd){
		case MOVE : //[MOVE][cmd_move](int: power)
			int cmd_move = reader.readByte();
			int power = reader.readInt();
			device.move(cmd_move, power);
			return(OK);
		case ACTION : //[ACTION](int: actuator_id)(val: value)
			actuator_id = reader.readInt();
   			int val_type = reader.readByte();
//System.out.println("ACTION !!! val_type = "+val_type);   
   			if(val_type<DataBytes.TINY_INT) { device.action(actuator_id, val_type); return(OK); }
   			switch(val_type){
   			case DataBytes.NBYTE: case DataBytes.BYTE: case DataBytes.NSHORT: case DataBytes.SHORT: case DataBytes.NWORD: case DataBytes.WORD:
   				int iv = reader.readInt(val_type);
//   				System.out.println("ACTION !!! iv = "+iv);  				
   				device.action(actuator_id, iv);
   				return(OK);	
   			case DataBytes.FLOAT:
   				float fv = reader.readFloat();
   				device.action(actuator_id, fv);
   				return(OK);	
   			case DataBytes.TBYTES:
   				byte[] tv = reader.readTbytes();
   				device.action(actuator_id, tv);
   				return(OK);	  				
   			}
 System.out.println("ACTION !!! --> bad type"); 			
   			return(BAD_TYPE);				
		case GET_STATE: //*GET_STATE*
			if(mon!=null) device.getState(mon);
			return(OK);
		case GET_PARAMS: //*GET_PARAMS*
			if(mon!=null) device.getParams(mon);
			return(OK);
		case GET_MOVE_CAPACITY: //*GET_MOVE_CAPACITY*
			if(mon!=null) device.getMoveCapacity(mon);
			return(OK);
		case GET_SENSOR_PARAMS: //[GET_SENSOR_PARAMS](int: sensor_id)
			sensor_id = reader.readInt();
			if(mon!=null) device.getSensorParam(mon, sensor_id);
			return(OK);
		case GET_ACTUATOR_PARAMS: //*GET_ACTUATOR_PARAMS*(int: actuator_id)
			actuator_id = reader.readInt();
			if(mon!=null) device.getActuatorParam(mon, actuator_id);
			return(OK);
		case READ_SENSOR: //[READ_SENSOR](int: sensor_id)
			sensor_id = reader.readInt();
			if(mon!=null) device.readSensor(mon, sensor_id);
			return(OK);
		case LOG : //[LOG](string : texte)
			String mess = reader.readString();
			if(mess==null)  return(BAD_SYNTAX); //valeur de type inconnu
			device.log(mess);
			return(OK);
		}
		//commande inconnu
		return(CMD_UNKNOW);
	}
		//décodeur sur Moniteur
	public static int decodeMonitor(Monitor monitor, int cmd, IReader reader, Device device_source){ //traite la commande encoder dans l'objet reader et l'applique au monitor, l'instance Device source sera la valeur device_source spécifié
		//retourne UDA.OK si la commande a été traité avec succés, ou l'un des codes d'erreur des constantes UDA du set 7.
		String name; int type_value; float dx, dy; int sensor_id;
		switch(cmd){
		case UDA.STATE:			//[STATE][ready][config_changed]
 			boolean ready = reader.readBool();
   			boolean config_changed = reader.readBool();
   			monitor.state(device_source, ready, config_changed);
   			return(OK);
		case UDA.PARAMS:		//[PARAMS](string : name)(int : type_device)(int: nb_sensor)(int: nb_actuator)[can_log]
   			name = reader.readString();
   			int type_device = reader.readInt();
   			int nb_sensor = reader.readInt();
   			int nb_actuator = reader.readInt();
   			boolean can_log = reader.readBool();
   			monitor.params(device_source, name, type_device, nb_sensor, nb_actuator, can_log);
   			return(OK);
		case UDA.MOVE_CAPACITY:		//[MOVE_CAPACITY][movement_type](int: power_min)(int: power_moy)(int: power_max)
   			int movement_type = reader.readByte();
   			int power_min = reader.readInt();
   			int power_moy = reader.readInt();
   			int power_max = reader.readInt();
   			monitor.moveCapacity(device_source, movement_type, power_min, power_moy, power_max);
   			return(OK);
		case UDA.SENSOR_PARAMS:		//[SENSOR_PARAMS](int : sensor_id)(string : name)(int : sensor_type)[type_value][observable]
   			sensor_id = reader.readInt();
   			name = reader.readString();
   			int sensor_type = reader.readInt();
   			type_value = reader.readByte();
   			boolean observable = reader.readBool();
   			monitor.sensorParam(device_source, sensor_id, name, sensor_type, type_value, observable);
   			return(OK);		
		case UDA.ACTUATOR_PARAMS:	//[ACTUATOR_PARAMS](int : actuator_id)(string : name)(int : actuator_type)[type_value]
   			int actuator_id = reader.readInt();
   			name = reader.readString();
   			int actuator_type = reader.readInt();
   			type_value = reader.readByte();
   			monitor.actuatorParam(device_source, actuator_id, name, actuator_type, type_value);
   			return(OK);	
		case UDA.SENSOR:	//[SENSOR](int: sensor_id)(val: value)
   			sensor_id = reader.readInt();
   			int val_type = reader.readByte();
   			if(val_type < DataBytes.TINY_INT) { monitor.sensorValue(device_source, sensor_id, val_type); return(OK); }
   			switch(val_type){
   			case DataBytes.NBYTE: case DataBytes.BYTE: case DataBytes.NSHORT: case DataBytes.SHORT: case DataBytes.NWORD: case DataBytes.WORD:
   				int iv = reader.readInt(val_type);
   				monitor.sensorValue(device_source, sensor_id, iv);
   				return(OK);	
   			case DataBytes.FLOAT:
   				float fv = reader.readFloat();
   				monitor.sensorValue(device_source, sensor_id, fv);
   				return(OK);	
   			case DataBytes.TBYTES:
   				byte[] tv = reader.readTbytes();
   				monitor.sensorValue(device_source, sensor_id, tv);
   				return(OK);	  				
   			}
   			return(BAD_TYPE);	
		case UDA.LOG:		//[LOG](string : texte)
   			String text = reader.readString();
   			monitor.log(device_source, text);
   			return(OK);	
		case UDA.MOVED2D:	//[MOVED2D](float: dx)(float: dy)(float: da)
			dx = reader.readFloat();
			dy = reader.readFloat();
			float da = reader.readFloat();
			monitor.moved2D(device_source, dx, dy, da);
			return(OK);	
		case UDA.MOVED3D:	//[MOVED3D](float: dx)(float: dy)(float: dz)(float: ax)(float: ay)(float: az)
			dx = reader.readFloat();
			dy = reader.readFloat();
			float dz = reader.readFloat();
			float ax = reader.readFloat();
			float ay = reader.readFloat();
			float az = reader.readFloat();			
			monitor.moved3D(device_source, dx, dy, dz, ax, ay, az);
			return(OK);
   		}		
		//commande inconnu
		return(CMD_UNKNOW);
	}
		//encodeur vers Device
	public static IWriter encodeMove(IWriter writer, int cmd, int power){  
		return(encodeMove( writer,  cmd,  power, true));
	}
	public static IWriter encodeMove(IWriter writer, int cmd, int power, boolean write_code_cmd){  
		//[MOVE][commande_move](int: value)
		if(write_code_cmd) writer.writeByte(UDA.MOVE);
		writer.writeByte(cmd);
		writer.writeInt(power);
		return(writer);
	}	
	public static IWriter encodeAction(IWriter writer, int actuator_id, int value){
		return(encodeAction( writer,  actuator_id,  value, true));
	}
	public static IWriter encodeAction(IWriter writer, int actuator_id, int value, boolean write_code_cmd){ 
		//[ACTION](int: actuator_id)(val: value)
		if(write_code_cmd) writer.writeByte(UDA.ACTION);
		writer.writeInt(actuator_id);
		writer.writeInt(value);
		return(writer);
	}
	public static IWriter encodeAction(IWriter writer, int actuator_id, float value){
		return(encodeAction( writer,  actuator_id,  value, true));
	}
	public static IWriter encodeAction(IWriter writer, int actuator_id, float value, boolean write_code_cmd){  
		//[ACTION](int: actuator_id)(val: value)
		if(write_code_cmd) writer.writeByte(UDA.ACTION);
		writer.writeInt(actuator_id);
		writer.writeByte(DataBytes.FLOAT);
		writer.writeFloat(value);
		return(writer);
	}	
	public static IWriter encodeAction(IWriter writer, int actuator_id, byte[] value){
		return(encodeAction( writer,  actuator_id,  value, true));
	}	
	public static IWriter encodeAction(IWriter writer, int actuator_id, byte[] value, boolean write_code_cmd){ 
		//[ACTION](int: actuator_id)(val: value)
		if(write_code_cmd) writer.writeByte(UDA.ACTION);
		writer.writeInt(actuator_id);
		writer.writeByte(DataBytes.TBYTES);
		writer.writeTbytes(value);
		return(writer);
	}
	public static IWriter encodeGetState(IWriter writer){
		return(encodeGetState(writer, true));
	}
	public static IWriter encodeGetState(IWriter writer, boolean write_code_cmd){ 
		//[GET_STATE]
		if(write_code_cmd) writer.writeByte(UDA.GET_STATE);
		return(writer);
	}
	public static IWriter encodeGetParams(IWriter writer){  
		return(encodeGetParams( writer, true));
	}
	public static IWriter encodeGetParams(IWriter writer, boolean write_code_cmd){  
		//[GET_PARAMS]
		if(write_code_cmd) writer.writeByte(UDA.GET_PARAMS);
		return(writer);
	}
	public static IWriter encodeGetMoveCapacity(IWriter writer){
		return(encodeGetMoveCapacity(writer, true));
	}
	public static IWriter encodeGetMoveCapacity(IWriter writer, boolean write_code_cmd){ 
		//[GET_MOVE_CAPACITY]
		if(write_code_cmd) writer.writeByte(UDA.GET_MOVE_CAPACITY);
		return(writer);
	}	
	public static IWriter encodeGetSensorParam(IWriter writer, int sensor_id){  
		return(encodeGetSensorParam( writer,  sensor_id, true));
	}
	public static IWriter encodeGetSensorParam(IWriter writer, int sensor_id, boolean write_code_cmd){  
		//[GET_SENSOR_PARAMS](int: sensor_id)
		if(write_code_cmd) writer.writeByte(UDA.GET_SENSOR_PARAMS);
		writer.writeInt(sensor_id);
		return(writer);
	}
	public static IWriter encodeGetActuatorParam(IWriter writer, int actuator_id){ 
		return(encodeGetActuatorParam( writer,  actuator_id, true));
	}
	public static IWriter encodeGetActuatorParam(IWriter writer, int actuator_id, boolean write_code_cmd){ 
		//[GET_ACTUATOR_PARAMS](int: actuator_id)
		if(write_code_cmd) writer.writeByte(UDA.GET_ACTUATOR_PARAMS);
		writer.writeInt(actuator_id);
		return(writer);
	}	
	public static IWriter encodeReadSensor(IWriter writer, int sensor_id){ 
		return(encodeReadSensor( writer,  sensor_id, true));
	}
	public static IWriter encodeReadSensor(IWriter writer, int sensor_id, boolean write_code_cmd){ 
		//[READ_SENSOR](int: sensor_id)
		if(write_code_cmd) writer.writeByte(UDA.READ_SENSOR);
		writer.writeInt(sensor_id);
		return(writer);
	}	
	
		//encodeur vers Monitor
	public static IWriter encodeState(IWriter writer, boolean ready, boolean config_changed){
		return(encodeState(writer, ready, config_changed, true));
	}
	public static IWriter encodeState(IWriter writer, boolean ready, boolean config_changed, boolean write_code_cmd){ 
		//[STATE][ready][config_changed]
		if(write_code_cmd) writer.writeByte(UDA.STATE);
		writer.writeBool(ready);
		writer.writeBool(config_changed);
		return(writer);
	}	
	public static IWriter encodeParams(IWriter writer, String name, int type_device, int nb_sensor, int nb_actuator, boolean can_log){
		return(encodeParams(writer, name, type_device, nb_sensor, nb_actuator, can_log, true));
	}
	public static IWriter encodeParams(IWriter writer, String name, int type_device, int nb_sensor, int nb_actuator, boolean can_log, boolean write_code_cmd){  
		//[PARAMS](string : name)(int : type_device)(int: nb_sensor)(int: nb_actuator)[can_log]
		if(write_code_cmd) writer.writeByte(UDA.PARAMS);
		writer.writeString(name);
		writer.writeInt(type_device);
		writer.writeInt(nb_sensor);
		writer.writeInt(nb_actuator);
		writer.writeBool(can_log);
		return(writer);
	}
	public static IWriter encodeMoveCapacity(IWriter writer, int movement_type, int power_min, int power_moy, int power_max){
		return(encodeMoveCapacity( writer,  movement_type,  power_min,  power_moy,  power_max, true));
	}
	public static IWriter encodeMoveCapacity(IWriter writer, int movement_type, int power_min, int power_moy, int power_max, boolean write_code_cmd){ 
		//[MOVE_CAPACITY][movement_type](int: power_min)(int: power_moy)(int: power_max)
		if(write_code_cmd) writer.writeByte(UDA.MOVE_CAPACITY);
		writer.writeByte(movement_type);
		writer.writeInt(power_min);
		writer.writeInt(power_moy);
		writer.writeInt(power_max);	 
		return(writer);
	}
	public static IWriter encodeSensorParam(IWriter writer, int sensor_id, String name, int type_sensor, int type_value, boolean observable){
		return(encodeSensorParam( writer, sensor_id ,  name,  type_sensor,  type_value,  observable, true));
	}
	public static IWriter encodeSensorParam(IWriter writer, int sensor_id, String name, int type_sensor, int type_value, boolean observable, boolean write_code_cmd){  //nb : type_value sur 1 octet
		//[SENSOR_PARAMS](int : sensor_id)(string : name)(int : sensor_type)[type_value][observable]
		if(write_code_cmd) writer.writeByte(UDA.SENSOR_PARAMS);
		writer.writeInt(sensor_id);
		writer.writeString(name);
		writer.writeInt(type_sensor);
		writer.writeByte(type_value);
		writer.writeBool(observable);
		return(writer);
	}
	public static IWriter encodeActuatorParam(IWriter writer, int actuator_id, String name, int type_actuator, int type_value){ 
		return(encodeActuatorParam( writer,  actuator_id,  name,  type_actuator,  type_value, true));
	}
	public static IWriter encodeActuatorParam(IWriter writer, int actuator_id, String name, int type_actuator, int type_value, boolean write_code_cmd){ 	//nb : type_value sur 1 octet
		//[ACTUATOR_PARAMS](int : actuator_id)(string : name)(int : actuator_type)[type_value]
		if(write_code_cmd) writer.writeByte(UDA.ACTUATOR_PARAMS);
		writer.writeInt(actuator_id);
		writer.writeString(name);
		writer.writeInt(type_actuator);
		writer.writeByte(type_value);
		return(writer);
	}
	public static IWriter encodeSensorValue(IWriter writer, int sensor_id, int value){ 
		return(encodeSensorValue(writer, sensor_id, value, true));
	}
	public static IWriter encodeSensorValue(IWriter writer, int sensor_id, int value, boolean write_code_cmd){  
		//[SENSOR](int: sensor_id)(val: value)
		if(write_code_cmd) writer.writeByte(UDA.SENSOR);
		writer.writeInt(sensor_id);
		writer.writeInt(value);
		return(writer);
	}
	public static IWriter encodeSensorValue(IWriter writer, int sensor_id, float value){ 
		return(encodeSensorValue(writer, sensor_id, value, true));
	}	
	public static IWriter encodeSensorValue(IWriter writer, int sensor_id, float value, boolean write_code_cmd){  
		//[SENSOR](int: sensor_id)(val: value)
		if(write_code_cmd) writer.writeByte(UDA.SENSOR);
		writer.writeInt(sensor_id);
		writer.writeByte(DataBytes.FLOAT);
		writer.writeFloat(value);
		return(writer);
	}	
	public static IWriter encodeSensorValue(IWriter writer, int sensor_id, byte[]  value){ 
		return(encodeSensorValue(writer, sensor_id, value, true));
	}	
	public static IWriter encodeSensorValue(IWriter writer, int sensor_id, byte[] value, boolean write_code_cmd){  
		//[SENSOR](int: sensor_id)(val: value)
		if(write_code_cmd) writer.writeByte(UDA.SENSOR);
		writer.writeInt(sensor_id);
		writer.writeByte(DataBytes.TBYTES);
		writer.writeTbytes(value);
		return(writer);
	}
	public static IWriter encodeLog(IWriter writer, String text){
		return(encodeLog(writer, text, true));
	}
	public static IWriter encodeLog(IWriter writer, String text, boolean write_code_cmd){ 
		//[LOG](string : texte)
		if(write_code_cmd) writer.writeByte(UDA.LOG);
		writer.writeString(text);
		return(writer);
	}	
	public static IWriter encodeMoved2D(IWriter writer, float dx, float dy, float da){ 
		return(encodeMoved2D( writer,  dx,  dy,  da, true));
	}
	public static IWriter encodeMoved2D(IWriter writer, float dx, float dy, float da, boolean write_code_cmd){ 
		//[MOVED2D](float: dx)(float: dy)(float: da)
		if(write_code_cmd) writer.writeByte(UDA.MOVED2D);
		writer.writeFloat(dx);
		writer.writeFloat(dy);
		writer.writeFloat(da);
		return(writer);		
	}
	public static IWriter encodeMoved3D(IWriter writer, float dx, float dy, float dz, float ax, float ay, float az){
		return(encodeMoved3D( writer,  dx,  dy,  dz,  ax,  ay,  az, true));
	}
	public static IWriter encodeMoved3D(IWriter writer, float dx, float dy, float dz, float ax, float ay, float az, boolean write_code_cmd){ 
		//[MOVED3D](float: dx)(float: dy)(float: dz)(float: ax)(float: ay)(float: az)
		if(write_code_cmd) writer.writeByte(UDA.MOVED3D);
		writer.writeFloat(dx);
		writer.writeFloat(dy);
		writer.writeFloat(dz);
		writer.writeFloat(ax);
		writer.writeFloat(ay);
		writer.writeFloat(az);	
		return(writer);
	}
	//]

	//[Fonctions utiles
	public static int sizeToCodingInt(int data){
		if(data==0 || data==1) return(1);
		if(data<0) data = -data;

		if(data<256) return(2);
		else if(data<65500) return(3);
		return(5);
	}
	public static int sizeToCodingString(String data){
		return(sizeToCodingInt(data.length())+data.length());
	}
		//Vers Device
	public static int sizeToEncodeMove(int cmd, int power){ return(1+1+sizeToCodingInt(power)); }
	public static int sizeToEncodeAction(int actuator_id, int value){ return(1+sizeToCodingInt(actuator_id)+sizeToCodingInt(value)); }
	public static int sizeToEncodeAction(int actuator_id, float value){ return(1+sizeToCodingInt(actuator_id)+5); }
	public static int sizeToEncodeAction(int actuator_id, byte[] value){ return(1+sizeToCodingInt(actuator_id)+1+sizeToCodingInt(value.length))+value.length; }
	public static int sizeToEncodeGetState(){ return(1); }
	public static int sizeToEncodeGetParams(){ return(1); }
	public static int sizeToEncodeGetMoveCapacity(){ return(1); }
	public static int sizeToEncodeGetSensorParam(int sensor_id){ return(1+sizeToCodingInt(sensor_id)); }
	public static int sizeToEncodeGetActuatorParam(int actuator_id){ return(1+sizeToCodingInt(actuator_id)); }
	public static int sizeToEncodeReadSensor(int sensor_id){ return(1+sizeToCodingInt(sensor_id)); }
		//vers Monitor
	public static int sizeToEncodeState(boolean ready, boolean config_changed){ return(3); }
	public static int sizeToEncodeParams(String name, int type_device, int nb_sensor, int nb_actuator, boolean can_log){
		 return(1+sizeToCodingString(name)+sizeToCodingInt(type_device)+sizeToCodingInt(nb_sensor)+sizeToCodingInt(nb_actuator)+1);
	}	
	public static int sizeToEncodeMoveCapacity(int movement_type, int power_min, int power_moy, int power_max){
		 return(1+1+sizeToCodingInt(power_min)+sizeToCodingInt(power_moy)+sizeToCodingInt(power_max));
	}	
	public static int sizeToEncodeSensorParam(int sensor_id, String name, int type_sensor, int type_value, boolean observable){
		return(1+sizeToCodingInt(sensor_id)+sizeToCodingString(name)+sizeToCodingInt(type_sensor)+1+1);
	}	
	public static int sizeToEncodeActuatorParam(int actuator_id, String name, int type_actuator, int type_value){ 
		return(1+sizeToCodingInt(actuator_id)+sizeToCodingString(name)+sizeToCodingInt(type_actuator)+1);
	}	
	public static int sizeToEncodeSensorValue(int sensor_id, int value){ return(1+sizeToCodingInt(sensor_id)+sizeToCodingInt(value)); }
	public static int sizeToEncodeSensorValue(int sensor_id, float value){ return(1+sizeToCodingInt(sensor_id)+5); }
	public static int sizeToEncodeSensorValue(int sensor_id, byte[]  value){ return(1+sizeToCodingInt(sensor_id)+1+sizeToCodingInt(value.length))+value.length; }
	public static int sizeToEncodeLog(String text){ return(1+sizeToCodingString(text)); }
	public static int sizeToEncodeMoved2D(float dx, float dy, float da){ return(1+4+4+4);	}
	public static int sizeToEncodeMoved3D(float dx, float dy, float dz, float ax, float ay, float az){ return(1+4+4+4+4+4+4);	}
	//]
}