/**
	API du service :
 
 		int UdaUdp.expose(Hub hub, Device d);	//<- inscrit le device sur le gestionnaire UdaUdp de l'hub spécifié, retourne l'id attribué au device d
 							//NB: si d est déjà inscrit, l'id qui lui est déjà attribué est retourné
 							//0 si l'inscription est impossible (un maximum de UdaUdp.MAX_EXPOSED_DEVICE devices peut être géré (typiquement 200))

 		int UdaUdp.unexpose(Hub hub, Device d);	//<- désinscrit le device sur le gestionnaire UdaUdp, retourne l'id qui avait été attribué au device d
 			//NB: les moniteurs distant qui écoute ce device seront désincrit de l'écoute et recevront le message : [STATE][ready=false]
 			 
 		Device UdaUdp.linkDevice(Spot target, int device_id)	//<- crée un lien vers le device distant géré par le spot target sous l'identifiant device_id
 			// le message [STATE][ready=false] sera automatiquement envoyé aux écouteur inscrits sur ce device quand ce dernier deviendra plus exposé (unexpose ou down de l'hote porteur ou appel à unlinkDevice(...))  


		boolean UdaUdp.unlinkDevice(Device d)		//<- supprime le lien créé préalablement avec linkDevice
								// le message [STATE][ready=false] sera envoyé aux écouteurs locaux enregistré sur ce device

		void UdaUdp.exposedDevices(Spot target, ExposedDeviceResponse cb)	//<- invoce cb.exposedDevice(Spot target, int[][] exposed_device) avec la liste des devices exposés par l'hote target 
						//(il peut s'agir de l'hote local): exposed_device[i][0]=device_id , exposed_device[i][1]=device_type
			
		void UdaUdp.requestDevice(Spot target, int device_id, DataReader paquet_request, ResponseDevice cb_response [, boolean auto_dispose = true]) //<- lance une requête paquet_request sur le device ciblé, 
			//cb_response.deviceResponse(Spot target, int device_id, int code_perform, DataReader paquet_response) sera déclenché en réponse	

NB: le fait d'exposer un device conduit automatiquement au marquage SMD d'un code SKUAD.SMD_DEVICE avec en valeur le numéro de version de la définition de la liste de device exposé sur l'hote 
(ce numéro de version changera à chaque fois que la cette liste sera modifié (ajout/suppression de device). Cette valeur sera à zero (ou supprimée) si aucun device n'est exposé. 
*/
package skuad.uda.net;

import skuad.hubudp.*;
import skuad.uda.*;
import skuad.util.databytes.*;
import java.util.*;
import skuad.uda.implementation.udaudp.*;

public class UdaUdp {
	
	public final static int MAX_EXPOSED_DEVICE = 65500;
	
	private final static Map<Hub, UdaUdpManager> managers = new HashMap<Hub, UdaUdpManager>();
	
	private synchronized static UdaUdpManager getManager(Hub hub) {
		if(hub==null) hub = HubUdp.joinLocalHub();
		UdaUdpManager mng = managers.get(hub);
		if(mng==null) { //instanciation du manager UdaUdp pour cet hub
			mng = new UdaUdpManager(hub);
			managers.put(hub, mng);
		}
		return(mng);
	}
	
	public static int expose(Hub hub, Device d){ //<- inscrit le device sur le gestionnaire UdaUdp du Hub indiqué, retourne l'id attribué au device d, 0 en cas d'erreur
		if(d==null) return(0);
		UdaUdpManager mng=getManager(hub);
		return(mng.expose(d));
	}

	public static int unexposeDevice(Hub hub, Device d){	//<- désinscrit le device sur le gestionnaire UdaUdp du Hub indiqué, retourne l'id qui avait été attribué au device d
		if(d==null) return(0);
		UdaUdpManager mng=managers.get(hub);
		if(mng!=null) return(mng.unexposeDevice(d));
		return(0);
	}

	public static Device linkDevice(Spot target, int device_id){	//<- crée un lien vers le device distant géré par le spot target sous l'identifiant device_id
		if(target.isUp()){
			Hub hub = target.getHub();
			UdaUdpManager mng = getManager(hub);
			return(mng.linkDevice(target, device_id));
		}
		return(null);
	}

	public static boolean unlinkDevice(Device d){	//<- supprime le lien créé préalablement avec linkDevice
			// le message [STATE][ready=false] sera envoyé aux écouteurs locaux enregistré sur ce device
		return(d!=null && (d instanceof VirtualDevice) && ((VirtualDevice) d).unlink() );
	}

	public static void exposedDevices(Spot target, ExposedDeviceResponse cb){	//<- invoce cb.exposedDevice(DynIpPort source, int[][] exposed_device) avec la liste des devices exposés par l'hote target 
				//(il peut s'agir de l'hote local): exposed_device[i][0]=device_id , exposed_device[i][1]=device_type
		if(target==null || !target.isUp()) cb.exposedDevice(target, new int[0][0]); //cible non existante ou down
		else getManager(target.getHub()).exposedDevices(target, cb);
	}
	
	public static boolean requestDevice(Spot target, int device_id, DataReader paquet_request, ResponseDevice cb_response){ return(requestDevice(target, device_id, paquet_request, cb_response, true));}
	public static boolean requestDevice(Spot target, int device_id, DataReader paquet_request, ResponseDevice cb_response , boolean auto_dispose){ //<- lance une requête paquet_request sur le device ciblé, 
		//cb_response.deviceResponse(DynIpPort target, int device_id, int code_perform, DataReader paquet_response) sera déclenché en réponse --> paquet_response est une insatnce de DataReader qu'il faudra diposé aprés lecture ( .dispose())	
		if(target==null || !target.isUp()){
			if(cb_response!=null) cb_response.deviceResponse(target, device_id, UDA.NOT_EXIST, null);
			return(false);
		}
		return( getManager(target.getHub()).requestDevice(target, device_id, paquet_request, cb_response, auto_dispose) );
	}
	//]
}