/**
 * Cette classe permet de parser le tableau des parèmetres de lancement d'un programme
 * afin de faciliter l'analyse des valeurs et des champs (couple nom, valeur) présents.
 * 
 * Syntaxe d'une chaine de paramètres de lancement :
 * 
 *  value_0  -fieldA [value0_fieldA value1_fieldA ... ] -fieldB [value0_fieldB ...] ...etc
 * 
 * -> un paramètre de la forme : value_X ou valueX_fieldY est dit paramètre de type valeur, ainsi tous les paramètres ne commençant pas
 *    par le symbole '-' sont concidérés comme une valeur
 * -> un paramètre de la forme : -fieldY [value0_fieldY ... ] est dit pramètre de type champ, de nom fieldY ayant pour ensemble de valeur 
 *    value0_fieldY, value1_fieldY,... qui peut éventuellement être vide. 
 *    Chaque élément de la suite de valeur d'un champ est associé à un numéro de rang (position de la valeur dans la suite) qui peut être
 *    utilisé pour leur consultation.
 * -> La consultation d'un champ se fera par son nom via la méthode val((String) name [, rang] [,String defaut_value]), avec rang=0
 *    par défaut.
 * -> La consultation d'une valeur (non champ) se fera via la méthode val((int) pos [,String defaut_value]). L'indice pos d'une
 *    valeur est sa position dans la chaine des paramètres par rapport aux autres valeurs (sans compter le nom des champs).
 *    
 * /!\ toute valeur qui suit le nom champ sera interprétée comme une des valeurs du champ. Si ce ne doit pas être le cas il faut explicitement
 *    terminer la liste des valeurs du champ via un symbole '-' isolé.
 *    
 * /!\ quand on consulte les valeurs, il faut garder à l'esprit que toutes les valeurs des champs sont aussi considérées comme des valeurs
 *     isolées.   
 * 
 * API :
 * 		//Consulation des valeurs
 *  		- int nbValue() : nombre de paramètre de type valeur (-> plage de valeur pour val((int) pos [,...]).
 *  		- boolean existValue(String) : true si la valeur indiqué existe.
 * 		- String val(int pos [,String defaut_value]) : la valeur de position pos.
 * 		- int intVal(int pos, int defaut_value)
 * 
 * 		//Consulation des champs
 *     	- int nbField() : nombre de paramètre de type champ.
 *  		- int existField(String) : true si le champ de nom name existe.
 *  		- int sizeField(String name) : nombre de valeur associé au champ de nom name (on doit utiliser rang< sizeField(name) pour val(String name, rang [,...] )
 *  		- String val(String name [, rang] [,String defaut_value]) : la valeur de position rang (=0 par défaut) du champ de nom name
 *  		- int intVal(String name [, rang] , int defaut_value) 
 */
package skuad.util.base;

import java.util.*;

public class LaunchOption {
	
	Map<String, Vector<String>> fields;	//map des champs
	Vector<String> values;				//tableau des valeurs
	
	public LaunchOption(String[] tab) {
		fields = new HashMap<String, Vector<String>>();
		values = new Vector<String>();
		
		//on parse le tableau des paramètre
		if(tab!=null) {
			Vector<String> cur_val_field = null;
			for(int i=0; i<tab.length; i++) {
				String cur = tab[i];
				if(cur!=null && cur.length()>0) {
					if(cur.charAt(0)=='-') { //c'est un champ
						cur = cur.substring(1);
						if(cur.length()>0) {
							cur_val_field = new Vector<String>();
							fields.put(cur, cur_val_field);
						} else cur_val_field=null;
					} else { //c'est une valeur
						values.add(cur); 
						if(cur_val_field!=null) cur_val_field.add(cur);
					}
				}
			}
		}
	}

	//Consulation des valeurs
	public int nbValue() { return(values.size()) ;}
	public boolean existValue(String val) {
		if(val!=null) for(int i=0; i<values.size(); i++) {
			if(val.equals(values.get(i))) return(true);
		}
		return(false);
	}
	public String val(int pos){
		return(pos>=0 && pos<values.size() ? values.get(pos) : null);
	}
	public String val(int pos, String def_val){
		return(pos>=0 && pos<values.size() ? values.get(pos) : def_val);
	}
	public int intVal(int pos, int def_val) { 
		String st_val = val(pos, null);
		if(st_val!=null) {
			try { return(Integer.parseInt(st_val)); }
			catch(Exception e) {}
		}
		return(def_val);
	}
	
	//Les champs
	public int nbField() { return(fields.size()) ;}
	public boolean existField(String name) { return(fields.containsKey(name)); }
	public int sizeField(String name) {
		Vector<String> tabv = fields.get(name);
		if(tabv!=null) return(tabv.size());
		else return(0);
	}
	public String val(String name){ return(val(name,0,null));}
	public String val(String name, String def_val){ return(val(name,0,def_val));}
	public String val(String name, int rang) { return(val(name, rang, null));}
	public String val(String name, int rang, String def_val){
		Vector<String> tabv = fields.get(name);
		if(tabv!=null && rang>=0 && rang<tabv.size()) return(tabv.get(rang));
		else return(def_val);
	}
	public int intVal(String name, int def_val) { 
		return(intVal(name, 0, def_val));
	}
	public int intVal(String name, int rang, int def_val) { 
		String st_val = val(name, rang, null);
		if(st_val!=null) {
			try { return(Integer.parseInt(st_val)); }
			catch(Exception e) {}
		}
		return(def_val);
	}
}