/** 
 Cette classe spécifie les éléments de protocole de communication UDA pour les flux binaire (Stream)
 
*/

package skuad.util.databytes;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import skuad.util.value.Value;

public class StreamWriter implements IWriter {

	OutputStream target;
	Lock sendlock;
	boolean error;
	
	public StreamWriter(OutputStream t){
		target=t;
		error=false;
		sendlock=new ReentrantLock();
	}
	
	public void lock() { sendlock.lock(); }
	public void unlock(){ sendlock.unlock(); }
	public boolean dispose() {
		try{ target.close(); target=null; }catch(Exception e){}
		return true;
	}
	
	//[Interface IWriter
	public boolean writeByte(int val){try{ 		//écrit une valeur octet
		target.write(val);
	} catch(Exception e){ error=true; return(false); } return(true); }		
	public boolean writeBool(boolean val){try{	//écrit une valeur 0 ou 1 en fonction de val
		target.write(val ? 1 : 0);
	} catch(Exception e){ error=true; return(false); } return(true); }
	public boolean writeInt(int data){try{ 		//écrit la valeur (int: label)
		boolean neg=data<0;
		if(neg) data = -data;
		if(!neg && data<DataBytes.TINY_INT) target.write(data);
		else if(data<256) { target.write(neg ? DataBytes.NBYTE : DataBytes.BYTE); target.write(data); }
		else if(data<65500) { target.write(neg ? DataBytes.NSHORT : DataBytes.SHORT); writeShort(data);  }
		else { target.write(neg ? DataBytes.NWORD : DataBytes.WORD); writeWord(data);  }
	} catch(Exception e){ error=true; return(false); } return(true); }
	public boolean writeFloat(float data){try{	//écrit la valeur (float: label)
		int bits = Float.floatToIntBits(data);
		target.write((bits >> 24) & 0xff);
		target.write((bits >> 16) & 0xff);
		target.write((bits >> 8) & 0xff);
		target.write(bits & 0xff);
	} catch(Exception e){ error=true; return(false); } return(true); }
	public boolean writeString(String data){try{ //écrit le codage (string: val)
		 byte[] text = data.getBytes(StandardCharsets.US_ASCII);	 
		 writeInt(text.length);		 
		 target.write(text);
	} catch(Exception e){ error=true; return(false); } return(true); }
	public boolean writeTbytes(byte[] data){try{ //écrit le codage (tab: val)
		writeInt(data.length);
		target.write(data);			
	} catch(Exception e){ error=true; return(false); } return(true); }
	public boolean writeValue(Value val){	//écrit une valeur (value : val) //[type_value](valeur)
		//		type_value -> NULL	--> valeur NULL
		int nb_element;
		if(val==null) writeByte(Value.NULL);
		else {
			byte type = val.getType();
			writeByte(type);			
			switch(type){ //{@link #STRING}, {@link #INT}, {@link #FLOAT}, {@link #BOOL}, {@link #ARRAY} ou {@link #HASH}
			case Value.STRING:	writeString(val.s()); break;
			case Value.INT: writeInt(val.toInt()); break;
			case Value.FLOAT: writeFloat(val.toFloat()); break;
			case Value.BOOL: writeBool(val.toBool()); break;
			case Value.BYTES:
				byte[] bs = val.bs();
				nb_element = bs.length;
				writeInt(nb_element);
				copyBytes(bs);
				break;		
			case Value.ARRAY:
				nb_element = val.size();
				writeInt(nb_element);
				for(int i=0; i<nb_element; i++) writeValue(val.get(i));
				break;
			case Value.HASH:
				nb_element = val.size();
				writeInt(nb_element);
				for(String name : val.keys()) {
					writeString(name);
					writeValue(val.get(name));
				}
				break;
			//ERROR, NULL ou UNDEFINED --> RAS
			}
		}
		return(true);
	}
	
	//[Primitive de lecture/écriture des valeurs brutes :
	public boolean writeShort(int data){try{    //entier sur 2 octets
		target.write((data >> 8) & 0xFF);
		target.write((data & 0xFF));		
	} catch(Exception e){ error=true; return(false); } return(true); }
	public boolean writeWord(int data){try{    //entier sur 4 octets
		target.write((data >> 24) & 0xff);
		target.write((data >> 16) & 0xff);
		target.write((data >> 8) & 0xff);
		target.write(data & 0xff);
	} catch(Exception e){ error=true; return(false); } return(true); }
	//]

	public int writeData(DataReader content) {
		int nbw=0;
		try { while(content.avaible()>0) { target.write(content.readByte()); nbw++; }
			return(nbw);
		} catch(Exception e){ error=true; return(nbw); }
	}

	public int writeData(DataReader content, int size) {
		int i=0;
		try { 	for( ; content.avaible()>0 && i<size; i++) target.write(content.readByte());
				return(i);
		} catch(Exception e){ return(i); }
	}

	public StreamWriter b(int value) {
		writeByte(value);
		return(this);
	}
	public StreamWriter b(boolean value) {
		writeByte(value ? 1 : 0);
		return(this);
	}
	public StreamWriter b(byte[] value) {
		copyBytes(value);
		return(this);
	}

	public StreamWriter i(int value) {
		writeInt(value); 
		return(this);
	}
	public StreamWriter s(String value) {
		writeString(value); 
		return(this);
	}
	public StreamWriter d(long value) {
		 writeWord((int)(value  >> 32) & 0xFFFFFFFF); 
		 writeWord((int)(value & 0xFFFFFFFF)); 
		 return(this);
	}
	
	public StreamWriter v(Value value) {
		writeValue(value); 
		return(this);
	}

	public boolean writeBruteString(String msg) {
		byte[] tab = msg.getBytes(StandardCharsets.US_ASCII);
		try { 	 target.write(tab); 
				return(true);
		} catch(Exception e){ error=true; return(false); }
	}

	public boolean copyBytes(byte[] data) {
		try { 	 target.write(data); 
			return(true);
		} catch(Exception e){ error=true; return(false); }
	}

	public boolean copyBytes(byte[] data, int start, int length) {
		try { 	 target.write(data, start, length);
				return(true);
		} catch(Exception e){ error=true; return(false); }
	}

	public StreamWriter sh(int value) {
		writeShort(value); 
		return(this); 
	}
	
	public StreamWriter w(int value) {
		writeWord(value); 
		return(this);
	}
	//]
}