/* BDB : Block DataBytes 
 
 Classe noyau (seed) d'un block de donnée sérialisé dans un Buffer d'octet
 
 Un BDB est l'association : d'un code de block (sur 1 octet, ce peut être le code d'une commande ou d'un protocole par exemple) 
 et d'un buffer de donnée encodé au format DataBytes
 
-> On ne peut pas instancier directement un objet de cette classe, on utilise plutôt les méthodes writer(code, maxSize) et seed(code, maxSize)
 	fournit par la librairie.
  A noter que cette classe BDB implémente une pile de recyclage d'instance pour optimiser la gestion mémoire. 
  
   ==> aprés utilisation il faut faire appel à dispose() pour libérer les ressources internes.
*/

package skuad.util.databytes.implementation;

import java.io.InputStream;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.nio.charset.StandardCharsets;
import java.util.LinkedList;

public class BDB {
	
	public final static int MAX_RECYCLED = 20; 				//taille maximum de la pile de recyclage des instances
	public final static int BUFFER_DATA_SIZE = 200; 		//taille des buffer recyclable (les tailles + grande ne seront pas concervés)
	protected static LinkedList<BDB> reclycledStack=new LinkedList<BDB>(); //pile de recyclage
	
	protected static BDB seed(){ return(seed(BUFFER_DATA_SIZE)); }
	protected static BDB seed(int maxSize){ 
		BDB seed=null;
		if(maxSize<=BUFFER_DATA_SIZE) try{ seed=reclycledStack.pop(); } catch(Exception e){}
		if(seed==null) seed=new BDB(maxSize); else seed.init();
		return(seed);
	}
	public static BDB seed(byte[] buff, int start, int size){ 
		BDB seed=null;
		if(size<=BUFFER_DATA_SIZE) try{ seed=reclycledStack.pop(); } catch(Exception e){}
		if(seed==null) seed=new BDB(size); else seed.init();
		seed.size=size;
		for(int s=start, d=0; d<size; ++s, ++d) seed.data[d]=buff[s];
		return(seed);
	}
	public static BDB seed(InputStream in, int size){ 
		BDB seed=null;
		if(size<=BUFFER_DATA_SIZE) try{ seed=reclycledStack.pop(); } catch(Exception e){}
		if(seed==null) seed=new BDB(size); else seed.init();
		seed.size=size;
		int cur=0;
		try{ while(size>0) {
			int r = in.read(seed.data, cur, size);
			cur+=r;
			size-=r;
		}} catch(Exception e) { seed.size=0; }
		return(seed);
	}	
	int size;
	int count;	//0 si seul une instance d'écriture détient le seed, >0 si 1 ou plusieurs instance de lecture détient le seed, <0 si aucune instance ne détient le seed
	byte[] data;
	String fullStringValue;

	private BDB(int maxSize){ 
		data = new byte[maxSize];
		init();
	}
	private void init(){
		size=0;
		count=1;
		fullStringValue=null;		
	}
	public void dispose(){ 	
		if(count>0){
			count--; 
			if(count==0 && reclycledStack.size()<=MAX_RECYCLED && data.length==BUFFER_DATA_SIZE) synchronized(reclycledStack){
				count=-1;
				reclycledStack.push(this); 
			}
		}
	}

	//[Méthodes internes pour les aspects lectures
	public String toString(){ //retourne l'ensemble du buffer de donnée sous forme de chaine de caractère (ne pas utiliser si les données sont suceptible d'être modifié)
		if(fullStringValue==null) fullStringValue=new String(data, 0, size, StandardCharsets.US_ASCII);
		return(fullStringValue);
	}
	public int getSize(){ return(size); }

	public DatagramPacket toDatagramPacket(InetAddress iaddr, int port, int vsize){
//System.out.println("-- seed size for DatagramPacket = "+vsize);		
		return(new DatagramPacket(data, vsize, iaddr, port));			
	}
	public byte[] copyBuff(int start, int end){ //retourne une copie du buffer sur la plage [start; end[
		int buff_size=end-start;
		byte[] rep = new byte[buff_size];
		for(int s=start, d=0; s<end; s++, d++) rep[d]=data[s];
		return(rep);
	}	
	//]
	
	//[Méthodes internes pour les aspects écritures (pour BdbWriter)
	protected BdbReader toReader(int vsize, boolean inc_count){ 
		size=vsize;
		if(inc_count) count++;
		return(new BdbReader(this));
	}
	protected BDB clone(int vsize){ 
		if(count>0){
			BDB clone =  seed(vsize);
			clone.size=vsize;
			for(int s=0; s<size; s++) clone.data[s]=data[s];
			return(clone);
		}
		return(null);		
	}
	//]
	
	//[Méthode pour les aspects lecture (pour BdbReader)
	public BdbReader newReader(boolean inc_count){ 
		if(inc_count) count++;
		return(new BdbReader(this));
	}
	//]
}