/**
 * Classe des gestionnaires de lecture de commande bufferisé
 * 
 *   -> aprés utilisation il faut faire appel à dispose() pour libérer les ressources interne.
 */
package skuad.util.databytes.implementation;

import java.io.OutputStream;
import java.net.InetAddress;
import java.nio.charset.StandardCharsets;

import skuad.util.value.Value;
import skuad.util.value.ValueWriter;
import skuad.util.databytes.*;
import skuad.util.base.Network;

public class BdbReader implements DataReader {
	BDB seed;
	boolean locked;
	int rpos;			//position de lecture dans le buffer
	int start_rpos;		//position de départ dans le buffer (celle ou sera replacer le pointeur de lecture lors d'un reset() )
	
	protected BdbReader(BDB vseed){
		seed=vseed;
		locked=false;
		rpos=0;
		start_rpos = 0;
	}
	
	public int getByte(int pos) {
		int real_pos = start_rpos+pos;
		if(!locked && real_pos<seed.size) return(0xFF & seed.data[real_pos]);
		return(0);
	}
	
	public int curPos(){			//retourne la position de lecture courante
		return(rpos-start_rpos);
	}
	public boolean toPos(int pos){	//repositionne le curseur de lecture sur la position indiqué
		if((start_rpos+pos)>=seed.size) return(false);
		rpos = start_rpos+pos;
		return(true);
	}
	
	public int size(){ return(seed.size-start_rpos); }	//taille total des données disponible
	public int avaible(){ return(seed.size-rpos); }		//taille des données restant à lire
	
	public void split(int start){ //supprime le tronçons [0;start[ de l'espace de lecture, et réinitialise le pointeur de lecture sur cette position
		start_rpos+=start;
		rpos = start_rpos;
	}
	public void reset(){ //réinitialise le pointeur de lecture
		rpos = start_rpos;
	}
	
	public boolean dispose(){ 
		if(!locked){
			locked=true;
			seed.dispose();
			return(true);
		}
		return(false);
	}	
	public DataReader clone(){ return(splitClone(0));}
	public DataReader cloneRest(){ return(splitClone(rpos-start_rpos)); }
	public DataReader splitClone(int pos_start){
		if(!locked) {
			BdbReader clone = seed.newReader(true);
			clone.start_rpos = start_rpos+pos_start;
			clone.rpos =  clone.start_rpos;
			return(clone);
		}
		return(this);
	}

	//[interface CmdReader
	public int readByte(){		//lecture du prochain octet, 0 si impossible ou erreur
		if(canRead()) return(0xFF & seed.data[rpos++]);
		return(0);
	}
	public int readInt(){ //lecture du prochain entier : [type]{value}
		return(readInt(readByte()));
	}	
	public int readInt(int type){ //lecture du prochain entier : {value}
		if(type<DataBytes.TINY_INT) return(type);
		int val=0;
		switch(type){
		case DataBytes.BYTE :  return(readByte());
		case DataBytes.NBYTE : return(-readByte());
		case DataBytes.SHORT : return(readShort());
		case DataBytes.NSHORT : return(-readShort());
		case DataBytes.WORD : return(readWord());
		case DataBytes.NWORD : return(-readWord());
		}
		return(val);		
	}	
	public boolean readBool(){ //lecture du prochain octet (non signé)
		 return(readByte()!=0);
	}
	public float readFloat(){ //lecture d'une valeur value sur 4 octets (poids faible -> poids fort) 
		if(canRead(4)){
			int rep;
			rep = readByte(); 
			rep = rep << 8 | readByte();
			rep = rep << 8 | readByte();
			rep = rep << 8 | readByte();
			return(Float.intBitsToFloat(rep));
		}
		return(0);
	}	
	
	public String readString(){ 	//lecture de la prochaine chaine de caractère : (string: label)
		int length = Math.min(readInt(), seed.size-rpos);
		String rep;
		if(length>0 && canRead(length)) {
			rep = new String(seed.data, rpos, length, StandardCharsets.US_ASCII);
			rpos+=length;
		} else rep ="";
		return(rep);
	}
	public String readString(int length){ //lecture de la prochaine chaine de caractère : {letter : length}
		String rep;
		if(length>0 && canRead(length)) {
			rep = new String(seed.data, rpos, length, StandardCharsets.US_ASCII);
			rpos+=length;
		} else rep ="";
		return(rep);
	}
	public byte[] readTbytes(){ //lecture de la valeur buffer : (int: size){size : contenu}
		int length = readInt();
		if(length>0 && canRead(length)){
			byte[] buff=new byte[length];
			for(int i=0; i<length; i++) buff[i]=seed.data[rpos++];
			return(buff);
		}
		return(null);
	}
	public Value readValue(){  
		int type = readByte();
		int nb_element;
		switch(type){
		case Value.STRING:	return(Value.s(readString()));
		case Value.INT: return(Value.i(readInt()));
		case Value.FLOAT: return(Value.f(readFloat()));
		case Value.BOOL:return(Value.b(readBool()));
		case Value.ARRAY:
			nb_element = readInt();
			ValueWriter va = Value.writer();
			for(int i=0; i<nb_element; i++) va.add(readValue());
			return(va.toConstant());
		case Value.HASH:
			nb_element = readInt();
			ValueWriter vh = Value.writer();
			for(int i=0; i<nb_element; i++) vh.set(readString(), readValue());
			return(vh.toConstant());
		case Value.NULL: return(Value.NULL_VALUE);
		case Value.UNDEFINED: return(Value.UNDEFINED_VALUE);
		}
		return(Value.ERROR_VALUE);
	}	
	//]
	
	//[Méthodes de lecture
	public boolean canRead(){
		return(!locked && rpos<seed.size); //e.i: rpos+1<=seed.size
	}
	public boolean canRead(int l){
		return(!locked && (rpos+l)<=seed.size);
	}
	
	public void skip(int l){ rpos+=l; }
	public String toString(){ //retourne l'ensemble du buffer de donnée sous forme de chaine de caractère (ne déplace pas le pointeur de lecture)
		if(locked) return("");
		return(seed.toString());
	}	
	public byte[] readBytes(byte[] buff, int length){ //lecture de length octets
		if(length>0 && canRead(length)){
			if(buff==null) buff=new byte[length];
			for(int i=0; i<length; i++) buff[i]=seed.data[rpos++];
		}
		return(buff);
	}
	public int writeToStream(OutputStream out, int length){ //transmission de length octets
		if(length>0 && canRead(length)){
			try{ out.write(seed.data, rpos, length); } catch(Exception e) { return(-1); }
			rpos+=length;
			return(length);
		}
		return(0);
	}
	public byte[] readBytes(byte[] buff, int length, int to_start_offset){ //lecture de length octets
		if(length>0 && canRead(length)){
			if(buff==null) buff=new byte[length+to_start_offset];
			for(int d=to_start_offset, i=0; i<length; i++) buff[d++]=seed.data[rpos++];
		}
		return(buff);		
	}	
	public int readShort(){ //lecture du prochain entier sur 2 octet (non signé)
		if(canRead(2)){
			int rep;
			rep = readByte();
			rep = rep << 8 | readByte(); 		
			return(rep);
		}
		return(0);
	}
	public int readWord(){ //lecture du prochain entier sur 4 octet (non signé)
		if(canRead(4)){
			int rep;
			rep = readByte(); 
			rep = rep << 8 | readByte();
			rep = rep << 8 | readByte();
			rep = rep << 8 | readByte();
			return(rep);
		}
		return(0);
	}	
	
	public long readDoubleWord(){ 	//leccture d'un double word (8 octets)
		long rep;
		rep = readWord(); 
		rep = rep << 32 | readWord();
		return(rep);
	}
	
	public String readStringIp(){
		return(readByte()+"."+readByte()+"."+readByte()+"."+readByte());
	}
	public InetAddress readIp(InetAddress sender){
		byte[] byte_ip = readBytes(null, 4);
		if(byte_ip!=null && byte_ip.length==4 && byte_ip[0]==127 && byte_ip[1]==0 && byte_ip[2]==0 && byte_ip[3]==1) return(sender);
		try { return(InetAddress.getByAddress(byte_ip)); } catch(Exception e){ return(null);}
	}
	public String readStringIp(InetAddress sender){
		return(Network.addressToIp(readIp(sender)));
	}
}