/**
 * Classe des gestionnaires d'écriture de commande bufferisé
 *   
 *  - aprés utilisation il faut faire appel à dispose() pour libérer les ressources interne.
 * 
 */
package skuad.util.databytes.implementation;

import java.net.DatagramPacket;
import java.net.InetAddress;
import skuad.util.databytes.*;
import skuad.util.value.Value;

import java.nio.charset.StandardCharsets;

public class BdbWriter implements DataWriter{
	BDB seed;
	boolean locked;
	int wpos;	//position d'écriture courante dans le buffer
	
	public BdbWriter(){
		seed=BDB.seed();
		locked=false;
		wpos=0;
	}
	public int size() { return(wpos); }		//taille total des données disponible
	public boolean setSize(int size) {
		if(size>=0) {
			if(size>wpos) expends(size-wpos);
			wpos=size;
			return(true);
		} else return(false);
	}
	public BdbWriter(int initiale_size){
		seed=BDB.seed(initiale_size);
		locked=false;
		wpos=0;
	}
	
	public boolean dispose(){
		if(!locked){
			locked = true;
			seed.dispose();
			return(true);
		}		
		return(false); //déjà disposé
	}
	
	public byte[] flushBuff(){ //retourne une copie du buffer + dispose
		byte[] buff=null;
		if(!locked){
			buff = seed.copyBuff(0, wpos);
			dispose();
		}
		return(buff);
	}

	public DataReader flushReader(){ //retourne une instance de lecture + dispose l'instance d'écriture courante
		BdbReader reader=null;
		if(!locked){
			reader = seed.toReader(wpos, false);
			locked = true; //dispose(); /!\ pas de dispose pour laisser l'instance seed vivante
		}
		return(reader);
	}
	
	public DataReader copyToReader(){ //retourne une instance de lecture sur une copie des donnée (l'instance courante n'est pas disposé)
		BdbReader reader=null;
		if(!locked) reader = seed.clone(wpos).toReader(wpos, false);
		return(reader);
	}

	public byte[] copyToBuff(){ //retourne une copie du buffer 
		if(!locked) return(seed.copyBuff(0, wpos));
		return(null);
	}
	
	public DatagramPacket toDatagramPacket(InetAddress iaddr, int port){
		if(!locked) return(seed.toDatagramPacket(iaddr, port, wpos));		
		return(null);
	}
	
	protected void expends(int size){ //s'assure que le buffer interne dispose de size octet de libre à partir de wpos, augmente la taille du buffer si ce n'est pas le cas.
		if((wpos+size)>=seed.data.length){ //on doit augmenter la taille du buffer
			int need = ((((wpos+size) - seed.data.length)/BDB.BUFFER_DATA_SIZE)+1)*BDB.BUFFER_DATA_SIZE;
			byte[] new_buff= new byte[need+seed.data.length];
			for(int i=0; i<wpos; i++) new_buff[i]=seed.data[i];
			seed.data=new_buff;
		}
	}
	
	//[API d'écriture
	public boolean setByte(int pos, int value){ //écrit l'octet value, à la position indiqué (sans déplacer le pointeur d'écriture), renvoie false et sans effet si pos>=position_courante_d_ecriture
		if(locked || pos>=wpos) return(false);
		seed.data[pos]=(byte) (value & 0xFF);
		return(true);
	}
	public int getByte(int pos){ if(locked || pos>=wpos) return(-1); else return(seed.data[pos]); }
	
	public void reset(){ wpos=0; }
	
	public int writeData(DataReader content){ //écrit les données de content (depuis sa position de lecture courante)
		if(locked) return(0);
		int size = content.avaible();
		if(content instanceof BdbReader){			
			BdbReader reader = (BdbReader) content;
			if(!copyBytes(reader.seed.data, reader.rpos, size)) return(0);
		} else {
			expends(size);
			int cur_pos_content = content.curPos();
			for(int i=0; i<size; i++) seed.data[wpos++]=(byte) (content.readByte() & 0xFF);
			content.toPos(cur_pos_content);
		}
		return(size);
	}

	public int writeData(DataReader content, int size){ //écrit size octetes de données de content (depuis sa position de lecture courante), retourne le nombre d'octet réellement écrit
		//,  ==> ne change pas la position de lecture de content
		if(locked) return(0);
		size = Math.min(size, content.avaible());
		if(content instanceof BdbReader){			
			BdbReader reader = (BdbReader) content;
			if(!copyBytes(reader.seed.data, reader.rpos, size)) return(0);
		} else {
			expends(size);
			int cur_pos_content = content.curPos();
			for(int i=0; i<size; i++) seed.data[wpos++]=(byte) (content.readByte() & 0xFF);
			content.toPos(cur_pos_content);
		}
		return(size);
	}
	
	//[Interface DataWriter	
	public boolean writeByte(int value){ //écrit l'octet value, puis déplace le pointeur d'écriture 
		if(locked) return(false);
		expends(1);
		seed.data[wpos++]=(byte) (value & 0xFF);
		return(true);
	}
	public boolean writeBool(boolean v){ //écrit l'octet value, puis déplace le pointeur d'écriture 
		if(locked) return(false);
		expends(1);
		seed.data[wpos++]=(byte) (v ? 1 : 0);
		return(true);
	}
	public boolean writeInt(int data){ //écriture d'un entier au format typé :  (int: size) = [type_int]{value}  (typage entier bas niveau UDA)
		if(locked) return(false);
		expends(5);
		if(data >= 0 && data <DataBytes.TINY_INT) this.writeByte(data);
		else {
			boolean neg=data<0;
			if(neg) data = -data;
			if(data<255) { 
				seed.data[wpos++]=(byte) (neg ? DataBytes.NBYTE : DataBytes.BYTE);
				seed.data[wpos++]=(byte) data;
			} else if(data<65500) { 
				seed.data[wpos++]=(byte) (neg ? DataBytes.NSHORT : DataBytes.SHORT);
				writeShort(data);
			} else { 
				seed.data[wpos++]=(byte) (neg ? DataBytes.NWORD : DataBytes.WORD);
				writeWord(data);
			}
		}
		return(true);
	}
	public boolean writeFloat(float value){ //écriture d'un float au format typé : (float: label) = { 4 : valeur } , NB: poids fort -> poids faible
		if(locked) return(false);
		expends(4);

		int bits = Float.floatToIntBits(value);
		seed.data[wpos++]=(byte) ((bits >> 24) & 0xff);
		seed.data[wpos++]=(byte) ((bits >> 16) & 0xff);
		seed.data[wpos++]=(byte) ((bits >> 8) & 0xFF);
		seed.data[wpos++]=(byte) ((bits & 0xFF));
		return(true);
	}	
	public boolean writeString(String msg){ 	//copie les octets de la chaine : (string: msg) = (int: size){size : contenu}
		if(locked) return(false);
		if(msg==null) msg="";
		int length = msg.length();
		expends(length+5);

		byte[] tab = msg.getBytes(StandardCharsets.US_ASCII);
		writeInt(length);
		for(int i=0; i<length; i++) seed.data[wpos++]=tab[i];
		return(true);
	}
	public boolean writeTbytes(byte[] data){ //écriture la valeur (tab: value)
		if(locked) return(false);
		expends(data.length+5);

		writeInt(data.length);
		for(int i=0; i<data.length; i++) seed.data[wpos++]=data[i];
		return(true);
	}	
	public boolean writeValue(Value val){	//écrit une valeur (value : val) //[type_value](valeur)
		//		type_value -> NULL	--> valeur NULL
		if(locked) return(false);
		expends(1);
		int nb_element;
		if(val==null) seed.data[wpos++]=(byte) (Value.NULL);
		else {
			int type = val.getType();
			 seed.data[wpos++]=(byte) (type);			
			switch(type){ //{@link #STRING}, {@link #INT}, {@link #FLOAT}, {@link #BOOL}, {@link #ARRAY} ou {@link #HASH}
			case Value.STRING:	writeString(val.s()); break;
			case Value.INT: writeInt(val.toInt()); break;
			case Value.FLOAT: writeFloat(val.toFloat()); break;
			case Value.BOOL: writeBool(val.toBool()); break;
			case Value.ARRAY:
				nb_element = val.size();
				writeInt(nb_element);
				for(int i=0; i<nb_element; i++) writeValue(val.get(i));
				break;
			case Value.HASH:
				nb_element = val.size();
				writeInt(nb_element);
				for(String name : val.keys()) {
					writeString(name);
					writeValue(val.get(name));
				}
				break;
			//ERROR, NULL ou UNDEFINED --> RAS
			}
		}
		return(true);
	}
	//]

	//[Méthode d'écriture courte pour appel en cascade (pas de vérification d'erreur)
	public BdbWriter b(int value){ writeByte(value); return(this); }	//écriture d'un byte
	public BdbWriter b(boolean value){ writeByte(value ? 1 : 0); return(this); }	//écriture d'un byte
	public DataWriter sh(int value){ writeShort(value); return(this); }
	public DataWriter b(byte[] value){ copyBytes(value, 0, value.length); return(this); } // -> copyBytes
	public BdbWriter i(int value){ writeInt(value); return(this); }	//écriture d'un entier
	public DataWriter s(String value){ writeString(value); return(this); } //écriture d'une chaine de catactère (string: val)
	public DataWriter d(long value){ writeWord((int)(value  >> 32) & 0xFFFFFFFF); writeWord((int)(value & 0xFFFFFFFF)); return(this); }  //écriture d'un double word
	public DataWriter v(Value value){ writeValue(value); return(this); } //écriture d'une value (value: val)
	public DataWriter w(int value){ writeWord(value); return(this); }  	//écriture d'un mot (4 octets)
	//]
	
	//[Méthodes d'écritures
	public int setDataString(String msg){ //remplie l'intégralité du buffer data avec la chaine de caractère passé en argument (le précédent contenu de data est perdu)
		if(locked) return(0);
		if(msg==null) msg="";
		//retourne le nombre de caractère effectivement fixé dans le buffer data.    NB: au maximum les data.length premier caractère sont conservés !
		byte[] tab = msg.getBytes(StandardCharsets.US_ASCII);
		if(tab.length>wpos) expends(tab.length-wpos);
		wpos=tab.length;
		for(int i=0; i<wpos; i++) seed.data[i]=tab[i];
		return(wpos);
	}
	public boolean writeShort(int value){ //écrit la valeur value sur 2 octets (poids faible, puis poids fort) 
		if(locked) return(false);
		expends(2);

		seed.data[wpos++]=(byte) ((value >> 8) & 0xFF);
		seed.data[wpos++]=(byte) ((value & 0xFF));
		return(true);
	}	
	public boolean writeWord(int value){ //écrit la valeur value sur 4 octets (poids faible -> poids fort) 
		if(locked) return(false);
		expends(4);

		seed.data[wpos++]=(byte) ((value >> 24) & 0xff);
		seed.data[wpos++]=(byte) ((value >> 16) & 0xff);
		seed.data[wpos++]=(byte) ((value >> 8) & 0xFF);
		seed.data[wpos++]=(byte) ((value & 0xFF));
		return(true);
	}
	public boolean writeBruteString(String msg){ 	//copie les octets de la chaine dans le buffer, sans la faire précédé de sa taille
		if(locked) return(false);
		if(msg==null) msg="";
		int length = msg.length();
		expends(length);

		byte[] tab = msg.getBytes(StandardCharsets.US_ASCII);
		for(int i=0; i<length; i++) seed.data[wpos++]=tab[i];
		return(true);
	}
	public boolean copyBytes(byte[] data){ //copie le buffer data
		return(copyBytes(data, 0, data.length));
	}	
/*	public boolean copyIp(IpPort target){ //copie le buffer data
		if(target.isLocalAddress()) return(copyTBytes(IpPort.byteLocalIp));
		else return(copyTBytes(target.byteIp()));
	}*/		
	
	public boolean copyBytes(byte[] data, int start, int length){ //copie le buffer data[start, start+length[ 
		if(locked) return(false);
		expends(length);
		int end = start+length;
		for(int s=start; s<end; s++) seed.data[wpos++]=data[s];
		return(true);
	}
	//]
}