package skuad.util.value;

import skuad.util.value.Value;
import skuad.util.value.ValueReader;
import skuad.util.value.ValueWriter;
import skuad.util.value.implementation.*;

/**
 * Interface générique d'un objet portant une information typée qui peut être atomique ou structurés.<br/>
 * <br/>
 * <ul>
 * 		<li>Atomique : pour une information unitaire de type primitive ({@link #BOOL}, {@link #INT}, {@link #FLOAT} ou {@link #STRING}).</li>
 * 		<li>Tableau : pour une information composite rassemblant des sous-informations ordonnées dans une liste à indice numérique ({@link #ARRAY}).</li>
 * 		<li>Tuple (ou table de hachage) : pour une information composite rassemblant des sous-informations étiquettées par une chaîne de caractères ({@link #HASH}).</li>
 * </ul>
 * <br/>
 * Ces méthodes de lecture vont renvoyer une réponse la plus adpatée possible au type réel de l'information qu'elle véhicule en suivant cette table de conversion :<br/><br/>
 * <table border=1>
 *  <tr><th>de (ligne) \ vers (colonne)</th>	<th>STRING</th>			<th>INT/FLOAT</th>		<th>BOOL</th>					<th>Array</th>	<th>Hash</th></tr>
 *  <tr>	<th>ERROR</th>						<td>"error"</td>		<td>0</td>				<td>false</td>					<td>null</td>	<td>null</td></tr>
 *  <tr>	<th>NULL</th>						<td>null</td>			<td>0</td>				<td>false</td>					<td>null</td>	<td>null</td></tr>
 *  <tr>	<th>UNDEFINED</th>					<td>null</td>			<td>0</td>				<td>false</td>					<td>null</td>	<td>null</td></tr>
 *  <tr>	<th>STRING</th>			    		<td>=</td>	  <td>convertion ou -1</td>   <td>!=null && !="0" && !="false"</td>	<td>null</td>	<td>null</td></tr>
 *  <tr>	<th>INT/FLOAT</th>					<td>to String</td>		<td>=</td>				<td>!=0</td>					<td>null</td>	<td>null</td></tr>
 *  <tr>	<th>BOOL</th>						<td>true/false</td>		<td>0/1</td>			<td>=</td>						<td>null</td>	<td>null</td></tr>
 *  <tr>	<th>Array</th>			  			<td>to String</td>		<td>0</td>				<td>false</td>					<td>=</td>		<td>null</td></tr>
 *  <tr>	<th>Hash</th>			 			<td>to String</td>		<td>0</td>				<td>false</td>					<td>null</td>	<td>=</td></tr>
 * </table>
 * <br/>
 * Le type réel de l'information peut être consulté via la méthode : {@link #getType()}.
 */
public abstract class Value extends ValueReader {
	//[définition des types (NB: leur valeur corresppond aux décalages binaires du nombre 1 afin d'être utiliser comme masque de convertion)
	/**
	 * Type de la valeur Erreur.
	 */
	public final static byte ERROR=(byte) 0xFF;
	/**
	 * Type pour une information indéfinie.
	 */
	public final static byte UNDEFINED=(byte) 0x00;
	/**
	 * Type d'une information unitaire sous forme de chaîne de caractère.
	 */
	public final static byte STRING	= (byte) (1 << 0);
	/**
	 * Type d'une information unitaire sous forme valeur entière.
	 */	
	public final static byte INT	= (byte) (1 << 1);
	/**
	 * Type d'une information unitaire sous forme de valeur booléenne.
	 */	
	public final static byte BOOL	= (byte) (1 << 2);
	/**
	 * Type d'une information unitaire sous forme valeur décimale.
	 */		
	public final static byte FLOAT = (byte) (1 << 3);
	/**
	 * Type d'une information composite structurée en tableau indexé numériquement.
	 */		
	public final static byte ARRAY = (byte) (1 << 4);
	/**
	 * Type d'une information composite dont les sous-informations sont indexées par une étiquette (chaîne de caractère).
	 */			
	public final static byte HASH	= (byte) (1 << 5);
	/**
	 * Type pour une value de type tableau d'octet
	 */			
	public final static byte BYTES	= (byte) (1 << 6);	
	/**
	 * Type pour une value === null
	 */			
	public final static byte NULL	= (byte) (1 << 7);
	
	
	/**
	 * Valeur constante Erreur
	 */
	public final static ErrorValue ERROR_VALUE=ErrorValue.value;
	/**
	 * Valeur constante pour une information indéfinie
	 */	
	public final static UndefinedValue UNDEFINED_VALUE=UndefinedValue.value;	
	/**
	 * Valeur constante NULL
	 */
	public final static NullValue NULL_VALUE=NullValue.value;
	//]
	
	//[Fabrique
	/**
	 * Valeur constante pour exprimer TRUE
	 */
	public final static Value TRUE=CBoolValue.TRUE;
	/**
	 * Valeur constante pour exprimer FALSE
	 */
	public final static Value FALSE=CBoolValue.FALSE;	
	
	public final static Value i(int v){ return(CIntValue.of(v)); }
	public final static Value b(boolean v){ return(v ? TRUE : FALSE); }
	public final static Value f(float f){ return(new CFloatValue(f)); }
	public final static Value s(String s){ return(new CAtome(s)); }
	public final static Value bs(byte[] bs){ return(new CAtome(bs)); }
	
	//Parseur : "chaine_ljson"	==> ljson = json sans les " imposé sur les noms des champs
	public final static Value parse(String desc){
		return(ValueParser.parse(desc));
	}
	
	public final static ValueWriter writer(){ return(new WValue()); }
}