package skuad.util.value;

import skuad.util.value.Value;

public abstract class ValueReader {
	
	public abstract Value toConstant();
	
	//[Méta-description de la valeur
	/**
	 * Lecture de la chaine de description/commentaire associée à la valeur.<br/>
	 * <br/>
	 * Si aucune description n'est spécifié pour la valeur, cette méthode retournera une chaine vide.<br/>
	 * @return la chaine de description ou commentaire associée à la valeur.
	 */
	public abstract String description();
	public boolean haveDescription(){
		String desc=description();
		return(desc!=null && desc.length()>0);
	}
	//]
	
	//[Analyse du type 
	/**
	 * Lecture du type réelle de l'information portée par l'objet.
	 * @return L'une valeur : {@link #ERROR}, {@link #UNDEFINED}, {@link #STRING}, {@link #INT}, {@link #FLOAT}, {@link #BOOL}, {@link #BYTES}, {@link #ARRAY} ou {@link #HASH} 
	 */
	public abstract byte getType();
	/**
	 * Teste si l'information portée par l'objet est de forme unitaire.
	 * @return <i>true si et seulement le type de l'information est {@link #STRING}, {@link #INT}, {@link #FLOAT} ou {@link #BOOL}
	 */
	public abstract boolean isAtome();
	/**
	 * Teste si l'information portée par l'objet est de forme composite structurée en tableau indexé numériquement.
	 * @return <i>true si et seulement le type de l'information est {@link #ARRAY}
	 */	
	public abstract boolean isArray();
	/**
	 * Teste si l'information portée par l'objet est de forme composite dont les sous-informations sont indexées par une étiquette (chaîne de caractères).
	 * @return <i>true si et seulement le type de l'information est {@link #HASH}
	 */		
	public abstract boolean isHash();
	
	public abstract boolean equals(Value v);
	//]
	
	//[Lecture de valeur
		//[si type est un ATOME (STRING -> s , INT -> i, BOOL -> b, FLOAT -> f)
	/**
	 * Lecture de l'information sous forme de chaine de caractère.
	 * @return Chaîne de caractère correspondante à la valeur de l'information.
	 */
	public abstract String toString();
	public String s(){ return(toString()); }
	/**
	 * Lecture de l'information sous forme de valeur entière.
	 * @return Valeur entière correspondante à la valeur de l'information.
	 */	
	public abstract int toInt();
	public int i(){ return(toInt()); }
	/**
	 * Lecture de l'information sous forme de valeur booléenne.
	 * @return valeur booléenne correspondante à la valeur de l'information.
	 */	
	public abstract boolean toBool();
	public boolean b(){ return(toBool()); }
	/**
	 * Lecture de l'information sous forme de valeur décimale.
	 * @return Valeur décimale correspondante à la valeur de l'information.
	 */
	public abstract float toFloat();
	public float f(){ return(toFloat()); }
	/**
	 * Lecture de l'information sous forme de tableau d'octet.
	 * @return Valeur tableau d'octet correspondante à la valeur de l'information.
	 */
	public abstract byte[] toBytes();
	public byte[] bs(){ return(toBytes()); }	
		//]
		//[si type==ARRAY 
	/**
	 * Test l'existence d'une position.<br/>
	 * @param p indice de la position à tester
	 * @return <b>true</b> si le champ existe, <b>false</b> sinon.
	 */
	public abstract boolean exist(int p);	
	/**
	 * Lecture du nombre de sous-informations portées par l'objet.
	 * @return 0 si l'information portée est de type unitaire (atome), sinon renvoie le nombre de sous-éléments de l'information composite portée par l'objet<br/>
	 * 		(soit le nombre d'éléments du tableau, ou le nombre de couples "étiquette x sous-information"). 
	 */
	public abstract int size();
	/**
	 * Retourne la sous-information placée à la position <i>p</i> du tableau d'informations portées par l'objet.
	 * @param p Position de la sous-information à récupérer.
	 * @return Valeur de la sous-information recherchée, ou {@link #UNDEFINED_VALUE} si la position demandée n'existe pas, ou {@link #ERROR_VALUE} si l'objet
	 * 		n'est pas du type {@link #ARRAY}.
	 */
	public abstract Value get(int p);
			//accesseur ARRAY rapide
	/**
	 * Retourne, sous forme de chaine de caractères, la sous-information placée à la position <i>p</i> du tableau d'informations portées par l'objet.
	 * @param p Position de la sous-information à récupérer.
	 * @return Chaine de caractères correspondante à la sous-information recherchée, ou <i>null</i> si la position demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #ARRAY}.
	 */
	public abstract String getString(int p);
	/**
	 * Retourne, sous forme de valeur entière, la sous-information placée à la position <i>p</i> du tableau d'informations portées par l'objet.
	 * @param p Position de la sous-information à récupérer.
	 * @return Valeur entière correspondante à la sous-information recherchée, ou <i>-1</i> si la position demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #ARRAY}.
	 */	
	public abstract int getInt(int p);
	/**
	 * Retourne, sous forme de valeur booléenne, la sous-information placée à la position <i>p</i> du tableau d'informations portées par l'objet.
	 * @param p Position de la sous-information à récupérer.
	 * @return Valeur booléenne correspondante à la sous-information recherchée, ou <i>null</i> si la position demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #ARRAY}.
	 */	
	public abstract boolean getBool(int p);
	/**
	 * Retourne, sous forme de valeur décimale, la sous-information placée à la position <i>p</i> du tableau d'informations portées par l'objet.
	 * @param p Position de la sous-information à récupérer.
	 * @return Valeur décimale correspondante à la sous-information recherchée, ou <i>-1</i> si la position demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #ARRAY}.
	 */	
	public abstract float getFloat(int p);
	
	public abstract byte[] getBytes(int p);
		//]	
		//[si type==HASH	
	//public int size();
	/**
	 * Test l'existence d'un champ.<br/>
	 * @param p Nom du champ à tester
	 * @return <b>true</b> si le champ existe, <b>false</b> sinon.
	 */
	public abstract boolean exist(String p);
	/**
	 * Retourne la sous-information associé à l'étiquette <i>p</i> dans l'information composite portée par l'objet.
	 * @param p Etiquette de la sous-information à récupérer.
	 * @return Valeur de la sous-information recherchée, ou {@link #UNDEFINED_VALUE} si l'étiquette demandée n'existe pas, ou {@link #ERROR_VALUE} si l'objet
	 * 		n'est pas du type {@link #HASH}.
	 */	
	public abstract Value get(String p);
	/**
	 * Retourne la liste des étiquettes existantes dans l'information composite portée par l'objet.
	 * @return Liste, sous forme d'itérateur, des étiquettes existantes dans l'information composite portée par l'objet, ou <i>null</i> si l'objet
	 * 		n'est pas du type {@link #HASH}. 
	 */		
	public abstract Iterable<String> keys();
			//accesseur HASH rapide
	/**
	 * Retourne, sous forme de chaine de caractères, la sous-information associée à l'étiquette <i>p</i> dans l'information composite portée par l'objet.
	 * @param p Etiquette de la sous-information à récupérer.
	 * @return Chaine de caractères correspondante à la sous-information recherchée, ou <i>null</i> si l'étiquette demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #HASH}.
	 */	
	public abstract String getString(String p);
	/**
	 * Retourne, sous forme de chaine de caractères, la sous-information associée à l'étiquette <i>p</i> dans l'information composite portée par l'objet.
	 * @param p Etiquette de la sous-information à récupérer.
	 * @param defaultValue valeur par défaut retourné si l'étiquette n'exite pas.
	 * @return Chaine de caractères correspondante à la sous-information recherchée, ou <i>null</i> si l'étiquette demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #HASH}.
	 */	
	public String getString(String p, String defaultValue){
		if(!exist(p)) return(defaultValue);
		else return(getString(p));
	}

	/**
	 * Retourne, sous forme de valeur entière, la sous-information associée à l'étiquette <i>p</i> dans l'information composite portée par l'objet.
	 * @param p Etiquette de la sous-information à récupérer.
	 * @return Valeur entière correspondante à la sous-information recherchée, ou <i>null</i> si l'étiquette demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #HASH}.
	 */		
	public abstract int getInt(String p);
	/**
	 * Retourne, sous forme de valeur entière, la sous-information associée à l'étiquette <i>p</i> dans l'information composite portée par l'objet.
	 * @param p Etiquette de la sous-information à récupérer.
	 * @param defaultValue valeur par défaut retourné si l'étiquette n'exite pas.
	 * @return Valeur entière correspondante à la sous-information recherchée, ou <i>null</i> si l'étiquette demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #HASH}.
	 */	
	public int getInt(String p, int defaultValue){
		if(!exist(p)) return(defaultValue);
		else return(getInt(p));
	}
	/**
	 * Retourne, sous forme de valeur booléenne, la sous-information associée à l'étiquette <i>p</i> dans l'information composite portée par l'objet.
	 * @param p Etiquette de la sous-information à récupérer.
	 * @return Valeur booléenne correspondante à la sous-information recherchée, ou <i>null</i> si l'étiquette demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #HASH}.
	 */		
	public abstract boolean getBool(String p);
	/**
	 * Retourne, sous forme de valeur booléenne, la sous-information associée à l'étiquette <i>p</i> dans l'information composite portée par l'objet.
	 * @param p Etiquette de la sous-information à récupérer.
	 * @param defaultValue valeur par défaut retourné si l'étiquette n'exite pas.
	 * @return Valeur booléenne correspondante à la sous-information recherchée, ou <i>null</i> si l'étiquette demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #HASH}.
	 */		
	public boolean getBool(String p, boolean defaultValue){
		if(!exist(p)) return(defaultValue);
		else return(getBool(p));
	}	
	/**
	 * Retourne, sous forme de valeur décimale,  la sous-information associée à l'étiquette <i>p</i> dans l'information composite portée par l'objet.
	 * @param p Etiquette de la sous-information à récupérer.
	 * @return Valeur décimale correspondante à la sous-information recherchée, ou <i>null</i> si l'étiquette demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #HASH}.
	 */		
	public abstract float getFloat(String p);
	/**
	 * Retourne, sous forme de valeur décimale,  la sous-information associée à l'étiquette <i>p</i> dans l'information composite portée par l'objet.
	 * @param p Etiquette de la sous-information à récupérer.
	 * @param defaultValue valeur par défaut retourné si l'étiquette n'exite pas.
	 * @return Valeur décimale correspondante à la sous-information recherchée, ou <i>null</i> si l'étiquette demandée n'existe pas
	 *    ou si l'objet	n'est pas du type {@link #HASH}.
	 */	
	public float getFloat(String p, float defaultValue){
		if(!exist(p)) return(defaultValue);
		else return(getFloat(p));
	}	
	
	public abstract byte[] getBytes(String p);
	public byte[] getBytes(String p, byte[] defaultValue){
		if(!exist(p)) return(defaultValue);
		else return(getBytes(p));
	}
		//]
	//]
}