/**
Définition et vocabulaire :
===========================
	- HubUdp est un intergiciel (middleware) qui donne la possibilité de créer des sous réseaux virtuels, appelé "hub", permettant à des contextes d'exécution
	(ensemble des éléments d'un programme en cours d'exécution) situés sur des machines différentes, ou sur une même machine, d'échanger des messages entre eux.
	- Chaque membre d'un hub (i.e. chaque noeud du réseau virtuel que constitue le hub) est appelé un "spot".
	Ainsi : un hub permet de créer un réseau virtuel interconnectant des spots entre eux.
	- Pour un contexte d'éxécution donnée, le spot qui le représente dans un hub est appelé le "spot local". Un contexte d'exécution dispose au maximum d'un spot
	local par hub (le spot local centralise les ressources de communication locale du contexte d'exécution pour ce hub).
	NB: un contexte d'exécution peut comporter d'autres spots que son spot local dans un hub donné, ce sont alors des "spots proxy" qui sont créés dans le cas ou le 
	spot local joue un rôle de relais pour un portail, ou de point de fusion entre 2 hubs (voir ci-dessous). Ces spots proxy représentent en réalité des contextes
	d'exécution distants (en général extérieur au réseau local) et permettent de rendre ces contextes accessibles pour échanger des messages via le hub.

	HubUdp offre les services suivant :
	- HubUdp est un intergiciel zero-config, aucun réglage n'est nécessaire pour son utilisation.
	- HubUdp est trans-plateforme (Windows, Linux, Mac, Arduino, ...) et multi-language (Java, C, Javascript, Swift, Android, ...), ainsi les spots de
	plateformes et de languages différents sont tous interopérables. 
	- plusieurs hubs différents peuvent être exploités sur le même réseau physique (1). Chaque hub est identifié par un nom (2), et les hubs sont virtuellement isolés 
	les uns des autres (ainsi un contexte d'exécution exploitant 2 hubs aura pour chacun d'eux un spot local distinct).
	- on peut créer des hubs publics (à accés libre, avec transmissions non cryptés) ou privés (accés contrôlé par mot de passe, et transmissions cryptés [*]).
	
	Un hub offre les services suivants :
	- C'est un système d'échange de message de niveau transport, c'est à dire qu'une fois qu'un message est transportés jusqu'à destination il est remis 
	à un gestionnaire de réception qui traitera le message (seul le transport est assuré par HubUdp). Ainsi, pour être en mesure de traiter les messages reçu dans 
	un contexte d'exécution on doit au minima enregistrer une instance de MessageReceiver sur l'hub concidéré.
	- Plusieurs gestionnaires de réception peuvent être enregistrés sur le même hub, chacun associé à un "canal" spécifique. Le numéro de canal est spécifié 
	au moment de l'envoie d'un message afin que le gestionnaire de réception approprié puisse être identifié à destination. Cela donne ainsi la possibilité d'avoir 
	plusieurs protocoles applicatifs distincts sur le même substrat de communication, chacun des protocoles devant être associé à un canal dédié.
	- Il est possible d'enregistrer plusieurs gestionnaires de réception (MessageReceiver) sur le même numéro de canal, il recevront alors tous les mêmes messages
	reçus sur ce canal.
	- le transport des messages utilisent de faible ressource réseau (mode UDP fiablisé par acquitement, réordonnancement, découpage et recomposition
	  automatique des messages).
	- l'envoie d'un message peut cibler un spot ou tous les spots du hub (broadcast)
	- on peut être informé de la présence des spots (connexion/déconnexion) par des notifications (pattern observeur)
	- des méta-données peuvent être fixé sur un spot afin de fournir des informations descriptives sur les caractéristiques ou les capacités du contexte
	d'exécution auquel il est associé.
	- 2 type de méta-donnée sont supportées :
	 		- les SMD (Small Méta Data), qui sont des méta-données primitives de petite taille (nombre, chaine de caractère ou petit tableau d'octet (<256 octets))
	 		- les TAG, qui sont des méta-données structurées (des instances de Value)
	- quand les méta-données SMD d'un spot sont modifiées, les modifications sont automatiquement notifiées sur l'ensemble du hub (pattern observeur).
	- les méta-données SMD décrivant le type de plateforme (Windows, Linux, Mac, Arduino, ...), le langage support (Java, C, Javascript, Swift, Android, ...),
	et l'id global de chaque spot sont automatiquement renseignées.
	- la zone de porté d'un hub est celle du réseau local, mais des clients externes à ce réseau peuvent se connecter au hub via un composant portail
	qui peut être activé par l'un des spots déjà présent dans le hub. Il est aussi possible de créer des points de "fusion" entre deux hubs afin que chacun des deux hubs
	perçois tous les spots de l'autre (et permet donc la communication sur l'ensemble des contextes d'exécution des deux hubs).
	- quand un contexte d'exécution se connecte à un hub via un portail on dira que le hub est activié, et du point de vue des codes qui se connecteront à ce portail 
	le hub sera qualifié de "hub distant".
		
	[*] le cryptage des échanges n'est pas encore effectif, il sera mis en place dans une prochaine version mais cela sera transparent
	    pour la couche application (aucune modification de code ne sera nécessaire).

Rmq :
=====
	(1) la seule limite est de 65000 spots (locaux et proxy, tout contexte d'exécution et hub confondus) par hote (machine), ainsi tous les contextes d'exécution d'un 
	    même hote sont liés par cette contrainte.
	(2) le nom d'un hub ne peut pas contenir le symbole ':'
	(3) -> le numéro de canal est un octet => un maximum de 255 canaux différents par hub peut être supporté.
		-> l'unicité de la correspondance sémantique "numéro de canal" -> "type de gestionnaire de réception" doit être garanti par la couche applicative sur l'ensemble du hub
		   /!\ faire attention à cela en cas de fusion de 2 hubs.
		-> une bonne pratique à ce niveau est dé choisir un numéro de canal de façon aléatoire pour chaque protocole applicatif (donc : type de gestionnaire de réception) défini,
		le risque de colision sera alors de (n-1)/255 si le hub ciblé comporte n protocole applicatif distinct.
		NB: le canal 0 est généralement employé comme un canal "libre" (-> simple échange de message texte).

Comportement d'un hub :
=======================
==> quand on rejoint un hub local (suite à l'appel à la méthode joinLocalHub(...)) :
- le hub entre dans l'état INITIALIZING (méthode getState()), et l'évènement hub du même nom est déclenché avec en paramêtre l'id du hub (nom + ':' + hash_key)
- Tant que le hub ne passe pas dans l'état HUB_READY : 
	- aucun envoie de message n'est possibe sur le hub et aucune notification n'est déclenchée.
 	- le spotlocal reste down, dans cet état les opérations concernant les méta-data sont possibles (lecture/écriture) sur le spot mais aucune notification de changement 
 	SMD/idContext n'est déclenché (elles le seront toutes une fois que le spot passera dans l'état up) et aucun envoi de message n'est possible vers ce spot.
- en cas d'erreur d'initialisation (nombre maximal des 65000 spots atteind sur l'hote), le hub passe dans l'état HUB_ERROR l'évènement hub du même nom est déclenché,
ce hub restera alors inopérationnel.
- une fois l'initialisation terminé le hub va passer dans l'état HUB_READY, le spotlocal passera dans l'état up et l'évènement hub HUB_READY sera déclenché. C'est aussi
à ce moment la que toutes les notifications up/updateSMD/IdContext nécessaire seront notitifé au écouteur déjà enregistré.
- le hub pourra alors être utilisé normalement.
- cette forme de hub restera dans l'état HUB_READY tout le long de l'exécution du contexte local.
	
==> quand on rejoint un hub distant (suite à l'appel à la méthode joinDistantHub(...)) : 
(1) - Le hub entre dans l'état TRY_JOIN_PORTAL (méthode getState()), et l'évènement hub du même nom est déclenché avec en paramêtre la chaine ip_portal:port_portal.
(2) - Tant que le hub ne passe pas dans l'état HUB_READY : 
	- aucun envoie de message n'est possibe sur le hub et aucune notification n'est déclenchée.
 	- le spotlocal reste down, dans cet état les opérations concernant les méta-data sont possibles (lecture/écriture) sur le spot mais aucune notification de changement 
 	SMD/idContext n'est déclenché (elles le seront toutes une fois que le spot passera dans l'état up) et aucun envoi de message n'est possible vers ce spot.
(3) - Une fois la connexion au portail établie, le hub passe dans l'état INITIALIZING et l'évènement hub du même nom est déclenché avec en paramêtre la chaine ip_portal:port_portal.
(4) - suite à cette connexion, Si le portail n'est pas activé : le hub va passer dans l'état PORTAL_DISABLED et l'évènement hub du même nom est déclenché avec en paramêtre la chaine ip_portal:port_portal.
(5) - dès que le portail sera activé le hub va passer dans l'état PORTAL_ENABLED et l'évènement hub du même nom est déclenché avec en paramêtre la chaine ip_portal:port_portal.
(6) - suite à cela le spot local va passer à l'état up, et ke hub va passer dans l'état HUB_READY et l'évènement hub du même nom est déclenché avec en paramêtre le nom du hub.
(7) - tous les spots existant vont être notifié et toutes les opérations seront possible
(8) - si le portail se désactive le hub va repasser dans l'état PORTAL_DISABLED et l'évènement hub du même nom est déclenché avec en paramêtre la chaine ip_portal:port_portal.
tous les spots passeront alors à down. On reprend alors à partr de (5)
(9) - si la connexion avec le portail tombe on reprend à (1)
	
API :
=====
  *************** HubUdp (méthodes statiques) ***
	HubUdp.joinLocalHub([name] [, key])		<- rejoint (si ce n'est pas déjà fait) le hub spécifié par (name, key).	Cela va entrainer la création du spot local pour ce hub.
	  							- si name n'est pas spécifié, sa valeur par défaut sera 'default'
	  							- si key est spécifié il s'agira d'un hub privé
	  			=> la réponse à cette demande est asynchrone et passera par une notification HubListener de code event==HubUdp.HUB_READY
	  							 
	HubUdp.joinDistantHub(ip_portal, port_portal [,key])	<- rejoint le hub distant relié au portail ip_portal:port_portal, si le hub est privé la bonne valeur de key doit être fournit
				=> la réponse à cette demande est asynchrone et passera par une notification HubListener de code event==HubUdp.HUB_READY
				=> de même, les activiations (concrétisation du lien entre le portail et le hub) et désactivations (suppresion du lien entre le portail et le hub)
					conduirons à des notifications HubListener de code event==HubUdp.PORTAL_ENABLED et HubUdp.PORTAL_DISABLED
				--> si le portail n'est pas actif les spots qui lui sont connectés restent isolés (ils se voient chacun seul dans le hub).

					
	HubUdp.newPortal(port_portal, link_key)	<- crée une instance de portail de jonction sur le port port_portal. Cette méthode ne crée pas de hub,
	 											le hub sera celui qui sera relié au portail via un appel à hup.linkPortal(...)  (voir ci-dessous)
	 											L'argument link_key spécifie le mot de passe qui devra être utilisé lors de l'activation du portail 
	 											(c'est à dire au moment de la demande de liaison du portail au hub via l'appel hup.linkPortal(...) ).
	
	void addHubListener(HubListener obs)		<- ajoute un écouteur de hub, en particulier :
												- la méthode hubEvent(...) de cet écouteur sera déclenché avec le code event==HubUdp.HUB_READY pour chacun des hubs 
												  reliés au contexte d'exécution local (également pour ceux déjà reliés avant l'enregistrement du listener).
												  
	void removeHubListener(HubListener obs)	<- supprime un listener précédemment enregistré
	
  *************** Hub : administration *** 
	hub.addReceiver(int channel, ProtocoleManager manager)		//ajoute un gestionnaire de réception pour le spot local 
										(-> manager.receive(sender, content_paquet, channel) sera invoqué sur chaque paquet recu)
	hub.removeReceiver(int channel, ProtocoleManager manager)	//supprime le gestionnaire précédemment enregistré
	hub.broadcast(....)	-> pour emettre des messages en mode broadcast (à tous les spots du hub)

	hub.linkPortal(ip_portal, port_portal , link_key)	<- relie le hub au portail de jonction ip_portal:port_portal, la bonne valeur de link_key doit être fournit.
		--> cette opération conduit à activer le portail. Si le portail est déjà activé l'appel hup.linkPortal(...) échoue  (un portail ne peut être relié qu'à un unique hub).
		--> le spot local devient le spot relais de ce portail : il va accueillir un connecteur (c'est à dire slot d'échange appelé "spot proxy") pour chacun des spots distant 
		(dit "spot client") qui sont connectés (ou se connecteront) au portail via la méthode HubUdp.joinDistantHub(...).
		--> si le spot relais arrête son exécution, le portail sera désactive (déconnecté du hub), et tous les spots clients (connectés au portail) seront isolé (ils seront perçus
		comme down par les autres spots du hub).

	hub.localPortal(port_portal) [*] <- crée un portail relié au hub, le spot local devient alors immédiatement le spot relais (cela revient à enchainer les appels HubUdp.newPortal(...)
		et hup.linkPortal(...) dans le même contexte d'exécution).
		-> on appel cela un portail intégré (directement porté par le spot relais), par opposition au portail précédent (qu'on qualifie de portail distant).
		Un portail intéré (s'il existe) est toujours activé, alors qu'un portail distant nécessite une opération de liaison explicite hup.linkPortal(...) pour être activé. 
		Mis à part les opérations de création et d'activation, il n'y a pas de différence d'usage entre un portail intégré et un portail distant : pour chacun d'eux 
		il faut utiliser l'appel HubUdp.distantHub(...) pour rejoindre le hub auquel ils sont liée.
		
		Avantages/inconvénients entre "portail intégré" et "portail distant" :
		- Un portail intégré comsomme moins de ressource car il ne mobilise qu'un unique contexte d'exécution pour le portail et le spot relais.
		- Un portail distant permet une connexion asynchrone des spots clients et du spot relais, alors qu'un portail intégré nécessite impérativement que le spot relais se 
		connecte toujours en premier (c'est lui qui crée le portail). Et de ce fait : en cas de désactivation du portail, tous les spots clients seront déconnectés et devront 
		renouveller l'appel HubUdp.distantHub(...) pour revenir dans le hub et il ne pourront pas le faire tant que le spot relais ne sera pas à nouveau UP.
		- Un portail distant facilite le déploiement dans une architecture WAN (tel qu'Internet) dans la mesure ou ce type de portail constitue un point fixe externe directement
		atteignable par des contextes d'exécution évoluant dans des LAN (réseaux locaux) distincts.
		Aussi, en général, on utilisera un portail intégré à l'interface de 2 reseaux locaux (sur la machine passerelle), et un portail distant pour relier une machine à un hub à 
		travers le réseau Internet.
		
	hub_a.fusion(hub_b) [*] <- crée un portail intégré bidirectionnel entre les hub_a et hub_b. Cela revient à créer un portail intégré A sur le hub_a et un portail intégér B sur le hub B,
		puis à produire un spot proxy pour chacun des spots du hub_b sur le portail A, et inversément un spot proxy pour chacun des spots du hub_a sur le portail B. Cela permet de 
		fusionner 2 hubs afin que chacun d'eux contiennent les mêmes spots. On appel ce portail particulier : un "portail de fusion".
		A noter que cette opération transforme le spot local du hub_b en un proxy du spot local du hub_a, les 2 instances n'en forme plus qu'une (elles portent en particulier les mêmes
		méta-donnée : lors de la fusion les méta-donnée précédemment porté par le spot local B sont supprimées, seul ceux du spot local A sont conservées).
		Vocabulaire: les spots local A et B sont appelés "spot de fusion" dans leur hub respectif.
	
	[*] ==> fonctionnalités prévues, elles seront disponibles dans une future version.
	
  ***************  Hub : envoie de message *** 
	boolean broadcast(int channel, DataReader content)	-> emission du message en mode broadcast sur l'ensemble du hub, avec toutes les options activées par défaut
	boolean broadcast(int channel, DataReader content, int option) -> emission du message en mode broadcast sur l'ensemble du hub

	boolean broadcastByte(int channel, byte content, int option)	-> emission du message ne contenant qu'un byte sur l'ensemble du hub
	
	boolean broadcastBuff(int channel, byte[] buff)	-> emission du buffer en mode broadcast sur l'ensemble du hub,  avec toutes les options activées par défaut
	boolean broadcastBuff(int channel, byte[] buff , int offset, int size, int option) -> emission du buffer en mode broadcast sur l'ensemble du réseau

	NB: pour envoyer un message vers un spot en particulier il faut utilise les méthodes send(...) sur l'instance du spot désigant le destintatire du message (spot.send(...)).
	
		*** Hub : gestion des spots *** 
	LocalSpot localSpot();							//-> retourne le descripteur du spot local
	void addSpotObserver(SpotObserver observer); 	//dès son enregistrement l'observer recoit tous les évènements up pour chaque spot UP et les IdContext déjà produit sur ces spots, 
													//ainsi qu'un updateSMD(spot, true) pour les spots qui ont déjà inintialisés leur SMD
	boolean removeSpotObserver(SpotObserver observer);
	void explore(SpotExplorer cb);
	
	Spot getSpotByIdTarget(String idTarget); 			//retourne le descripteur du spot correspondant à l'id idTarget indiqué, null si non trouvé
	Spot getSpotByIdLocal(String local_id);				//retourne le descripteur du spot correspondant à l'id local indiqué, null si non trouvé
	Spot getSpotByIdGlobal(String global_id);			//retourne le descripteur du spot correspondant à l'id global indiqué, null si non trouvé
	Spot getSpotByIdContext(String chaine_contexte, String id_context);
	
	NB:
	Les 4 identifiants d'un spot :
	=============================
- idLocal = identifiant dynamique relatif (différent d'un point à l'autre du réseau) permettant de désigner un spot d'un hub de façon unique sur un hote donné 
  	--> persistant : non	, 	différenciation dynamique : moyenne (1 sur 65000)
  	--> portée : tous les contextes d'exécutions d'une même machine hote
  	--> encodage minimal : 2 octets (un entier court)
  	--> politique de collision : - collision impossible -
  	--> disponibilité : immédiate
  	--> comportement pour des spots proxy : différenciant 
  	(dans le code d'implémentation interne, la valeur idLocal est appelé ctarget (pour Code Target, qu'il ne faut pas confondre avec l'idTarget décrit ci-dessous)

- idTarget = ip#idLocal : identifiant dynamique (change potentiellement quand le contexte d'exécution porteur est relancé) et relatif (différent d'un point à l'autre du réseau), 
  qui permet de désigner de façon unique un spot sur un réseau sans fusion (tout contexte d'exécution, hub et portail confondu). 
  	--> persistant : non 	, 	différenciation dynamique : moyenne (1 sur 65000)
  	--> portée : l'ensemble du réseu hubudp, sans hub fusionné
  	--> encodage minimal : 6 octets  
  	--> politique de collision : - collision impossible -
  	--> disponibilité : immédiate
  	--> comportement pour des spots proxy : différenciant (la valeur pour le spot proxy sera différente de celle du spot terminal)

- idGlobal = identifiant dynamique permettant de désigner un spot de façon unique sur l'ensemble d'un hub 
  	--> persistant : non	,	différenciation dynamique : trés forte ( + 200 000 milliards) 
  	--> portée : l'ensemble d'un hub donnée
  	--> encodage minimal : 8 octets
  	--> politique de collision : spot A et spot B 
  			- A non proxy et B non proxy : celui de plus haut idTarget subi une déco-reco (le cas d'égalité d'idTarget n'est pas possible) et change son diGlobal
  			- A non proxy et B proxy : B est masqué (assimilé à A)
  			- A et B proxy :  l'un d'eux (au hasard) est masqué (assimilé à l'autre).
	--> disponibilité : dès réception de la 1ère définition SMD du spot
	--> comportement pour des spots proxy : transparent (même valeur pour le spot proxy et le spot terminal)
	
- idContext = identifiant semi-static permettant de désigner un contexte d'exécution (tout hub confondu) de façon unique sur l'ensemble du réseau 
  	--> persistant : oui	, méthode : par chaine contextuelle (un identifiant idContext par chaine contextuelle)
  	--> portée : l'ensemble du réseau
  	--> encodage minimal : 6 octets
	--> politique de collision : celui de plus fort idGlobal modifie son idContext	
	--> disponibilité : immédiate
	--> comportement pour des spots proxy : transparent
	NB: tous les spot locaux d'un même contexte (quand celui-ci est connecté à plusieurs hubs) partagent les mêmes valeur d'idContext,
	c'est à dire que quand on fixe la même chaine contextuelle sur 2 spots locaux d'un même contexte, ils auront tous les 2 la même valeur d'idContext
	pour cette chaine contextuelle. Mais le fait de fixer l'idContext sur un spot, ne conduit pas à le fixer sur tous les autres.
	
	
  *************** API Spot *** 
 	boolean isLocalSpot();	//true s'il s'agit de le spot local
	boolean isUp();
	String getIpPort();		//ip:port du contexte portant ce spot
	void addSpotObserver(SpotObserver observer);	//dès son enregistrement l'observer recoit tous les évènements IdContext déjà produit sur le spot, 
													//ainsi qu'un updateSMD(spot, true) si le spot à déjà inintialisé ses SMD
	boolean removeSpotObserver(SpotObserver observer);
	
	//les identifiants
	String getIdTarget();	//id target du spot
	int getIdLocal();		//id locale du spot
	String getIdGlobal();	//id globale du spot (null si cet identifant n'est pas encore disponible, il ne l'est qu'à partir de la première notification updateSMD(...)) 
	String getIdContext(String chaine_context);

	 //lecture des méta-donnée SMD (Small Méta-Data)
	boolean readySMD();								//	-> retourne true ssi les SMD du spot sont disponible (première notification updateSMD(...) déjà réalisé)
	int count(int code);							//	-> retourne le nombre d'occurence existante pour le code SMD indiqué
	int length(int code [, int rang]);				//	-> retourne le nombre d'octet de la valeur du code SMD indiqué
	int getInt(int code [, int rang]);				//	-> lecture de la valeur du code SMD indiqué sous forme d'un entier
	boolean haveInt(int code , int value);			//	-> true ssi le code SMD indiqué possède la valeur value
	String getString(int code [, int rang]);		//	-> lecture de la valeur du code SMD indiqué sous la forme d'une chaine de caractère
	byte[] getBuff(int code [, int rang]);			//	-> lecture du buffer de la valeur du code SMD indiqué
	int readBuff(int code, int pos [, int rang]); 	//	-> lecture de l'octet pos de la valeur du code SMD indiqué
 	boolean getBool(int code)
 	
 	//lecture des méta-donnée TAG (méta-données structurées)
 	void getTag(String name, TagReader callback)
 	
 	//envoie de message vers ce spot
	boolean send(int channel, DataReader content)	-> emission du message, avec acquitement et content sera disposé aprés émission
	boolean send(int channel, DataReader content, boolean need_ack, boolean auto_dispose) -> emission du message, dispose() sera appliquée à l'instance content que si auto_dispose==true

	boolean send(int channel, byte content, boolean need_ack)	-> emission du message ne contenant qu'un byte
	
	boolean send(int channel, byte[] buff)	-> emission du buffer buff avec acquittement
	boolean send(int channel, byte[] buff , int offset, int size, boolean need_ack) -> emission du buffer buff
	
  *************** API LocalSpot ***
   		//Spot + :
	String fixIdContext(String ch_contexte); //	<- fixe un idContext sur le spot local pour le chaine contextuelle ch_contexte,
	 										 //retourne la valeur de l'idContext attribué
	
	//écriture de la définition SMD de l'host local
	void setInt(int code, int value [,boolean mode_add]);	//		-> fixe la valeur du code à value sur 1 octet
	void setString(int code, String value [, boolean mode_add]);
	void setBuff(int code, byte[] value [, boolean mode_add]);	//-> fixe la valeur du code à value au format tableau d'octet, value est un buffer ou un tableau	
	void setBool(int code, boolean v)
	
		 //suppression
	void removeAll(int code);				//	-> supprime toute les valeurs pour le code indiqué
	void removeInt(int code, int value);	//	-> supprime le couple (code, value) s'il existe
	
	//gestion des Tag
	void addTag(String name, Value value);	 	//ajoute la valeur 'value' au tag de nom 'name'
	boolean haveTag(String name, Value value);	//teste si le tag name possède la valeur value
	boolean addOnce(String name, Value value);	//ajoute la valeur value au tag name s'il ne la possède pas déjà, true si l'ajout est effectif
	boolean removeTag(String name, Value value);//supprime la valeur 'value' du tag de nom 'name'
	boolean clearTag(String name);				//efface toutes les valeurs du tag de nom 'name'
 */
package skuad.hubudp;

import skuad.hubudp.implementation.*;
import skuad.util.base.*;
import java.util.*;

/**
 * The main class to use <i>HubUDP</i>.
 * 
 *
 */
public class HubUdp {
	public final static String VERSION 	 = "0.1_b09";
	
	private static HubManager hubManager = new HubManager();
	
	//[API statique 
	public static String hubOptions(String tab, boolean include_portal) { //retourne la chaine de caractère décrivant les options possibles pour la configuration d'un hub
		String rep =
				tab+	"** Arguments de configuration du hub et du spot local : \n"
				+tab+	" -hub_name HUB_NAME : nom du hub local à rejoindre \n"
				+tab+	" -hub_pass HUB_PASS : mot de passe du hub à rejoindre (local"+(include_portal ? " ou distant":"")+")\n"
				+tab+	" -loghub : activation des logs des évènements hub \n"
				+tab+ 	" -smd CODE VALUE : fixe la chaine VALUE sur le code smd CODE du spot local\n"
				+tab+ 	" -spot_name SPOT_NAME : fixe le nom du spot local\n"
				+tab+ 	" -host_name HOST_NAME : fixe le nom de l'hote\n"
		;
		if(include_portal) {
			rep+= tab+	" -portal IP PORT : IP:PORT du portail du hub à rejoindre [ignoré si -hub_name]\n";
		}
		return(rep);
	}
	
	/**
	 * Join the <i>Hub</i> with the given config. If there is no config, it's the local default <i>Hub</i> that will be joined.
	 * It's possible to specify the name, the password, a portal to join and to activate the events logs.
	 * @param config The <i>String</i> that contains the parameters : -hubname NAME; -hubpass PASS; -portal IP PORT; -loghub
	 * @return <b>Hub</b> Return the <i>Hub</i> that you joined.
	 */
	public static Hub joinHub(String[] config) { //--> connexion à un hub dont la désignation est décrit par le tableau config
		//le tableau config peut typiquement être le trableau des arguments (String[] args) de la fonction main(...) et donc 
		//contenir les paramêtre de lancement du programme.
		//- si config = [ ... "-hub_name", "HUBNAME" , ...  "-hub_pass", "PASSWORD" ...  ] --> le hub local de nom HUBNAME sera rejoint 
		//- sinon, si config = [ ... "-portal", "IP", "PORT"  ... "-hub_pass", "PASSWORD"...  ] --> le hub distant connecté au portail IP:PORT sera rejoint
		//- sinon, c'est le hub local par défaut qui sera rejoint
		return(joinHub(new LaunchOption(config), true));
	}
	
	/**
	 * Join the <i>Hub</i> with the given config. If there is no config, it's the local default <i>Hub</i> that will be joined.
	 * It's possible to specify the name, the password, a portal to join and to activate the events logs.
	 * @param config The <i>String</i> that contains the parameters : -hubname NAME; -hubpass PASS; -portal IP PORT; -loghub
	 * @param can_use_portal If this <i>boolean</i> is on false, it can't use portals.
	 * @return <b>Hub</b> Return the <i>Hub</i> that you joined.
	 */
	public static Hub joinHub(LaunchOption config, boolean can_use_portal) { //--> connexion à un hub dont la désignation est décrit par le tableau config
		Hub hub = null;
		if(config.existField("loghub")) HubUdp.logHubEvent(); //on active les logs des évènements hub
		if(config.existField("hub_name")) {
			String hub_name = config.val("hub_name");
			String hub_pass = config.val("hub_pass", null);
			hub = hubManager.joinLocalHub(hub_name, hub_pass); 	
		} else if(can_use_portal && config.existField("portal")) {
			String ip_portal = config.val("portal");
			int port_portal = config.intVal("portal",1,0);
			String hub_pass = config.val("hub_pass", null);
			if(port_portal>0) hub = hubManager.joinDistantHub(ip_portal, port_portal, hub_pass);
		}
		if(hub==null) hub = hubManager.joinLocalHub(Hub.DEFAULT_HUB, null);
		
		if(hub!=null && (config.existField("smd") || config.existField("spot_name") || config.existField("host_name"))) {
			hub.whenReady((LocalSpot localspot) -> {
				int code = config.intVal("smd",0,-1);
				if(code>0) {
					String value = config.val("smd", 1, null);
					if(value!=null) localspot.setString(code, value);
				}
				String spot_name = config.val("spot_name", null);
				if(spot_name!=null) localspot.setString(Spot.SPOT_NAME, spot_name);
				String host_name = config.val("host_name", null);
				if(host_name!=null) localspot.setString(Spot.HOST_NAME, host_name);
			});
		}
		return(hub); 	
	}
	
	//joinLocalHub([String name [, String key]]) 
	/**
	 * Join the <i>Hub</i> with the given config. If there is no config, it's the local default <i>Hub</i> that will be joined.
	 * It's possible to specify the name, the password, a portal to join and to activate the events logs.
	 * @return <b>Hub</b> Return the <i>Hub</i> that you joined.
	 */
	public static Hub joinLocalHub() { return( hubManager.joinLocalHub(Hub.DEFAULT_HUB, null)); } 
	/**
	 * Join the <i>Hub</i> with the given config. If there is no config, it's the local default <i>Hub</i> that will be joined.
	 * It's possible to specify the name, the password, a portal to join and to activate the events logs.
	 * @param name The <i>String</i> that contains the parameter : -hubname NAME.
	 * @return <b>Hub</b> Return the <i>Hub</i> that you joined.
	 */
	public static Hub joinLocalHub(String name) { return( hubManager.joinLocalHub(name, null)); } 
	/**
	 * Join the <i>Hub</i> with the given config. If there is no config, it's the local default <i>Hub</i> that will be joined.
	 * It's possible to specify the name, the password, a portal to join and to activate the events logs.
	 * @param name The <i>String</i> that contains the parameter : -hubname NAME.
	 * @param key The <i>String</i> that contains the parameter : -hubpass KEY
	 * @return <b>Hub</b> Return the <i>Hub</i> that you joined.
	 */
	public static Hub joinLocalHub(String name, String key) { return( hubManager.joinLocalHub(name, key)); }

	/**
	 * Join the distant <i>Hub</i>(through a portal) with the given IP, Port and Key.
	 * @param ip_portal The IP of the Portal, attached to the <i>Hub</i>, that you want to join.
	 * @param port_portal The PORT of the Portal, attached to the <i>Hub</i>, that you want to join.
	 * @param join_key The password of the Portal, attached to the <i>Hub</i>, that you want to join.
	 * @return <b>Hub</b> Return the <i>Hub</i> that you joined.
	 */
	public static Hub joinDistantHub(String ip_portal, int port_portal , String join_key) { 
		return( hubManager.joinDistantHub(ip_portal, port_portal, join_key));
	}
	/**
	 * Create a new <i>Portal</i> to link to the <i>Hub</i>
	 * @param port_portal The Port which will be use to join this <i>Portal</i>.
	 * @param link_key The password to validate the joining to the <i>Hub</i>.
	 * @return <b>boolean</b> Return <b>true</b> if the portal creation correctly happened.
	 */
	public static boolean newPortal(int port_portal, String link_key) {	//<- crée une instance de portail de jonction sur le port port_portal. Cette méthode ne crée pas de hub,
		return( hubManager.newPortal(port_portal, link_key));
	}
	
	/**
	 * Add the given <i>HubListener</i> to the <i>Hub</i>.
	 * @param obs The <i>HubListener</i> that you want to assign to the <i>Hub</i>.
	 * @see HubListener
	 */
	public static void addHubListener(HubListener obs) { hubManager.addListener(obs); }
	/**
	 * Remove the given <i>HubListener</i> to the <i>Hub</i>.
	 * @param obs The <i>HubListener</i> that you want to unassign to the <i>Hub</i>.
	 * @see HubListener
	 */
	public static boolean removeHubListener(HubListener obs) { return(hubManager.removeListener(obs)); }
	
	public Vector<LocalSpot> haveIdContext(String ch_contex){ return(hubManager.haveIdContext(ch_contex)); } //retourne tous les spots locaux qui ont fixé l'idContext correspondant à la chaine contextuelle indiqué
	
	public static void logHubEvent() { hubManager.logHubEvent(); }	
	//]	
}