package skuad.ubiquity.space.activity;

import skuad.util.value.Value;

/**
 * Interface d'écriture de la facette Activity des agents Ubiquity.<br/>
 * <br/>
 * Cette interface d'écriture est fournie par la méthode {@link Ubiquity.agent.Interactor#myActivity()} de l'handle de l'agent.<br/>
 * Elle permet à l'agent de définir ses activités et de planifier leurs exécutions.<br/>
 * Pour un agent conçu en héritant de la classe {@link Ubiquity.agent.FullParticipantAdpator}, cette interface est aussi disponible sous l'attribut <b>this.activity</b>.<br/>
 * <br/>
 * Pour définir une activité il faut utiliser la méthode {@link #set(String, ActivityRunner)}.<br/>
 * Il y a ensuite trois façons de planifier le déclenchement d'une activité :<br/>
 * 	<ul>
 *  	<li>- une unique fois au bout d'un certain délai (ou immédiatement si le delay=0)  => méthode {@link #run(String, Value, long)}</li>
 *  	<li>- périodiquement à partir d'un certain délai (ou immédiatement si le delay=0)  => méthode {@link #repeat(String, Value, long, long)}</li>
 *  	<li>- à la suite de l'exécution en cours ou déjà planifié => méthode {@link #chain(String, Value, long)}</li>
 *  </ul>
 * <br/>
 * A tout moment le traitement d'une activité peut être annulée avec les méthodes {@link #cancel(String)} ou {@link #remove(String)}.
 * Une activité reste définie (associée et désignable par son identifiant) tant qu'elle n'a pas été supprimée via la méthode {@link #remove(String)}. 
 * <br/><br/>
 * <b>NB :</b> les identifiants des activités sont propres à chaque agent (deux agents différents peuvent exploiter la même valeur d'identifiant d'activité, il s'agira bien de 2 définitions 
 * d'activité distinctes). <br/>
 * <b>NB :</b> toutes les activités liées à un agent seront annulées au moment où l'agent sera retiré du space.
 * <b>NB :</b> si on définit une activité avec un identifiant déjà utilisé, la nouvelle unité d'exécution viendra remplacer celle précédemment associée à l'identifiant.
 */
public interface WActivity extends RActivity {
	
	/**
	 * Facette de lecture.<br/>
	 * <br/>
	 * Cette méthode permet d'obtenir le descripteur de la facette activity en lecture seule.<br/>
	 * C'est ce même objet qui est fournit par la méthode {@link Ubiquity.agent.AIdentity#getActivity()} du descripteur d'identité des agents.<br/>  
	 * @return le descripteur en lecture seule de la facette activity.
	 */
	public RActivity reader();
	
	//[Ajout suppression d'activité
	public byte set(String activity, ActivityRunner runner);
	public byte remove(String activity);
	//]
	
	//[Manipulation des activitées
	/**
	 * Planifie l'exécution d'une activité pour qu'elle soit déclenchée une fois dans 'delay' ms.<br/>
	 * Si l'activité est en cours d'attente d'exécution run/repeat, cette exécution en attente sera annulée et seule cette nouvelle planification sera concidérée.<br/>
	 * Si l'activité est en cours d'exécution, le CancelFlag sera positionné, il n'y aura plus d'occurence repeat sur l'ancienne exécution et seule cette nouvelle planification sera considérée.<br/>
	 * @param id_activity L'identifiant de l'activité concernée.
	 * @param arg Argument qui sera transmis à l'activité au moment de son lancement (dans sa méthode {@link Ubiquity.agent.activity.Activity#action(Interactor, String, Value, CancelFlag)}).
	 * @param delay Durée au bout de laquelle l'activité sera déclenchée (0 si elle doit être déclenché immédiatement).
	 * @return L'identifiant associée à l'activité.
	 */
	public byte run(String activity,Value arg, long delay); 	
	
	/**
	 * L'activité indiquée va être déclenchée dans 'delay' ms puis répétée toutes les 'period' ms.<br/>
	 * Si l'activité est en cours d'attente d'exécution run/repeat, cette exécution en attente sera annulée et seule cette nouvelle planification sera considérée.<br/>
	 * Si l'activité est en cours d'exécution, le CancelFlag sera positionné, il n'y aura plus d'occurence repeat sur l'ancienne exécution et seule cette nouvelle planification sera considérée.<br/>
	 * @param id_activity L'identifiant de l'activité concernée.
	 * @param arg Argument qui sera transmis à l'activité au moment de son lancement (dans sa méthode {@link Ubiquity.agent.activity.Activity#action(Interactor, String, Value, CancelFlag)}).
	 * @param delay Durée en milliseconde au bout de laquelle l'activité sera déclenchée.
	 * @param period Durée en milliseconde entre de deux activations de l'activité.
	 * @return L'identifiant associé à l'activité.
	 */
	public byte repeat(String activity, Value arg, long delay, long period);
	
	/**
	 * Chaine une prochaine exécution de l'activité pour un déclenchement avec l'argument arg.<br/>
	 * L'activité poursuit son éventuelle attente, exécution ou précédent ordre de chainage. Cette commande
	 * fait en sorte que dès que l'activité est disponible son exécution soit relancée une nouvelle fois avec l'argument indiqué.<br/>
	 * Si rien n'est actuellement prévu pour l'activité, elle sera immédiatement déclenchée avec l'argument fourni.<br/>
	 * Si l'activité est en exécution repeat, le cycle en cours se termine et le prochain cycle se fera
	 * avec l'argument fourni (l'activité continuera ensuite son exécution cyclique avec ce même argument).<br/> 
	 * Des appels {@link #chain(String, Value)} peuvent être émis en série, les ordres s'empilent alors de sorte à ce que l'exécution
	 * se fasse les unes après les autres.<br/>
	 * NB: tout autre appel run ou repeat sur l'activité casse la chaine et tout son empilement (tous les ordres de chainage en attente), pour 
	 * laisser place au nouvel ordre de planification.
	 * @param id_activity L'identifiant de l'activité concernée.
	 * @param arg Argument qui sera transmis à l'activité au moment de son lancement (dans sa méthode {@link Ubiquity.agent.activity.Activity#action(Interactor, String, Value, CancelFlag)}).
	 * @param delay Durée minimale d'attente en milliseconde au bout de laquelle l'activité sera déclenchée.
	 * @return L'identifiant associé à l'activité.
	 */
	public byte chain(String activity, Value arg, long delay);
	
	/**
	 * Stop l'activité associé à l'identifiant indiqué.<br/> 
	 * Si l'activité n'a pas été lancée cela n'aura aucune incidence. Par contre si l'activité est lancée par run/repeat/chain :
	 *  <ul>
	 * 		<li>- si l'activité est en attente d'exécution (par exemple si le délai avant lancement n'a pas encore expiré, ou si la période d'attente entre deux cycles n'est pas écoulée) : 
	 *      cette exécution n'aura pas lieu et il n'y aura plus d'occurence repeat ou chain</li>
	 * 		<li>- si l'activité est en cours de traitement : l'indicateur <b>cancel</b> sera positionné sur le moniteur {@link Ubiquity.agent.activity.ProgressFlag} de l'activité et il n'y aura plus d'occurence repeat</li>
	 *  </ul>
	 * @param id_activity L'identifiant de l'activité concernée.
	 * @return <b>true</b> si l'activité existe, <b>false</b> sinon.
	 */	

	public byte cancel(String activity); 
	//]	
}