package skuad.util.thread;

import java.util.*;

/**
 * Pile auto-recyclé de processus.<br/>
 * <br/>
 * Cette classe est un gestionnaire de processus utilisant une pile de Thread gérée avec une stratégie de recyclage automatisée. Son utilitée est d'optimiser les traitements
 * nécessitant de nombreux processus.<br/>
 * <br/>
 * Au lieu de créer un nouveau processus à chaque fois qu'un nouveau traitement doit être initié, cette classe recycle les processus qui ont terminés l'exécution du traitement
 * qui leur ont été préallablement attribué. En pratique un gain significatif est obtenu en terme d'économie de création de thread. Par exemple, 3 à 4 threads peuvent suffire
 * pour des traitements qui en aurait nécessité plusieurs centaines sans cette optimisation.<br/> 
 * <br/>
 * Notez que cette classe est une classe statique, en ce sens elle ne s'instancie pas. Pour l'uitliser on exploite les différents fonctions (méthodes statiques) qu'elle expose.<br/>
 */
public class ThreadStack extends Thread {
		
	/**
	 * Constante fixant le temps d'attente entre chaque déclenchement des activités des waiters.
	 * <br/>
	 * Par défaut la valeur de cette durée est de 5 secondes.<br/>
	 * @see utility.thread.Waiter
	 */
	public final static int WAITERS_FREQUENCY=5000;
	
	/**
	 * Constante spécifiant la durée de conservation maximal d'un thread dans la pile de recyclage.<br/>
	 * <br/>
	 * Par défaut la valeur de cette durée est de 10 secondes.<br/>
	 */
	public final static int BUFFERED_TIMEOUT=10*1000; //durée de conservation des instances bufférisé dans la pile
	
	
	private static Runnable destascker=null;			//tache de dépilage des taches en attente d'exécution
	private static TimerTask destascker_tasck=null;		//timer du processus de dépilage des taches
	public final static int DESTACKER_PERIOD=200;		//période de déclenchement de la tache de dépilage des taches
	
	/**
	 * Lancement immédiat d'un traitement ponctuel.<br/>
	 * <br/>
	 * Cette fonction affecte un thread à l'exécution du traitement fournit en argument. Le traitement est lancé immédiatement.<br/>
	 * 
	 * @param r Instance dont la méthode <b>run</b> spécifie le traitement à réaliser.
	 */
	public synchronized static void perform(Runnable r){
		ThreadStack athread = null;
		if(thread_stack==null){ //on initialise la pile de recyclage
			thread_stack = new LinkedList<ThreadStack>();
			tasck_stack = new LinkedList<Runnable>();
			//on procéde aussi à l'enregistrement du processus de vidange de la pile de recyclage
			ThreadStack.addWaiter(new Waiter(){
				public void run(long ct){
					synchronized(thread_stack) { while(!thread_stack.isEmpty()){
						if((ct-thread_stack.peek().last_push)>BUFFERED_TIMEOUT) { //on supprime toutes les instances périmées
							//ThreadStack t=thread_stack.poll(); if(t!=null) t.close();
							thread_stack.pop().close();
						} else return; //fin de traitement
					} }
				}
			});			
		} else synchronized(thread_stack) { athread = thread_stack.poll(); }
		if(athread==null) { //pas de thread disponible sur la pile
			if(destascker==null) { //on doit créer la tache de dépilage des tâche en attente d'exécution
				destascker = new Runnable(){
					public void run() {
						synchronized(tasck_stack) {
							while(!tasck_stack.isEmpty()) {
								ThreadStack athread;
								synchronized(thread_stack) { athread = thread_stack.poll(); }
								if(athread==null) { athread=new ThreadStack(); nb_thread++; } //on va essayer de créer un nouveau thread
								Runnable tasck = tasck_stack.pop(); 
								try { athread.go(tasck); }
								catch(OutOfMemoryError ome) { //impossible de créer le thread
									tasck_stack.addFirst(tasck); //on replace la tache sur la pile d'attente
									nb_thread--;
									return; //on va attendre le prochain cycle DESTACKER_PERIOD pour une nouvelle tentative
								}
							}
							//on met fin au processus de dépilage des tâche en attente d'exécution
							destascker_tasck.cancel();
							destascker_tasck = null;
						}
					}
				};
			}
			synchronized(tasck_stack) {
				tasck_stack.add(r); 
				if(destascker_tasck==null) destascker_tasck = perform(destascker, 40, DESTACKER_PERIOD);
			}
		} else athread.go(r); //on lance l'activité
	}
	
	/**
	 * Lancement différé d'un traitement ponctuel.<br/>
	 * <br/>
	 * Cette fonction planifie l'exécution du traitement fournit en argument pour qu'il soit déclenché au terme de l'écoulement
	 * du délai spécifié.<br/>
	 * Un thread sera affecté à cette exécution le moment venu.<br/>
	 * 
	 * @param r Instance dont la méthode <b>run</b> spécifie le traitement à réaliser.
	 * @param delay Temps d'attente (en milliseconde) avant de produire l'exécution du traitement.
	 * @return L'instance <b>TimerTask</b> correspondant à cette plannification. Cette instance pourra notamment être utilisé pour annuler
	 * la plannification si le délai avant l'exécution n'est pas écoulé.
	 */
	public static TimerTask perform(Runnable r, long delay){
		TimerTask tt=new DifferedPerfomedStask(r);
		timer.schedule(tt, delay); //mise en place de la planification
		return(tt);
	}
	
	/**
	 * Lancement différé d'un traitement cyclique.<br/>
	 * <br/>
	 * Cette fonction planifie l'exécution du traitement fournit en argument pour qu'il soit déclenché au terme de l'écoulement
	 * du délai spécifié, puis à chaque écoulement de la période indiquée.<br/>
	 * Un thread sera affecté à chacune des exécutions.<br/>
	 * 
	 * @param r Instance dont la méthode <b>run</b> spécifie le traitement à réaliser.
	 * @param delay Temps d'attente (en milliseconde) avant de produire la première exécution du traitement.
	 * @param periode Période du clycle de répétition de l'exécution
	 * @return L'instance <b>TimerTask</b> correspondant à cette plannification. Cette instance pourra notamment être utilisé pour arréter
	 * l'exécution cyclique.
	 */
	public static TimerTask perform(Runnable r, long delay, long periode){
		TimerTask tt=new DifferedPerfomedStask(r);
		timer.schedule(tt, delay, periode); //mise en place de la planification
		return(tt);
	}	

	//[Enregistrement des waiters (traitements à déclencher toutes les WAITERS_FREQUENCY millisecondes, par défaut 5 secondes)
	private static Vector<Waiter> waiters=null;
	private static TimerTask timer_waiters=null;
	
	/**
	 * Enregistrement d'un waiter.<br/>
	 * <br/>
	 * Un waiter est un traitement réccurent qui sera exécuté toutes les {@link #WAITERS_FREQUENCY} millisecondes.<br/>
	 * 
	 * @param w Instance du waiter définissant le traitement à réaliser à chaque déclenchement.
	 * @see utility.thread.Waiter 
	 */
	public synchronized static void addWaiter(Waiter w){ //ajoute un waiter
		if(waiters==null) waiters=new Vector<Waiter>();
		waiters.add(w);
		if(waiters.size()==1){ //on déclenche le timer waiters
			if(timer_waiters==null) timer_waiters=new TimerTask(){
				public void run(){
					synchronized(waiters) { 
						long cur_time=System.currentTimeMillis();
						for(Waiter waiter : waiters) waiter.run(cur_time); 
					}
				}
			};
			timer.schedule(timer_waiters, WAITERS_FREQUENCY, WAITERS_FREQUENCY);
		}
	}
	
	/**
	 * Désenregistrement d'un Waiter.<br/>
	 * <br/>
	 * Cette fonction désenregistre un waiter préallablement enregistré par la fonction {@link #addWaiter(Waiter)}. Cette action met alors fin
	 * à l'exécution récurrente de ce waiter.<br/>
	 * 
	 * @param w Instance du waiter à désenregistrer.
	 * @return <b>true</b> si le désenregistrement a bien été effectué, <b>false</b> si le waiter spécifié n'est pas connu.
	 */
	public synchronized static boolean removeWaiter(Waiter w){ //supprime un waiter
		if(waiters!=null) {
			 if(waiters.remove(w)){
				if(waiters.size()==0){ //on arrête le timer waiters
					timer_waiters.cancel();
					timer_waiters=null;
				}
				return(true);
			}
		}
		return(false);
	}
	//]
	
	//[Gestion du recyclage des threads tournants
	private static Timer timer=new Timer(); //timer interne 
	private static LinkedList<ThreadStack> thread_stack=null;	//pile de recyclage des threads
	private static LinkedList<Runnable> tasck_stack=null;		//pile des taches en attente d'exécution
	private static int nb_thread=0;
	private static int nb_actif_thread=0;
	
	/**
	 * Nombre total de thread créé.<br/>
	 * <br/>
	 * Cette fonction, a  but statistique, permet de consulter le nombre total de threads qui ont été créés depuis le lancement du programme.<br/>
	 * 
	 * @return Nombre total de threads créés.
	 */
	public static int totalNumberOfThread(){ return(nb_thread); }
	
	/**
	 * Nombre de threads qui se trouvent actuellement de la pile de recyclage.<br/>
	 * <br/>
	 * Cette fonction, a  but statistique, permet de consulter le nombre de thread qui sont actuellement en attente dans la pile de recyclage.<br/>
	 * 
	 * @return Nombre threads présent dans la pile de recyclage.
	 */	
	public static int numberWaiterThread(){ return(thread_stack.size()); }
	
	/**
	 * Nombre de threads actifs.<br/>
	 * <br/>
	 * Cette fonction, a  but statistique, permet de consulter le nombre de thread actif.<br/>
	 * <br/>
	 * Un thread actif est un thread qui exécute actuellement un traitement suite à un appel à {@link #perform(Runnable)} ou {@link #perform(Runnable, long)}.<br/>
	 * 
	 * @return Nombre threads actuellement en activité.
	 */		
	public static int numberActifThread(){ return(nb_actif_thread); }
	//]	
	
	//[Implémentation des thread	
	boolean in_run=false;
	Runnable tasck;
	long last_push;			//date de la dernière mise en cache de cet instance
	
	private ThreadStack(){} //cette classe ne s'instancie pas de l'extérieur
	
	private synchronized void go(Runnable r){
		tasck=r;
		if(in_run) notify(); //on réveille le processus
		else start();
	}
	private synchronized void close(){
		if(in_run) {
			in_run=false;
			notify(); //on réveille le processus
		}
	}
	
	/**
	 * Méthode d'éxécution interne.<br/>
	 * <br/>
	 * La classe TheadStack n'étant pas publiquement instanciable (aucun constructeur publique), cette méthode ne peut pas être déclenché. Elle joue un rôle
	 * au niveau du traitement interne du gestionnaire ThreadStack.<br/>
	 */
	public void run(){
		in_run=true;
		while(in_run){
			nb_actif_thread++;
			while(tasck!=null) {
				tasck.run(); 
				synchronized(tasck_stack) { //on regarde si on peu repartir sur une nouvelle tache
					tasck = tasck_stack.poll();
				}
			}
			if(in_run) {
				last_push=System.currentTimeMillis();
				synchronized(thread_stack) {thread_stack.add(this); } //on se replace sur la pile
				nb_actif_thread--;
				try{ synchronized(this) { wait(); } } catch(Exception e){ e.printStackTrace(); }  //on va attendre une prochaine tache à faire
			}
		}
		nb_thread--;
	}
}

class DifferedPerfomedStask extends TimerTask { //classe interne
	Runnable client;
	
	public DifferedPerfomedStask(Runnable c){ client=c; }
	
	public void run(){
		client.run();
		//ThreadStack.perform(client); 
	}
}